--[[
	DuckMod - A World of Warcraft add-on library
	Copyright (C) 2009-2012  Dag Bakken

	DuckMod is free software: you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.

	DuckMod is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with DuckMod.  If not, see <http://www.gnu.org/licenses/>.
]]


local TV=2.9991;

--local DUCKNET_DATA="\2";				-- NEW, old="!"
--local DUCKNET_ENTRY="\3";				-- NEW, old="#"
--local DUCKNET_COMMAND="\4";				-- NEW, old=":"
--local DUCKNET_FREESPLIT1="\5";			-- Free splitter for addons
--local DUCKNET_QUOTATION="\6";			-- NEW, old="\""
--local DUCKNET_WNSPLIT1="\7";			-- WoWnet splitter 1
--local DUCKNET_TREESEP="\8";				-- Tree separator
--local DUCKNET_DISCONNECT="\10";			-- Never use this

local DUCKNET_HEARTBEAT = 1/5;			-- 200ms ticker (5 fps)
local CV2_TICKER = 1/16;				-- 62,5ms ticker (16 fps -> 4KB/s)
local Ambiguate = Ambiguate


if (not DuckMod) then DuckMod = {}; end			-- Make it if empty
if (not DuckMod[TV]) then					-- Don't mess with it if it's loaded
	DuckMod[TV]={};
end

local DM=DuckMod[TV];
local _;

DM.Talk={};
DM.SentBlindWhisperListOfReceivers={};

--
--  DuckMod Communication v2
--
DM.Com2={
	LastTick=0,			-- the ticker
	API={},				-- user API
	Interface={},		-- registered interfaces
	MT={},				-- multi-threading stuff
	PingList={},		-- lists to run ping for
	Sent={},			-- data I've sent
	Build={},			-- Rebuilding input
	Incoming={},		-- incoming addon data
	Loopback={},		-- loopback tester
};
local c2=DM.Com2;


local function dbg(h,t)
	print(h,":");
	if (type(t)~="table") then t={t}; end
	for _i,_d in pairs(t) do print(_i,_d); end
end

-- Set up for handling
function c2:Init()
	if (not self.TheFrame) then
		self.TheFrame=CreateFrame("Frame","DuckMod-Com2-Messager"..TV,UIParent);
		if (not self.TheFrame) then DM:Chat("Could not create DuckMod Com2 UI frame",1); return; end
		self.TheFrame:RegisterEvent("CHAT_MSG_ADDON");				-- Addon channel
		self.TheFrame:SetScript("OnEvent",self.OnEvent);
		self.TheFrame:SetScript("OnUpdate",self.Tick);
	end
	if (not self.TheFrameWorldMap) then
		self.TheFrameWorldMap=CreateFrame("Frame","DuckMod-Com2WM-Messager"..TV,WorldMapDetailFrame);
		if (not self.TheFrameWorldMap) then DM:Chat("Could not create DuckMod Com2 WM frame",1); return; end
		self.TheFrameWorldMap:RegisterEvent("CHAT_MSG_ADDON");				-- Addon channel
--		self.TheFrameWorldMap:SetScript("OnEvent",self.OnEvent);
		self.TheFrameWorldMap:SetScript("OnUpdate",self.Tick);
	end

	if (not DM.DefaultChatHook_AddMessage_Original) then
        DM.DefaultChatHook_AddMessage_Original=DEFAULT_CHAT_FRAME.AddMessage;
        DEFAULT_CHAT_FRAME.AddMessage=DM.DefaultChatHook_AddMessage;
    end
end

function c2:Valid(prefix,ctype)
	if (not prefix) then return nil; end							-- none provided
	for _,inter in pairs(c2.Interface) do						-- walk interfaces
		if (prefix==inter.ID) then									-- found one
			if (ctype and inter.Channel~=ctype) then return nil; end	-- wrong channel
			return true;
		end
	end
	return nil;
end

-- An event has been received
function c2:OnEvent(event,arg1,arg2,arg3,arg4,_,_,_,_,arg9)
	if (event=="CHAT_MSG_ADDON") then
		if arg4 then arg4 = Ambiguate(arg4,"none") end
		if (arg2=="LOOPBACK") then
			c2.Loopback[arg3]=time();
--print("LOOPBACK: Loopback received for",arg3,"at",date("%c",c2.Loopback[arg3]));
			return;
		end
		-- Normal interface input
--		if (c2.Sent[arg2]) then c2.Sent[arg2]=nil; return; end			-- own feedback
		if (arg4==UnitName("player")) then return; end
		if (not c2:Valid(arg1,arg3)) then return; end;	-- Quick way out without further function-calls - which ChannelInput will do
		table.insert(c2.Incoming,{pr=arg1,me=arg2,se=arg4,ch=arg3});	-- feed fifo
		return;
	end
end
--
--	name,guild,level,race,class,zone,classFileName,sex=GetWhoInfo(index)



--	t.cbRX=cbRX;	-- data RX
--		cbRX(interface,data name,data,metadata,sender toon,api used);
--	t.cbVC=cbVC;	-- data version check typically for negotiation
--		cbVC(interface,data name,metadata);
--	t.cbDR=cbDR;	-- a request for data to be sent (typ. from a negotiation)
--		cbDR(interface,data name,metadata);
function c2:CreateInterface(id,cbRX,cbVC,cbDR,eh,ch)
	c2:Init();
	local t={};
--	local i,e,name=1,UnitName("player"):len(),UnitName("player");
--	t.Hash=0; repeat t.Hash=t.Hash+name:byte(i); i=i+1; until (i>e);	-- hash for transmission ID
	t.Hash=DM:Hash(id..UnitName("player"));
	t.Counter=0;														-- transmission counter
	t.Lines={};
	t.Burst={};
	t.BurstToon={};
	t.Incoming={};
	t.Neg={};				-- negotiation
	t.cbRX=cbRX;	-- data RX
	t.cbVC=cbVC;	-- data version check typically for negotiation
	t.cbDR=cbDR;	-- a request for data to be sent (typ. from a negotiation)
	t.EventHandler=eh;
	t.ShareEvent=c2.API.ShareEvent;
	t.SetChannel=c2.API.SetChannel;
	t.SetID=c2.API.SetID;
	t.SetReceiver=c2.API.SetReceiver;
	t.Send=c2.API.Send;
	t.Request=c2.API.Request;
	t.Ping=c2.API.Ping;
	t.IsIdle=c2.API.IsIdle;
	t.Close=c2.API.Close;
	t.ProgressIB=c2.API.ProgressIB;
	table.insert(c2.Interface,t);
	t:SetID(id);					-- prefix and register it with the server
	t:SetChannel(ch or "GUILD");	-- channel
	t:SetReceiver({});				-- no receiver(s)
	return t;
end

function c2:Tick(elapsed)
	c2.LastTick=c2.LastTick-elapsed;
	if (c2.LastTick>0) then return; end
	c2.LastTick=CV2_TICKER;
	local now=time();
	local inter;
    c2.LastTickInterface,inter=next(c2.Interface,c2.LastTickInterface);
	if (c2.LastTickInterface) then
		for hash,neg in pairs(inter.Neg) do								-- walk negotiations
			if (neg.End and neg.End<=now) then							-- valid and timed out
--dbg(inter.ID.." ==>> Negotiation timer done: "..hash,inter.Neg[hash]);
				-- negotiation done
				if (not neg.MyRequest and neg.High==neg.MyVer and neg.Roll==neg.MyRoll) then
					if (inter.cbDR) then inter:cbDR(neg.na,neg.md); end
				end
				wipe(inter.Neg[hash]);									-- wipe negotiate status
				inter.Neg[hash]=nil;									-- done this one
			end
		end

		if (not inter.HaltOutput) then									-- output not halted
			if (inter.Lines[1]) then									-- stuff waiting for transmission
				if (inter.Channel~="WHISPER" and not c2.Loopback[inter.Channel]) then
--print("LOOPBACK: Loopback added to",inter.Channel);
					table.insert(inter.Lines,1,{d="LOOPBACK"});
				end
--print("inter.Lines[1]");
				if (inter.Lines[1].t) then inter.Receiver=inter.Lines[1].t; end	-- there'sa toon for this line
				if (inter.Receiver) then DM.SentBlindWhisperListOfReceivers[inter.Receiver]=now;	-- got one
				elseif (inter.Channel=="WHISPER") then					-- no active receiver, so get one
					inter.HaltOutput=true;								-- halt output until receiver has been found
					DM.MT:Run(tostring(inter).." -> Find pipe",c2.MT.FindPipe,c2.MT,inter);	-- find a piper
					return;
				end
				local temp,_,extra=c2:GetNextOutput(inter.Lines);				-- don't need "toon" at this point as it's gotten earlier if any
--		{ Neg={ na=name, md=meta, High=ver, Roll=0, MyVer=ver, MyRoll=0, MyRequest=true, },	-- extra meta for the engine
				if (extra and extra.Neg) then
--	inter.Neg[hash]={ na=name, md=meta, High=ver, Roll=0, MyVer=ver, MyRoll=0, MyRequest=true, };
--	inter.Neg[hash].End=time()+2;							-- I spoke up, so restart 2 seconds silence
					local _hash=temp:match("^(%d+)_"); _hash=tonumber(_hash);
					inter.Neg[_hash]=extra.Neg;
					inter.Neg[_hash].End=time()+2;
				end
--print("Out:",inter.ID,inter.Receiver,inter.Channel);
--print(temp);
				SendAddonMessage(inter.ID,temp,inter.Channel,inter.Receiver);	-- send it
--				c2.Sent[temp]=true;										-- mark it
			else inter.Receiver=nil; end								-- out-buffer empty, so also remove receiver-toon
		elseif (type(inter.HaltOutput)=="number") then					-- output is halted and timed
--print("inter.HaltOutput");
			if (now>=inter.HaltOutput) then								-- timer done
				wipe(inter.Lines);										-- remove output
				inter.HaltOutput=nil;									-- allow output
			end
		end
	end

	-- Maintain offline user supressor list
	if (next(DM.SentBlindWhisperListOfReceivers)) then
		local toon,val=next(DM.SentBlindWhisperListOfReceivers);
		if (now-val>5) then DM.SentBlindWhisperListOfReceivers[toon]=nil; end
	end

	-- handle incoming buffer
	if (c2.Incoming[1]) then
		local buf=table.remove(c2.Incoming,1);
		c2:ChannelInput(buf.pr,buf.me,buf.se,buf.ch);	-- prefix,message,sender,channel
	end

	-- Run global pings
	if (not c2.RunningPingAll) then
		local _,inter=next(c2.PingList);	-- pick one
		if (inter and inter.ReceiverList and next(inter.ReceiverList)) then
			DM.MT:Run(tostring(inter).." -> Ping Everyone",c2.MT.PingAll,c2.MT,inter);
		end
	end
end

function c2:Tick_old(elapsed)
	c2.LastTick=c2.LastTick-elapsed;
	if (c2.LastTick>0) then return; end
	c2.LastTick=DUCKNET_HEARTBEAT;
	local now=time();
	for _,inter in pairs(c2.Interface) do								-- walk registered interfaces
		for hash,neg in pairs(inter.Neg) do								-- walk negotiations
			if (neg.End and neg.End<=now) then							-- valid and timed out
--dbg(inter.ID.." ==>> Negotiation timer done: "..hash,inter.Neg[hash]);
				-- negotiation done
				if (not neg.MyRequest and neg.High==neg.MyVer and neg.Roll==neg.MyRoll) then
					if (inter.cbDR) then inter:cbDR(neg.na,neg.md); end
				end
				wipe(inter.Neg[hash]);									-- wipe negotiate status
				inter.Neg[hash]=nil;									-- done this one
			end
		end

		if (not inter.HaltOutput) then									-- output not halted
			if (inter.Lines[1]) then									-- stuff waiting for transmission
				if (inter.Channel~="WHISPER" and not c2.Loopback[inter.Channel]) then
--print("LOOPBACK: Loopback added to",inter.Channel);
					table.insert(inter.Lines,1,{d="LOOPBACK"});
				end
--print("inter.Lines[1]");
				if (inter.Lines[1].t) then inter.Receiver=inter.Lines[1].t; end	-- there'sa toon for this line
				if (inter.Receiver) then DM.SentBlindWhisperListOfReceivers[inter.Receiver]=now;	-- got one
				elseif (inter.Channel=="WHISPER") then					-- no active receiver, so get one
					inter.HaltOutput=true;								-- halt output until receiver has been found
					DM.MT:Run(tostring(inter).." -> Find pipe",c2.MT.FindPipe,c2.MT,inter);	-- find a piper
					return;
				end
				local temp,_,extra=c2:GetNextOutput(inter.Lines);				-- don't need "toon" at this point as it's gotten earlier if any
--		{ Neg={ na=name, md=meta, High=ver, Roll=0, MyVer=ver, MyRoll=0, MyRequest=true, },	-- extra meta for the engine
				if (extra and extra.Neg) then
--	inter.Neg[hash]={ na=name, md=meta, High=ver, Roll=0, MyVer=ver, MyRoll=0, MyRequest=true, };
--	inter.Neg[hash].End=time()+2;							-- I spoke up, so restart 2 seconds silence
					local _hash=temp:match("^(%d+)_"); _hash=tonumber(_hash);
					inter.Neg[_hash]=extra.Neg;
					inter.Neg[_hash].End=time()+2;
				end
--print("Out:",inter.ID,inter.Receiver,inter.Channel);
--print(temp);
				SendAddonMessage(inter.ID,temp,inter.Channel,inter.Receiver);	-- send it
--				c2.Sent[temp]=true;										-- mark it
			else inter.Receiver=nil; end								-- out-buffer empty, so also remove receiver-toon
		elseif (type(inter.HaltOutput)=="number") then					-- output is halted and timed
--print("inter.HaltOutput");
			if (now>=inter.HaltOutput) then								-- timer done
				wipe(inter.Lines);										-- remove output
				inter.HaltOutput=nil;									-- allow output
			end
		end
	end

	-- Maintain offline user supressor list
	if (next(DM.SentBlindWhisperListOfReceivers)) then
		local toon,val=next(DM.SentBlindWhisperListOfReceivers);
		if (now-val>5) then DM.SentBlindWhisperListOfReceivers[toon]=nil; end
	end

	-- handle incoming buffer
	if (c2.Incoming[1]) then
		local buf=table.remove(c2.Incoming,1);
		c2:ChannelInput(buf.pr,buf.me,buf.se,buf.ch);	-- prefix,message,sender,channel
	end

	-- Run global pings
	if (not c2.RunningPingAll) then
		local _,inter=next(c2.PingList);	-- pick one
		if (inter and inter.ReceiverList and next(inter.ReceiverList)) then
			DM.MT:Run(tostring(inter).." -> Ping Everyone",c2.MT.PingAll,c2.MT,inter);
		end
	end
end

-- hash_line:lines:data
-- Remove any not complete transmission in the queue
-- Used when a piper logs off during transmission
function c2:UnwindOutput(inter)
	while (inter.Lines[1] and not inter.Lines[1]:match("^%d-_%d-:%d-:")) do
		table.remove(inter.Lines,1);
	end
end

function c2.API.Close(inter)
	for i,j in pairs(c2.Interface) do
		if (j==inter) then table.remove(c2.Interface,i); return true; end
	end
	return nil;
end

function c2.API.IsIdle(inter)
	if (inter.Lines[1]) then return nil; end
	if (next(inter.Neg)) then return nil; end
	if (next(inter.Incoming)) then return nil; end
	return true;
end

function c2.API.SetReceiver(inter,rec,name)
	if (type(rec)=="string") then rec={ [rec]=0, };		-- single toon, so convert to valid list
	elseif (type(rec)~="table") then return; end		-- i have no idea, so abort
--	rec=DM:CopyTable(rec);								-- get a copy in case of direct storage reference
	for dude in pairs(rec) do rec[dude]=0; end			-- set all unused (zero uses)
	inter.ReceiverList=rec;								-- set list in interface
	inter.ReceiverListName=name;						-- identify list
	if (inter.ListOfToonsExhausted) then				-- the list has been exhausted for some reason
		inter.ListOfToonsExhausted=nil;
		if (type(inter.HaltOutput)=="number") then		-- sending has been inhibited and timed
			inter.HaltOutput=time();					--=====> release output and wipe buffer now since we have a new list
		end
	end
	table.insert(c2.PingList,inter);					-- add this interface to ping-list
end

function c2.API.SetChannel(inter,ch,rec)
	inter.Channel=ch:upper();
	if (inter.Channel=="WHISPER") then inter.Receiver=rec; else inter.Receiver=nil; end
end

function c2.API.SetID(inter,id)
	inter.ID=id;
	RegisterAddonMessagePrefix(id);
end

function c2.API.SetBar(inter,bar)
	inter.PB=bar;
end

-- inter: interface
-- data : any data
-- name : string
-- meta : any data
function c2.API.Send(inter,data,name,meta)
	c2:AddToSendQueue(inter,{ api="Send", md=meta, na=name, da=data });
--	c2:SendBurst(inter,{ api="Send", md=meta, na=name, da=data },toon);		-- for burst testing
end

function c2.API.Ping(inter,toon)
	c2:SendBurst(inter,{ api="Ping", },toon);
end

function c2.API.Request(inter,name,meta,ver)
	if (not ver) then ver=inter:cbVC(name,meta) or 0; end
--print("ver",ver);
--	local hash=c2:SendBurst(inter,{ api="Request", md=meta, na=name, ve=ver });
--	inter.Neg[hash]={ na=name, md=meta, High=ver, Roll=0, MyVer=ver, MyRoll=0, MyRequest=true, };
--	inter.Neg[hash].End=time()+2;							-- I spoke up, so restart 2 seconds silence
	local hash=c2:AddToSendQueue(	inter,											-- interface
									{ api="Request", md=meta, na=name, ve=ver },	-- data
									nil,											-- queue, if not default
									nil,											-- toon, if applicable
									nil,											-- hash, if not default
									{ Neg={ na=name, md=meta, High=ver, Roll=0, MyVer=ver, MyRoll=0, MyRequest=true, }, }	-- extra meta for the engine
									);
	return true;
end

function c2.API.ShareEvent(inter,setself,widget,...)
	local count=select("#",...);
	local buf={};
	for i=1,count do
		local arg=select(i,...);				-- select has multiple return
		buf[i]=arg;
	end
	inter:Send(buf,"ShareEvent",{widget=widget,setself=setself});
end

function c2.API.ProgressIB(inter)
	if (not c2.Build[inter.ID]) then return 1; end
	local total,lines=0,0;
	local tmp;
	for h,_h in pairs(c2.Build[inter.ID]) do			-- hash
		for s,_s in pairs(_h) do						-- sender
			total=total+(_s.Lines or 0);				-- add number of lines expected
			tmp=0; for i in ipairs(_s) do tmp=i; end	-- find last line received
			lines=lines+tmp;							-- add number of lines received
		end
	end
	if (total==0) then return -1; end					-- nothing expected
	if (lines==0) then return 0; end					-- nothing received
	return total/lines;									-- get current
end

function c2:Roll(inter,hash,receiver)
	if (not inter.cbVC) then								-- no callback, can't participate
		inter.Neg[hash].MyVer=0;
		inter.Neg[hash].MyRoll=0;
		return;
	end
	inter.Neg[hash].MyVer=inter:cbVC(inter.Neg[hash].na,inter.Neg[hash].md) or 0;	-- set my version
	inter.Neg[hash].MyRoll=math.random(2147483647);			-- 31-bit random
	if (inter.Neg[hash].MyVer<inter.Neg[hash].High) then return; end	-- too low, don't announce it
	if (inter.Neg[hash].MyVer==inter.Neg[hash].High) then				-- same version
		if (inter.Neg[hash].MyRoll<=inter.Neg[hash].Roll) then return; end	-- too low roll
	end
	inter.Neg[hash].High=inter.Neg[hash].MyVer;
	inter.Neg[hash].Roll=inter.Neg[hash].MyRoll;
--print("Com: Sending roll at hash",hash);
	c2:SendBurst(inter,{ api="Negotiate", ve=inter.Neg[hash].High, ro=inter.Neg[hash].Roll, md=inter.Neg[hash].md, na=inter.Neg[hash].na, },receiver,hash);
end

function c2:UpdateNegotiation(inter,hash,in_ver,in_roll,in_name,in_meta)
--print("Com: UpdateNegotiation at hash",hash);
	if (not inter.Neg[hash] or not next(inter.Neg[hash])) then return nil; end
--print(" = = > Got inter.Neg at",hash);
	if (inter.Neg[hash].na~=in_name or not DM:CompareTables(inter.Neg[hash].md,in_meta)) then return nil; end
--print(" = = > Got na amd md at",hash);
	if (not inter.Neg[hash].High or not inter.Neg[hash].Roll) then return nil; end
--print(" = = > Got High and Roll at",hash);
	if (inter.Neg[hash].High>in_ver) then return nil; end		-- new is lower
--print(" = = > Got low High at",hash);
	if (inter.Neg[hash].High==in_ver) then						-- new is equal
		if (inter.Neg[hash].Roll>in_roll) then return nil; end	-- new is lower
	end
--print(" = = > Can update neg data at",hash);
	inter.Neg[hash].High=in_ver;			-- set highest data
	inter.Neg[hash].Roll=in_roll;			-- set highest data
	return true;							-- updated
end

function c2:ChannelInput(prefix,text,sender,channel)
--print(text);
	local hash,line,data=text:match("^(%d-)_(%d-):(.*)$");				-- split basic format
--print(hash,line,data);
	if (not hash or not line or not data) then return; end

	-- handle input
	local buf=c2.Build;
	local now=time(); hash=tonumber(hash); line=tonumber(line);			-- get ready
	if (line==1) then													-- first line, so init reception
		buf[prefix]=buf[prefix] or {}; buf=buf[prefix];					-- init new prefix
		buf[hash]=buf[hash] or {}; buf=buf[hash];						-- init new hash
		buf[sender]=buf[sender] or {}; buf=buf[sender];					-- init sender for this hash
		if (buf.Lines) then print("COM_V2 ERROR: hash/sender pair reused while in use.",hash,sender); end
		wipe(buf);														-- wipe it
		buf.Lines,buf[1]=data:match("^(%d-):(.*)$");					-- grab line-count and data
		buf.Lines=tonumber(buf.Lines);									-- number
	else
		if (not buf[prefix] or not buf[prefix][hash] or not buf[prefix][hash][sender]) then return; end	-- don't have it, ignore it
		buf=buf[prefix][hash][sender];
		buf[line]=data;													-- insert it
	end
	buf.LastTime=now;													-- mark last valid reception
	if (line~=buf.Lines) then return; end								-- not done, so abort for now
	-- rebuild input
	-- (table.concat would be faster, but this way detects missing data)
	local build="";
	line=0; repeat
		line=line+1;
		if (not buf[line]) then return; end								-- missing data
		build=build..buf[line];											-- add it
	until (line>=buf.Lines);

	local bcr=DM:CopyTable(buf);
	local bc=build;

	wipe(c2.Build[prefix][hash][sender]); c2.Build[prefix][hash][sender]=nil;	-- clear this sender
	if (not next(c2.Build[prefix][hash])) then c2.Build[prefix][hash]=nil; end	-- nothing else waiting, so clear this hash
	if (not next(c2.Build[prefix])) then c2.Build[prefix]=nil; end				-- nothing else waiting, so clear this prefix
--	build=c2:DecodeTable(DM.Code:Decode(build));
	local decoded;
	build,decoded=DM.Code:Decode(build);								-- rebuild binary data from transmission coding
	if (not decoded) then
		MessageBoard_LS.BrokenCom_raw=bcr;
		MessageBoard_LS.BrokenCom=bc;
		print("Broken input data. Length:",build:len());
		return;
	end	-- could not decode data
	build=c2:DecodeTable(build);										-- rebuild lua data

--print("Decoding:");
--print(text);
--print(hash,line);

	-- handle input for interfaces
	for _,inter in pairs(self.Interface) do repeat						-- do all interfaces
		if (prefix~=inter.ID) then break; end							-- has wrong prefix, do next
		-- check previous massive fail - correct it if sender is present
		if (inter.ListOfToonsExhausted) then							-- our list is exhaused
			if (type(inter.ReceiverList)=="table" and inter.ReceiverList[sender]) then		-- got this sender
				if (inter.ReceiverList[sender]<0) then inter.ReceiverList[sender]=0; end	-- marked unresponsive, so set unknown
				inter.ListOfToonsExhausted=nil;							-- enable further attempts when needed
				if (type(inter.HaltOutput)=="number") then				-- output is halted with timer
					inter.HaltOutput=time();							--=====> release output and wipe buffer now
				end
			end
		end
--dbg(prefix.." incoming:",build);
--for _i,_d in pairs(build) do print(_i,_d); end
		if (build.api=="Negotiate") then
--print("Com: Negotiate");
			local neg=inter.Neg[hash];
			if (not neg or not next(neg)) then break; end									-- no running request/negotiation for this hash, so abort
			if (neg.na~=build.na or not DM:CompareTables(neg.md,build.md)) then break; end	-- not equal names or metadata
			self:UpdateNegotiation(inter,hash,build.ve,build.ro,build.na,build.md);			-- update negotiation with incoming stuff
			neg.End=time()+2;																-- someone spoke up, so restart 2 seconds silence
		elseif (build.api=="Request") then
--print("Com: Request");
			inter.Neg[hash]={ na=build.na, md=build.md, High=build.ve, Roll=0 };	-- create new negotiation
			self:Roll(inter,hash,sender);											-- roll for it and burst negotiation
			inter.Neg[hash].End=time()+2;											-- someone spoke up, so restart 2 seconds silence
		elseif (build.api=="Send") then															-- someone sent something
--dbg("Received at Send",build);
			if (type(build.da)=="table" and build.na=="ShareEvent" and type(build.md)=="table") then	-- A shared event has been received
				c2:InsertEvent(build.md,build.da); break;										-- dispatch event and done this interface
			end
			if (inter.cbRX) then inter:cbRX(build.da,build.na,build.md,sender,build.api); end	-- call addon RX callback
			-- optionally notify caller
			if (inter.EventHandler) then														-- there's an eventhandler widget
				local handler=inter.EventHandler:GetScript("OnEvent");							-- find function
				if (handler) then handler(inter.EventHandler,"COM2_DATA_RECEIVED",inter,build.da,build.na,build.md,sender,build.api); end
			end
		elseif (build.api=="Ping") then													-- it's a ping (always burst)
			c2:SendBurst(inter,{ api="Pinged", },sender);								-- reply burst blindly
		elseif (build.api=="Pinged") then												-- it's a ping-reply (always burst)
			if (type(inter.ReceiverList)=="table" and inter.ReceiverList[sender]) then	-- verify that a receiver list exists for the interface
				inter.ReceiverList[sender]=inter.ReceiverList[sender]+1;				-- mark sender as "responding"
			end
		end
	break; until (true); end		-- fake loop
end

function c2:InsertEvent(md,buf)
	local widget=_G[md.widget]; if (not widget or not widget.GetScript) then return; end
	local handler=widget:GetScript("OnEvent"); if (not handler) then return; end
	if (md.setself) then handler(widget,unpack(buf));
	else handler(unpack(buf)); end
end

-- hash_line:lines:data
function c2:SendBurst(inter,t,toon,hash)
	toon=toon or inter.Receiver;
	if (toon) then DM.SentBlindWhisperListOfReceivers[toon]=time(); end
	toon=toon or "";
	hash=c2:AddToSendQueue(inter,t,inter.Burst,toon,hash);		-- fill the burst buffer
	DM.MT:Run(tostring(inter).." -> Send burst",c2.MT.SendBurst);
	return hash;
end

function c2.MT:SendBurst()
	for _,inter in pairs(c2.Interface) do								-- walk all interfaces
		while (inter.Burst[1]) do										-- is there burst data for this interface
			local temp,toon=c2:GetNextOutput(inter.Burst); if (not temp) then return; end
--			c2.Sent[temp]=true;											-- set data as sent
			SendAddonMessage(inter.ID,temp,inter.Channel,toon);			-- send it
--			repeat DM.MT:Yield(true); until (not c2.Sent[temp]);		-- wait for feedback
			DM.MT:Yield(true);
		end
	end
end

function c2:GetNextOutput(buf)
	if (not buf[1]) then return nil,nil,nil; end
	local temp=table.remove(buf,1);
	return temp.d,temp.t,temp.extra;
end

function c2:AddToSendQueue(inter,t,queue,toon,hash,extra)
--dbg("function c2:AddToSendQueue(inter,t,queue,toon,hash)",t.da);
	if (not queue) then queue=inter.Lines; end
	local out="0000000000:"..DM.Code:Code(c2:CodeTable(t));	-- stringify, UU-encode, add line-count
	inter.Counter=inter.Counter+1;
	hash=hash or (inter.Hash+inter.Counter);
	local pre=tostring(hash).."_";
	local head=tostring(inter.ID..pre):len()+2;		-- only length of line numbers missing
	local first=(#queue)+1;
	local pos,linn,lins=1,0,"0";
	repeat
		linn=linn+1; lins=tostring(linn);
		table.insert(queue,{ t=toon, d=pre..lins..":"..out:sub(pos,pos+(254-(head+lins:len()))), } );
		pos=pos+(254-(head+lins:len()))+1;
	until (pos>out:len());
local add=0;
for _i=first,#queue do add=add+queue[_i].d:match("^.-:(.*)$"):len(); end
if (add~=out:len()) then
	print("ERROR IN ENCODING");
	print("add:",add," - out:len:",out:len());
end
	queue[first].d=queue[first].d:gsub(":0000000000:",":"..lins..":",1);
	queue[linn].extra=extra;
--print(inter.ID,hash,first,linn,lins);

	return hash;		-- last hash
end

-- "match" will fail due to miscount in this if byte-coding for the saved file has been changed.
-- speedy version
function c2:DecodeTable(s)
	local _match,_tonumber,_settype,_sub,_byte=string.match,tonumber,DM.Table.v03.SetType,string.sub,string.byte;	-- make everything local
	local nTY,nLE,name,dTY,dLE;
	local pos,maxlen,t=1,s:len(),{};
	repeat
		nTY=_byte(s,pos);													-- get type with numeric representation to not create a new string
		nLE,pos=_match(s,".(%x-):().*",pos); nLE=_tonumber(nLE,16);			-- capture data-length,position - convert length to number
		name=_settype(nTY,_sub(s,pos,pos+(nLE-1))); pos=pos+nLE;			-- grab it and advance to next
		dTY=_byte(s,pos);													-- get type with numeric representation to not create a new string
		dLE,pos=_match(s,".(%x-):().*",pos); dLE=_tonumber(dLE,16);			-- empty capture reads character position
		t[name]=_settype(dTY,_sub(s,pos,pos+(dLE-1))); pos=pos+dLE;			-- grab, shove, advance
	until (pos>maxlen);
	return t;
end

-- third letter of native type string - lower case, hex representation for avoiding string
function c2:SetType(ty,da)
	if (ty==0x72) then return tostring(da); end		-- st-r-ing
	if (ty==0x6D) then return tonumber(da); end		-- nu-m-ber
	if (ty==0x6F) then return (da=="true"); end		-- bo-o-l
	if (ty==0x62) then if (da=="") then return {}; end return DM.Table.v03.DecodeTable(da); end	-- ta-b-le
	if (ty==0x6C) then return nil; end				-- ni-l-
	if (ty==0x6E) then return loadstring(da); end	-- fu-n-ction
	return da;
end

-- this function:
--   accesses all global functions locally for speed
--   uses table concat for extreme speed in repetetive storage, like cloth with 700+ creatures codes in well under 5ms
--   In case of further timing problems: further work can be to store all strings separately to remove all string concatenations
function c2:CodeTable(t)
	local _string_format,_codetable,_tostring,_type,_len,_sub,_dump=string.format,c2.CodeTable,tostring,type,string.len,string.sub,string.dump;	-- make everything local
	local list={};
	local tmp,add;
	for en,ed in pairs(t) do
		tmp=_tostring(en);
--		if (_type(ed)=="table") then add=_codetable(c2,ed); else add=_tostring(ed); end
		if (_type(ed)=="table") then add=_codetable(c2,ed); elseif (_type(ed)=="function") then add=_dump(ed); else add=_tostring(ed); end
		list[#list + 1]=_sub(_type(en),3,3).._string_format("%X",_len(tmp))..":"..tmp.._sub(_type(ed),3,3).._string_format("%X",_len(add))..":"..add;
	end
	return table.concat(list);
end

function c2:DecodeTable_old(s)
	local t={};
	local nTY,nLE,name,dTY,dLE,data;
	local pos=1;
	repeat
		nTY,nLE=s:match("(..)(%d-):.*",pos); nLE=tonumber(nLE);
		pos=pos+2+tostring(nLE):len()+1;
		name=self:SetType(nTY,s:sub(pos,pos+(nLE-1)));
		pos=pos+nLE;
		dTY,dLE=s:match("(..)(%d-):.*",pos); dLE=tonumber(dLE);
		pos=pos+2+tostring(dLE):len()+1;
		data=self:SetType(dTY,s:sub(pos,pos+(dLE-1)));
		t[name]=data;
		pos=pos+dLE;
	until (pos>s:len());
	return t;
end

function c2:SetType_old(ty,da)
	if (ty=="st") then return tostring(da); end
	if (ty=="nu") then return tonumber(da); end
	if (ty=="bo") then return da=="true"; end
	if (ty=="ta") then if (da=="") then return {}; end return self:DecodeTable(da); end
	if (ty=="ni") then return nil; end
	return da;
end

function c2:CodeTable_old(t)
	local line="";
	for en,ed in pairs(t) do
		local tmp=tostring(en);
		line=line..type(en):sub(1,2)..tmp:len()..":"..tmp;
		local add="";
		if (type(ed)=="table") then add=self:CodeTable(ed);
		else add=tostring(ed); end
		line=line..type(ed):sub(1,2)..add:len()..":"..add;
	end
	return line;
end

function c2.MT:PingAll(inter)
	c2.RunningPingAll=true;
print("Ping all");
	for toon in pairs(inter.ReceiverList) do
print("Ping out:",toon);
		local now=debugprofilestop();
		inter.ReceiverList[toon]=-1;				-- set not present
		inter:Ping(toon);
		repeat DM.MT:Yield(true); until (debugprofilestop()-now>300);
	end
	c2.RunningPingAll=nil;
end

function c2.MT:FindPipe(inter,didfullreset)
	if (not inter.ReceiverList[inter.Receiver]) then inter.Receiver=nil; end
	local toon,stat=inter.Receiver,nil;
	repeat
		toon,stat=next(inter.ReceiverList,toon);		-- get next in line
		if (toon and stat>=0) then						-- valid toon
			local now=debugprofilestop();				-- start window
			inter.ReceiverList[toon]=-1;				-- set not present
			inter:Ping(toon);							-- ping it (burst)
			repeat DM.MT:Yield(true); until (debugprofilestop()-now>2000 or inter.ReceiverList[toon]>=0);
			if (inter.ReceiverList[toon]>=0) then		-- got a valid answer
				inter.Receiver=toon;					-- found one
				inter.HaltOutput=nil;					--=====> release output now
				return;									-- continue with this receiver activated
			end
		end
	until (toon==inter.Receiver);						-- cycled once
	-- no toons found
	if (not didfullreset) then							-- not a cycle after reset
		for toon,stat in pairs(inter.ReceiverList) do inter.ReceiverList[toon]=0; end	-- reset all toons
		return self:FindPipe(inter,true);				-- try again with closure
	end
	inter.Receiver=nil;									-- no receiver found
	inter.ListOfToonsExhausted=true;					-- so list has been exhausted
	inter.HaltOutput=time()+(15*60);					--=====> release output in 15 minutes
	-- optionally notify caller
	if (inter.EventHandler) then
		local handler=inter.EventHandler:GetScript("OnEvent");
		if (handler) then
			handler(inter.EventHandler,"COM2_RECIPIENTS_EXHAUSTED",inter,inter.ReceiverListName);
		end
	end
end

function DM.DefaultChatHook_AddMessage(...)
	local _,msg=...;        -- self,msg,r,g,b
	if (not msg or not msg.match) then return; end
	local checker=ERR_CHAT_PLAYER_NOT_FOUND_S:format("(.-)");
	checker=msg:match(checker);
	if (checker and DM.SentBlindWhisperListOfReceivers[checker]) then	-- user offline
		DM.SentBlindWhisperListOfReceivers[checker]=nil;
		for _,inter in pairs(c2.Interface) do				-- do all interfaces
			if (inter.ReceiverList[checker]) then
				inter.ReceiverList[checker]=-1;				-- proven not here
			end
			if (inter.Receiver==checker) then
				inter.Receiver=nil;							-- don't use it anymore
				c2:UnwindOutput(inter);						-- throw away any remaining data
			end
		end
		return;
	end
	DM.DefaultChatHook_AddMessage_Original(...);
end



-- prefix: standard addon-channel prefix
-- [tab] : tabulator inserted by the addon channel
-- hash+ : hash plus counter
-- _     : underscore
-- line  : line number for current block
-- :     : data separator

-- first line data -=> Lines:0000000000
-- line beginning -=> prefix[tab]hash+_line:
-- data coding -=> TY[len]:nameTY[len]:data

--[[
	anon = {
		md = <meta-data if any>,
		na = <data name if any>,
		ty = <data type>,
		da = <the data>,
	}
]]
