--[[
	DuckMod - A World of Warcraft add-on library
	Copyright (C) 2009-2012  Dag Bakken

	DuckMod is free software: you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.

	DuckMod is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with DuckMod.  If not, see <http://www.gnu.org/licenses/>.
]]


local TV=2.9991;

if (not DuckMod) then DuckMod = {}; end				-- Make it if empty
if (not DuckMod[TV]) then DuckMod[TV]={}; end		-- Don't mess with it if it's loaded
local DM=DuckMod[TV];
local _;


-- Make a random number of integer size with fractional part
function DM:Random(bottom,top)
	if (not bottom) then
		bottom=0;
		top=1;
	elseif (not top) then
		top=bottom;
		bottom=0;
	end

	local num=math.random();		-- Generate a fractional number
	num=num*(top-bottom);			-- Set span
	num=num+bottom;					-- Set floor
	return num;
end

-- Table size
function DM:Length(t)
	if (not t) then return 0; end
	local count=0;
	if (d_type and d_type(t)=="database") then
		for _,v in pairs(t._str_) do if (v:match("^%d+:(%a%a):[%+%-].*$")~="ni") then count=count+1; end end
	else
		for _ in pairs(t) do count=count+1; end
	end
	return count;
end

-- Safe table-copy with optional merge (equal entries will be overwritten, arg1 has pri)
function DM:CopyTable(t,new)
	if (not t and not new) then return nil; end
	if (not t) then t={}; end
	if (not new) then new={}; end
	local i,v;
	for i,v in pairs(t) do
		if (type(v)=="table") then new[i]=self:CopyTable(v,new[i]); else new[i]=v; end
	end
	return new;
end

function DM:CompareTables(one,two)
	if (type(one)~="table") then one={one}; end
	if (type(two)~="table") then two={two}; end
	for _e in pairs(one) do if (one[_e]~=two[_e]) then return nil; end end
	for _e in pairs(two) do if (one[_e]~=two[_e]) then return nil; end end
	return  true;
end

-- create string hash
function DM:Hash(text)
	local hash,i,len=0,1,text:len();
	repeat hash=hash+text:byte(i); i=i+1; until (i>len);
	return hash;
end

-- Return the common DuckMod ID
function DM:GetID(link)
	if (type(link)~="string") then
		if (type(link)~="number") then return nil,nil; end
		_,link=GetItemInfo(link);
	end
	local itemID,i1,i2,i3,i4,i5,i6=link:match("|Hitem:(%p?%d+):(%p?%d+):(%p?%d+):(%p?%d+):(%p?%d+):(%p?%d+):(%p?%d+):(%p?%d+)|h%[(.-)%]|h");
	if (not i6) then
		itemID,i1,i2,i3,i4,i5,i6=link:match("item:(%p?%d+):(%p?%d+):(%p?%d+):(%p?%d+):(%p?%d+):(%p?%d+):(%p?%d+)");
	end
	if (not i6) then return; end
	return "item:"..itemID..":"..i1..":"..i2..":"..i3..":"..i4..":"..i5..":"..i6,tonumber(itemID);
end

function DM:StoreByText(loc,...)
	loc=strtrim(loc);
	local f=loc:match("(.+)%(.*%)");
	if (f) then
		f=strtrim(f);
		local fu={strsplit(":",f)};
		local p;
		local fun={strsplit(".",fu[1])};
		for i,iD in ipairs(fun) do if (i==1) then p=_G[iD]; else p=p[iD]; end end
		pSelf=p;
		if (fu[2]) then
			fun={strsplit(".",fu[2])};
			for i,iD in ipairs(fun) do p=p[iD]; end
		else
			pSelf=nil;
		end
		if (pSelf) then p(pSelf,...);
		else p(...); end
	else
		if (loc:find(".")) then
			local fun={strsplit(".",loc)};
			for i,iD in ipairs(fun) do
				if (i==1) then p=_G[iD];			-- first table, so get its reference
				elseif (fun[i+1]) then p=p[iD];		-- there's more
				else p[fun[i]]=...; end				-- last one, so index previous table
			end
		else
			_G[loc]=...;		-- single global variable
		end
	end
end

-- Print text
function DM:Chat(msg,r,g,b)
	if (DEFAULT_CHAT_FRAME) then
		if (not r and not g and not b) then r=1; g=1; b=1; end;
		if (not r) then r=0; end;
		if (not g) then g=0; end;
		if (not b) then b=0; end;
		DEFAULT_CHAT_FRAME:AddMessage(msg,r,g,b);
	end
end


-- Multi-threading utility
DM.MT={
	LastStack="",
	Current=0,
	Count=0,
	Speed=(1/30)*1000,
	LastTime=0,
	Threads={
	},
};

function DM.MT:Run(name,func,...)
	for _,tTable in pairs(self.Threads) do
		if (tTable.orig==func and tTable.name==name) then
			if (dbg) then dbg.print(name,"MT> Already running in another thread. Aborted."); end
			return nil;		-- not started
		end
	end

	self.RunningCo=true;
	self.Count=self.Count+1;
--DM:Chat("Starting "..name);
	self.Threads[self.Count]={};
	self.Threads[self.Count].name=name;
	self.Threads[self.Count].orig=func;
	self.Threads[self.Count].cr=coroutine.create(func);
	self.LastTime=GetTime();
	self.LastStack="Running "..name;
	self.Speed=(1/30)*1000;
	-- For running without timer
	self.PassCounter=0;
	if (not self.Multiplier) then self.Multiplier=500; end

	local succeeded,result=coroutine.resume(self.Threads[self.Count].cr,...);
	if (not succeeded and Swatter) then
		if (Swatter) then Swatter.OnError(result,nil,self.LastStack);
		else DM:Chat(result); DM:Chat(self.LastStack); end
	end
	self.RunningCo=nil;
	return true;		-- started
end

-- Timer version
function DM.MT:Yield(immediate,dbdata)
	self.LastStack=debugstack(2);
	local now=debugprofilestop();
	if (not immediate and not InCombatLockdown()) then
		if (now-self.LastTime<self.Speed) then return; end
	end
	self.LastTime=now;		-- Inaccurate to account for other snags
	coroutine.yield();
	self.LastStack=debugstack(2);
end

function DM.MT:Next(elapsed)
	if (self.Count<1) then return; end
	if (self.RunningCo) then return; end	-- Don't if we are already doing it. In case of real MT.
	-- Set timing for MT yield
	if (elapsed<self.Speed) then self.Multiplier=self.Multiplier+5;
	else self.Multiplier=self.Multiplier-6; end
	if (self.Multiplier<1) then self.Multiplier=1; end
--DM:Chat(self.Multiplier);

	if (not self.Threads[self.Current+1]) then self.Current=0; end	-- Wrap
	self.Current=self.Current+1;
	if (not self.Threads[self.Current]) then return; end	-- Nothing to do
	if (coroutine.status(self.Threads[self.Current].cr)=="dead") then
		local removeIt=self.Current;
--DM:Chat("Ending "..self.Threads[removeIt].name);
		while (self.Threads[removeIt]) do
			self.Threads[removeIt]=self.Threads[removeIt+1];
			removeIt=removeIt+1;
		end
		self.Current=self.Current-1;
		self.Count=self.Count-1;
	else
		self.RunningCo=true;
		local succeeded,result=coroutine.resume(self.Threads[self.Current].cr);
		if (not succeeded) then
			if (Swatter) then Swatter.OnError(result,nil,self.LastStack);
			else DM:Chat(result); DM:Chat(self.LastStack); end
		end
 		self.RunningCo=nil;
	end
end

function DM.MT:Processes()
	return self.Count;
end


-- Table compression utility
-- API:
-- DuckMod.Table:Init(who,UseCache,Base)
--	Set default options for your addon.
--	   who	  - An identifier for your add-on.
--	   UseCache - Use cached mode
--	   Base	 - Set a default table for read/write
--
-- DuckMod.Table:Read(who,entry,base)
--	Read a compressed entry
--	   who   - Your addon's ID, if any
--	   entry - The entry you want to read
--	   base  - The base table where 'entry' is located. If omitted,
--			   the default setting for your addon will be used.
--
-- DuckMod.Table:Write(who,entry,tData,base)
--	 Write an uncompressed entry
--	   who   - Your addon's ID, if any
--	   entry - The entry you want to write
--	   tData - Data write to 'entry'
--	   base  - The base table where 'entry' is located. If omitted,
--			   the default setting for your addon will be used.

-- Table handling
DM.Table={
	v03={},
	LastPackerVersion="03",
	tableSlowCompress={},
	Scrap={},
	Cache={},
};
-- Backward compression compatibility
DM.Table[1]={ Entry ="\1", String="\1\1", Number="\1\2", Bool="\1\3", Nil="\1\4", Other ="\1\9", sTable="\2", eTable="\3", Version="\4", ThisVersion="1", Last="\4", };
DM.Table[2]={ Entry ="\2", String="\2\7", Number="\2\2", Bool="\2\3", Nil="\2\4", Other ="\2\9", sTable="\3", eTable="\4", Version="\7", ThisVersion="2", Last="\7", };
DM.Table.CV=2;
DM.Table.Entry=DM.Table[DM.Table.CV].Entry; DM.Table.String=DM.Table[DM.Table.CV].String; DM.Table.Number=DM.Table[DM.Table.CV].Number; DM.Table.Bool=DM.Table[DM.Table.CV].Bool;
DM.Table.Nil=DM.Table[DM.Table.CV].Nil; DM.Table.Other=DM.Table[DM.Table.CV].Other; DM.Table.sTable=DM.Table[DM.Table.CV].sTable; DM.Table.eTable=DM.Table[DM.Table.CV].eTable;
DM.Table.Version=DM.Table[DM.Table.CV].Version; DM.Table.ThisVersion=DM.Table[DM.Table.CV].ThisVersion; DM.Table.Last=DM.Table[DM.Table.CV].Last;


-- "match" will fail due to miscount in this if byte-coding for the saved file has been changed.
-- speedy version
function DM.Table.v03.DecodeTable(s)
	local _match,_tonumber,_settype,_sub,_byte=string.match,tonumber,DM.Table.v03.SetType,string.sub,string.byte;	-- make everything local
	local nTY,nLE,name,dTY,dLE;
	local pos,maxlen,t=1,s:len(),{};
	repeat
		nTY=_byte(s,pos);													-- get type with numeric representation to not create a new string
		nLE,pos=_match(s,".(%x-):().*",pos); nLE=_tonumber(nLE,16);			-- capture data-length,position - convert length to number
		name=_settype(nTY,_sub(s,pos,pos+(nLE-1))); pos=pos+nLE;			-- grab it and advance to next
		dTY=_byte(s,pos);													-- get type with numeric representation to not create a new string
		dLE,pos=_match(s,".(%x-):().*",pos); dLE=_tonumber(dLE,16);			-- empty capture reads character position
		t[name]=_settype(dTY,_sub(s,pos,pos+(dLE-1))); pos=pos+dLE;			-- grab, shove, advance
	until (pos>maxlen);
	return t;
end

-- third letter of native type string - lower case, hex representation for avoiding string
function DM.Table.v03.SetType(ty,da)
	if (ty==0x72) then return tostring(da); end		-- st-r-ing
	if (ty==0x6D) then return tonumber(da); end		-- nu-m-ber
	if (ty==0x6F) then return (da=="true"); end		-- bo-o-l
	if (ty==0x62) then if (da=="") then return {}; end return DM.Table.v03.DecodeTable(da); end	-- ta-b-le
	if (ty==0x6C) then return nil; end				-- ni-l-
	if (ty==0x6E) then return loadstring(da); end	-- fu-n-ction
	return da;
end

-- this function:
--   accesses all global functions locally for speed
--   uses table concat for extreme speed in repetetive storage, like cloth with 700+ creatures codes in well under 5ms
--   In case of further timing problems: further work can be to store all strings separately to remove all string concatenations
function DM.Table.v03.CodeTable(t)
	local _string_format,_codetable,_tostring,_type,_len,_sub,_dump=string.format,DM.Table.v03.CodeTable,tostring,type,string.len,string.sub,string.dump;	-- make everything local
	local list={};
	local tmp,add;
	for en,ed in pairs(t) do
		tmp=_tostring(en);
--		if (_type(ed)=="table") then add=_codetable(ed); else add=_tostring(ed); end
		if (_type(ed)=="table") then add=_codetable(ed); elseif (_type(ed)=="function") then add=_dump(ed); else add=_tostring(ed); end
		list[#list + 1]=_sub(_type(en),3,3).._string_format("%X",_len(tmp))..":"..tmp.._sub(_type(ed),3,3).._string_format("%X",_len(add))..":"..add;
	end
	return table.concat(list);
end


-- For non-default uses
function DM.Table:Init(who,UseCache,Base)
	if (who=="Default") then return; end
	if (not Base) then return nil; end
	if (not self.Default[who]) then self.Default[who]={}; end
	wipe(self.Default[who]);
	self.Default[who].UseCache = UseCache;
	self.Default[who].Base	 = Base;
	self.Default[who].Cache	= {};
end

function DM.Table:GetType(typedata,tdExtra,CV)
	if (not CV) then CV=self.CV; end
	if (type(typedata)=="string") then return self[CV].String; end
	if (type(typedata)=="number") then return self[CV].Number; end
	if (type(typedata)=="boolean") then return self[CV].Bool; end
	if (type(typedata)=="table") then return self[CV].sTable..string.char(tdExtra); end
	if (not typedata) then return self[CV].Nil; end
	return self[CV].Other;
end

function DM.Table:CompressV3(tData,s,e)
	local now=debugprofilestop();
	local tmp=DM.Table.v03.CodeTable(tData);
	table.insert(DM.Table.tableSlowCompress,{t=debugprofilestop()-now,s=s,e=e});
	table.sort(DM.Table.tableSlowCompress,function (a,b) return (a.t>b.t); end);
	table.remove(DM.Table.tableSlowCompress,6);	-- keep list of five
	return tmp;
end


--tostring(123)  --Returns "123"
--tostring({})  --Returns "table: ###"
--tostring(function() end) --Returns "function: ###"
--tostring(nil) --Returns "nil"
--tostring(true) --Returns "true"

--[[
	MyTable = {
		First = {
			fdata=5,
		}
		second = {
			sdata="text",
		}
	}

	base[MyTable]="
	'Version''ThisVersion'sTable1'	-> Write
	'String'First'sTable2'			-> Compress (1)
	'String'fdata'Number'			-> Compress (2)
	5								-> Compress (3)
	'eTable2'						-> Compress (2)
	'String'second'sTable2'			-> Compress (1)
	'String'sdata'String'			-> Compress (2)
	text							-> Compress (3)
	'eTable2'						-> Compress (2)
	'eTable1'						-> Compress (1)
]]


function DM.Table:Write(who,entry,tData,base,section)
	local cache=self.Default[who].UseCache;
	if (not who) then who="Default"; end
	if (not base) then
		base=self.Default[who].Base;
	elseif (section and base~=self.Default[who].Base[section]) then		-- Only cache for base data
		cache=false
	end
	base[entry]=self[self.CV].Version..self[self.CV].ThisVersion..self:GetType(tData,100,self.CV)..self:CompressV2(tData,101);
	base[entry]=base[entry]..self[self.CV].eTable..string.char(100);

	if (cache) then
		if (base==self.Default[who].Base) then
			if (section) then
				if (not self.Default[who].Cache[section]) then self.Default[who].Cache[section]={}; end
				cache=self.Default[who].Cache[section];
			else
				cache=self.Default[who].Cache;
			end
			cache[entry]=tData;
		end
	end
	return true;
end

--[[
	index
	  1   2 -> Entry-type
	  3   a -> Entry name
	a+1 a+2 -> Data type
	a+3   b -> Data
]]

function DM.Table:PullEntry(cString,CV)
	if (not CV) then CV=self.CV; end
	local stop1,stop2,stop3;
	if (cString:sub(1,1)==self[CV].sTable) then
		stop1=cString:find(self[CV].eTable..cString:sub(2,2),3,true);
		if (not stop1) then return nil; end	-- Missing data
		return cString:sub(1,2),cString:sub(3,stop1-1),cString:sub(stop1+2);
	end
	stop1=cString:find(self[CV].Entry,3,true); if (not stop1) then stop1=cString:len()+1; end	-- Virtual entry position
	stop2=cString:find(self[CV].sTable,3,true); if (not stop2) then stop2=cString:len()+1; end	-- Virtual entry position
	stop3=cString:find(self[CV].eTable,3,true); if (not stop3) then stop3=cString:len()+1; end	-- Virtual entry position
	if (stop1>stop2) then stop1=stop2; end
	if (stop1>stop3) then stop1=stop3; end
	return cString:sub(1,2),cString:sub(3,stop1-1),cString:sub(stop1);
end

function DM.Table:DecompressV1(base,cString,anon)
--DM:Chat("Decompressing...");
	while(cString:len()>0) do
		local eType,eData=nil,nil;
		local dType,dData=nil,nil;
		if (anon) then
			dType,dData,cString=self:PullEntry(cString,1);
			dData=self:DecompressV1(base,dData);	-- dData is now without table-tags
		else
			eType,eData,cString=self:PullEntry(cString,1);		-- Entry name
			if (eType==self[1].Number) then eData=tonumber(eData); end
			dType,dData,cString=self:PullEntry(cString,1);		-- Entry data
			if (dType==self[1].Number) then dData=tonumber(dData);
			elseif (dType==self[1].Bool) then if (dData=="true") then dData=true; else dData=false; end
			elseif (dType==self[1].Nil) then dData=nil;
			elseif (dType:sub(1,1)==self[1].sTable) then
				base[eData]={};
				dData=self:DecompressV1(base[eData],dData);	-- dData is now without table-tags
			end
		end
		if (not base) then return dData;
		else
			if (anon) then base=dData;
			else base[eData]=dData; end
		end
	end
	return base;
end

function DM.Table:DecompressV2(base,cString,anon)
--DM:Chat("Decompressing...");
	while(cString:len()>0) do
		local eType,eData=nil,nil;
		local dType,dData=nil,nil;
		if (anon) then
			dType,dData,cString=self:PullEntry(cString,2);
			dData=self:DecompressV2(base,dData);	-- dData is now without table-tags
		else
			eType,eData,cString=self:PullEntry(cString,2);		-- Entry name
			if (eType==self[2].Number) then eData=tonumber(eData); end
			dType,dData,cString=self:PullEntry(cString,2);		-- Entry data
			if (dType==self[2].Number) then dData=tonumber(dData);
			elseif (dType==self[2].Bool) then if (dData=="true") then dData=true; else dData=false; end
			elseif (dType==self[2].Nil) then dData=nil;
			elseif (dType:sub(1,1)==self[2].sTable) then
				base[eData]={};
				dData=self:DecompressV2(base[eData],dData);	-- dData is now without table-tags
			end
		end
		if (not base) then return dData;
		else
			if (anon) then base=dData;
			else base[eData]=dData; end
		end
	end
	return base;
end

function DM.Table:DecompressV3(ref,cString,anon)
	if (cString=="") then return {}; end
	ref=DM.Table.v03.DecodeTable(cString);			-- store at given reference
	return ref;
end


function DM.Table:Read(who,entry,base,section)
	local cache=self.Default[who].UseCache;
	if (not who) then who="Default"; end								-- Using default settings
	if (not base) then
		base=self.Default[who].Base;
	elseif (section and base~=self.Default[who].Base[section]) then		-- Only cache for base data
		cache=false
	end
	if (not base[entry]) then return nil; end							-- Entry does not exist
	if (type(base[entry])~="string") then			-- Not a proper entry
		DM:Chat("Got "..type(base[entry]).." as entry");
		return nil;
	end
--	if (base[entry]:byte(1)~=self.Version) then return base[entry]; end	-- It's uncompressed data
	local CV=tonumber(base[entry]:sub(2,2));
	local fnDecompress=self["DecompressV"..CV];		-- Look for decompressor
	if (not fnDecompress) then return nil; end		-- We don't have a decompressor for this data
	-- Not using cache, so decompress to scrap-book and return it
	if (not cache) then
		wipe(self.Scrap);
		return DM:CopyTable(fnDecompress(self,self.Scrap,base[entry]:sub(3),true));		-- Strip version
	end
	if (section) then
		if (not self.Default[who].Cache[section]) then self.Default[who].Cache[section]={}; end
		cache=self.Default[who].Cache[section];
	else
		cache=self.Default[who].Cache;
	end
	-- Not decompressed yet, so decompress to cache and return it
	if (not cache[entry]) then
--		cache[entry]=DM:CopyTable(fnDecompress(self,self.Scrap,base[entry]:sub(3),true));
		cache[entry]={};
		fnDecompress(self,cache[entry],base[entry]:sub(3),true);
	end
	return cache[entry];
end

function DM.Table:PurgeCache(who)
	wipe(self.Default[who].Cache);
	collectgarbage("collect");
end


DM.Code={};

--[[
	Code with password. It's not complicated, but needs the password
	to decode, and is totally unreadable after coding.
		code=mod(byte+seed+pw[next],256);
		byte=mod(code-pw[next]-seed,256);
]]
function DM.Code:Protect(text,pw)
	if (not pw or pw=="") then return text; end
	local seed=0;
	local pwlen=pw:len();
	for i=1,pwlen do seed=seed+pw:byte(i); end
	local code="";
	local pi=1;
	for i=1,text:len() do
		code=code..string.char(mod(text:byte(i)+(seed+pw:byte(pi)),256));
		pi=pi+1; if (pi>pwlen) then pi=1; end
	end
	return self:Code(code);		-- Results in pure 8-bit, so we must code it
end

function DM.Code:Unprotect(code,pw)
	if (not pw or pw=="") then return code; end
	local decoded;
	code,decoded=self:Decode(code);
	if (decoded~=true) then return code; end
	local seed=0;
	local pwlen=pw:len();
	for i=1,pwlen do seed=seed+pw:byte(i); end
	local text="";
	local pi=1;
	for i=1,code:len() do
		local value=mod(code:byte(i)-(seed+pw:byte(pi)),256);
		while(value<0) do value=value+256; end
		text=text..string.char(value);
		pi=pi+1; if (pi>pwlen) then pi=1; end
	end
	return text;
end

DM.Code.Base64scrap={};
DM.Code.Base64={"A","B","C","D","E","F","G","H","I","J","K","L","M","N","O","P","Q","R","S","T","U","V","W","X","Y","Z","a","b","c","d","e","f","g","h","i","j","k","l","m","n","o","p","q","r","s","t","u","v","w","x","y","z","0","1","2","3","4","5","6","7","8","9","+","/"};
DM.Code.Base64rev={};
for k,v in ipairs(DM.Code.Base64) do DM.Code.Base64rev[v]=k-1; end

function DM.Code:Code(text)
	local pad=text:len()%3; if (pad>0) then pad=3-pad; end
	text=text..string.rep(string.char(0),pad);
	local list=self.Base64scrap; wipe(list);
	local p1,p2,p3,p4;
	for i=1,text:len(),3 do
		p1=bit.lshift(text:byte(i),16)+bit.lshift(text:byte(i+1),8)+text:byte(i+2);
		p4=bit.band(p1,63); p1=bit.rshift(p1,6);
		p3=bit.band(p1,63); p1=bit.rshift(p1,6);
		p2=bit.band(p1,63); p1=bit.rshift(p1,6);
		table.insert(list,self.Base64[p1+1]);	-- lua offsets by 1
		table.insert(list,self.Base64[p2+1]);
		table.insert(list,self.Base64[p3+1]);
		table.insert(list,self.Base64[p4+1]);
	end
	if (pad>=1) then list[(#list)]="="; end
	if (pad>=2) then list[(#list)-1]="="; end
	return table.concat(list);
end

function DM.Code:Decode(code)
	if (code:len()%4~=0) then return code,false; end
	local pad=0; if (code:find("==",-2,true)) then pad=2; elseif (code:find("=",-1,true)) then pad=1; end
	local com;
	local text="";
	local p1,p2,p3,p4;
	for i=1,code:len(),4 do
		p1=self.Base64rev[code:sub(i  ,i  )];
		p2=self.Base64rev[code:sub(i+1,i+1)];
		p3=self.Base64rev[code:sub(i+2,i+2)] or 0;
		p4=self.Base64rev[code:sub(i+3,i+3)] or 0;
		com=bit.lshift(p1,18);
		com=com+bit.lshift(p2,12);
		com=com+bit.lshift(p3,6);
		com=com+p4;
		text=text..string.char(bit.rshift(com,16)%256)
		text=text..string.char(bit.rshift(com,8)%256)
		text=text..string.char(com%256)
	end
	if (pad>0) then
		text=text:sub(1,-1-pad);
	end
	return text,true;
end


--[[
	Text-code any three 8-bit values to four values ranging from 64-127
	inclusive.
		composite=byte1*65536+byte2*256+byte3
		char1=mod(composite,64)+64;
		composite=composite/64;
		char2=mod(composite,64)+64;
		composite=composite/64;
		char3=mod(composite,64)+64;
		char4=composite/64+64;

		composite=(char4-64)*262144+(char3-64)*4096+(char2-64)*64+(char1-64);
		char3=mod(composite,256);
		composite=composite/256;
		char2=mod(composite,256);
		char1=composite/256;
]]

--	raw string bytes:
--	11111111 22222222 33333333

--	Binary coding: +64 >> 7
--	111111110000000000000000
--	000000002222222200000000
--	000000000000000033333333

-->	111111112222222233333333 -> com[0]=0333333d
--	000000111111112222222233 -> com[1]=0222233c
--	000000000000111111112222 -> com[2]=0112222b
--	000000000000000000111111 -> com[3]=0111111a

function DM.Code:Code_old(text)
	text=text:len()..":"..text;
	while (math.floor(text:len()/3)*3~=text:len()) do text=text.." "; end
	local code="";
	for i=1,text:len(),3 do
		local com=bit.lshift(text:byte(i),16)+bit.lshift(text:byte(i+1),8)+text:byte(i+2);
		code=code..string.char(mod(com,64)+64); com=bit.rshift(com,6);
		code=code..string.char(mod(com,64)+64); com=bit.rshift(com,6);
		code=code..string.char(mod(com,64)+64); com=bit.rshift(com,6);
		code=code..string.char(com+64);
	end
	return code;
end

--	raw string bytes:
--	0333333d 0222233c 0112222b 0111111a

--	Binary coding:
--	111111000000000000000000
--	000000112222000000000000
--	000000000000222233000000
--	000000000000000000333333
-->	111111112222222233333333 -> inp=33333333
--	000000001111111122222222 -> inp=22222222 33333333
--	000000000000000011111111 -> inp=11111111 22222222 33333333

-- Return: text (string), decoded (true|false)
function DM.Code:Decode_old(code)
	local text="";
	if (mod(code:len(),4)~=0) then return code,false; end
	for i=1,code:len(),4 do
		local com;
		com=bit.lshift(code:byte(i+3)-64,18);
		com=com+bit.lshift(code:byte(i+2)-64,12);
		com=com+bit.lshift(code:byte(i+1)-64,6);
		com=com+(code:byte(i)-64);
		local inp=string.char(mod(com,256)); com=bit.rshift(com,8);
		inp=string.char(mod(com,256))..inp; com=bit.rshift(com,8);
		inp=string.char(mod(com,256))..inp;
--		inp=string.char(com)..inp;
		text=text..inp;
	end
	local here=text:find(":");
	if (not here) then return code,false; end
	local size=text:sub(1,here-1); size=tonumber(size);
	if (not size or size<1) then return code,false; end
	text=text:sub(here+1);
	size=text:len()-size;
	if (size>3 or size<0) then return "",true; end
	if (size>0) then text=text:sub(1,text:len()-size); end
	return text,true;
end


function packtable(t)
	if (not t) then t={}; end
	return "03"..DM.Table:CompressV3(t);
end
function unpacktable(s)
	if (type(s)=="table") then return s; end
	if (not s or s=="") then return {}; end
--print(type(s),s);
	local fnDecompress=DM.Table["DecompressV"..s:sub(2,2)];	-- Look for decompressor
	return fnDecompress(DM.Table,{},s:sub(3),true);
end

function DM:Tick(elapsed)
	DM.MT:Next(elapsed);		-- Run multi-threading
end

function DM:Init()
	if (not self.TheFrame) then
		self.TheFrame=CreateFrame("Frame","DuckMod-Core-Messager"..TV,UIParent);
		if (not self.TheFrame) then DM:Chat("Could not create DuckMod Core UI frame",1); return; end
	else
		return;		-- already inited
	end
	if (not self.TheFrameWorldMap) then
		self.TheFrameWorldMap=CreateFrame("Frame","DuckMod-CoreWM-Messager"..TV,WorldMapDetailFrame);
		if (not self.TheFrameWorldMap) then DM:Chat("Could not create DuckMod Core WM frame",1); return; end
	end
--	self.TheFrame:SetScript("OnEvent",self.OnEvent);
	self.TheFrame:SetScript("OnUpdate",self.Tick);
	self.TheFrameWorldMap:SetScript("OnUpdate",self.Tick);

	-- As LUA random can be dodgy at times, it is recommended to make a few calls to it to get it going.
	math.random(); math.random(); math.random();

	if (DM.Com2) then DM.Com2:Init(); end
	if (DM.Database) then DM.Database:Init(); end
	if (DM.WoWnet) then DM.WoWnet:Init(); end
	if (DM.Render) then DM.Render:Init(); end
	if (DM.BrowserServer) then DM.BrowserServer:Init(); end
end
