--[[
	DuckMod - A World of Warcraft add-on library
	Copyright (C) 2009-2014  Dag Bakken

	DuckMod is free software: you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.

	DuckMod is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with DuckMod.  If not, see <http://www.gnu.org/licenses/>.
]]


local TV=2.9991;

if (not DuckMod) then DuckMod = {}; end			-- Make it if empty
if (not DuckMod[TV]) then					-- Don't mess with it if it's loaded
	DuckMod[TV]={};
end

local DM=DuckMod[TV];
local _;

globalDMROOT=DM;


local DUCKNET_TREESEP="\8";				-- Tree separator
local hiddenKey="_DMDB-";
local membersDB=hiddenKey.."share-guildlistdatabases";
local guildsDB=hiddenKey.."share-guildlists";
local pf="dm"..tostring(TV);
local interface=nil;

DM.Database={
	Separator=DUCKNET_TREESEP,
	DataVersion=1,
	LastStamp=0,			-- Last out of the database
	LastRedundant=nil,		-- Last out of the database
	Redundant=nil,			-- Setting for written data
	BootDelay=15,
	StartBootDelay=15,
	PIPER_IDLE=0,			-- No attempt made
	PIPER_USED=1,			-- 1=used once, 2=twice, etc
	PIPER_UNAVAILABLE=-1,	-- Do not attempt this
	Com2={},
	API={},
	Util={ MT={}, },
	Info={ WhoList={}, },
	Addon={},
	Queue={},
	WQ={},
};
DM_PRUNE_NEVER=0;
DM_PRUNE_NORMAL=1;
DM_PRUNE_FIRST=10;			-- 11 second, 12 third, etc

DM.DatabaseQueue={};
DM.DatabaseData={};
local dbd=DM.DatabaseData;		-- database data
local dbc=DM.Database;			-- database class
local api=dbc.API;
local util=dbc.Util;
local wq=dbc.WQ;


local function dbg(h,t)
	print(h,":");
	for _i,_d in pairs(t) do print(_i,_d); end
end

local function say(...)
	print("|cFF00FFFF",...);
end

local function saydate(a,b)
	say("Current:",date("%c",a),"and external:",date("%c",b));
end

-- Set up for handling
function dbc:Init()
	if (not self.TheFrame) then
		self.TheFrame=CreateFrame("Frame","DuckMod-DB-Messager"..TV,UIParent);
		if (not self.TheFrame) then
			DM:Chat("Could not create DuckMod DB UI frame",1);
			return;
		end
		self.TheFrame:SetScript("OnEvent",dbc.OnEvent);
		self.TheFrame:SetScript("OnUpdate",dbc.Heartbeat);
	end
	if (not self.TheFrameWorldMap) then
		self.TheFrameWorldMap=CreateFrame("Frame","DuckMod-DBWM-Messager"..TV,WorldMapDetailFrame);
		if (not self.TheFrameWorldMap) then
			DM:Chat("Could not create DuckMod DB WM frame",1);
			return;
		end
--		self.TheFrameWorldMap:SetScript("OnEvent",dbc.OnEvent);
		self.TheFrameWorldMap:SetScript("OnUpdate",dbc.Heartbeat);
	end
end

-- An event has been received
function dbc:OnEvent(event,arg1,arg2,arg3,arg4,arg5,arg6,_,_,arg9)
	-- Database inter-guild local roster update
	if (event=="GUILD_ROSTER_UPDATE") then
--print("GUILD_ROSTER_UPDATE");
		if (arg1) then return; end		-- Some change, so don't all start sync together
		local list={};
		local num=GetNumGuildMembers();
		for i=1,num do
			local name=GetGuildRosterInfo(i);
			list[name]=dbc.PIPER_IDLE;
		end
		local guild=GetGuildInfo("player");
		if (not guild) then return; end
--print(num.." members in "..guild);
		dbc.Info:UpdateGuildMemberList(guild,list);
	end

	-- Database inter-guild external roster update
	if (event=="WHO_LIST_UPDATE") then
print("WHO_LIST_UPDATE");
		local list={};
		local num=GetNumWhoResults();
		local guild;
		for i=1,num do
			_,guild=GetWhoInfo(i);
			if (not list[guild]) then list[guild]=1;
			else list[guild]=list[guild]+1; end
		end
		local goal,amount="",0;
		for guild,count in pairs(list) do
			if (count>amount) then goal=guild; amount=count; end
		end
		guild=goal;
		wipe(list);
		for i=1,num do
			local name,gname=GetWhoInfo(i);
			if (gname==guild) then list[name]=dbc.PIPER_IDLE; end
		end
		-- this function MUST be called, so can not abort between WHO_LIST_UPDATE and UpdateGuildMemberList
		dbc.Info:UpdateGuildMemberList(guild,list,true);	-- Merge it with the other list, if any
	end

	if (event=="COM2_DATA_RECEIVED") then
-- arg1 inter
-- arg2 build.da
-- arg3 build.na
-- arg4 build.md
-- arg5 build.api
-- arg6 sender
	end
end


-- The data distribution code
DM.DatabaseMetaTable={
	__index = function (t,k)
		if (rawget(t,"_lnk_")) then
			if (not t:DM_Get("PauseLinks")) then return util:Handle(rawget(t,"_lnk_"))[k]; end
			return nil;
		end
		dbc.LastStamp=0;
		-- if there's no data-key, return nil
		if (not t._str_[k]) then return nil; end
		local typ,r,val;
		dbc.LastStamp,typ,r,val=t._str_[k]:match("^(%d+):(%a%a):([%+%-])(.*)$");
		dbc.LastStamp=tonumber(dbc.LastStamp);
		dbc.LastRedundant=(r=="-" and true or nil);
		if (typ=="nu") then return tonumber(val);
		elseif (typ=="bo") then return (val=="true");
		elseif (typ=="ni") then return nil;
		elseif (typ=="da") then return rawget(t,val);	-- return real handle for database
		elseif (typ=="ta") then return unpacktable(val); end
		return val;		-- string
	end,
	__newindex = function (t,k,v)
		local _rawget=rawget;
		if (_rawget(t,"_lnk_")) then
			if (not t:DM_Get("PauseLinks")) then util:Handle(_rawget(t,"_lnk_"))[k]=v; end
			return;
		end
		local typ,tim=d_type(v),util.LockedTime or time();
		local ty,ref;
		if (t._str_[k]) then										-- there's something there
			ty,ref=t._str_[k]:match("^%d+:(%a%a):[%+%-](.*)$");		-- get metainfo
			if (ty~="da") then ref=nil; end							-- no ref here
		elseif (not util.LockedTime) then tim=tim-1000000000; end	-- Fake very old for other sync later
		local register=nil;
		if (typ=="table") then										-- a table has been provided
			if (not _rawget(t,"_endpoint_")) then typ="database";	-- override for table-to-database storage
			else v=packtable(v); end							-- create packed table
		end
		if (typ=="database") then							-- doing ref-write
			if (not ref) then ref=t:DM_UniqueKey(); end		-- must have ref
			v=todatabase(v);								-- if table, go database
			rawset(t,ref,v);								-- shove it directly in there
			register=true;
			typ=d_type(v);									-- update after forging
			v=ref;											-- set data to database name for save
		elseif (ref) then									-- doing normal write, and have old ref
			rawset(t,ref,nil);								-- overwrite with nil like a table would have done
		end
		-- do redundancy
		local r="+";
		if (dbc.Redundant~=nil) then	-- accept true and false
			if (dbc.Redundant) then r="-"; else r="+"; end
		else
			if (t._str_[k]) then r=t._str_[k]:match("^%d+:%a%a:([%+%-]).*$"); end
		end
		-- save
		t._str_[k]=tim..":"..typ:sub(1,2)..":"..r..tostring(v);
		dbc.Redundant=nil;
		if (register) then util:RegisterSubDatabases(t); end
		-- Written, so queue for send
		if (not util.LockedTime and ((_rawget(t,"_tree_") or "")~="")) then
			util:QueueForSend(nil,nil,_rawget(t,"_tree_"),k);
		end
	end,
	__tostring = function (t)
		return "database: "..util:GetLuaName(t._tree_);
	end,
	__type = "database",

-- metamethod "len" cannot be redefined. DM:Length(t) will handle it.
--	__len = function (t)
--		local count=0;
--		for _,v in pairs(t._str_) do if (v:match("^%d+:(%a%a):[%+%-].*$")~="ni") then count=count+1; end end
--		return count;
--	end,

	-- function dispatcher
	__call = function(t,index)
		index=index:lower();
		if (index=="event_queue_peek") then
			local q=rawget(util:GetRoot(t),"_queue_");
			return q[1];
		end
	end,

	-- insert new event
	__add = function(t,v)
		local function different(one,two)								-- check if two 1-level tables are in any unequal
			for k in pairs(one) do if (one[k]~=two[k]) then return true; end end
			for k in pairs(two) do if (one[k]~=two[k]) then return true; end end
			return nil;
		end
		local q=rawget(util:GetRoot(t),"_queue_");						-- get queue for this database
		local found;
		for i,e in ipairs(q) do if (not different(e,v)) then found=i; break; end end	-- grab old equal entry 
		if (found) then table.remove(q,found); end						-- remove it to simulate it being moved to the end
		table.insert(q,v);												-- add event to queue
		if (#q>1) then													-- more than one entry in queue
			local start=(q[1].busy and 2) or 1; found=nil;				-- skip nr 1 if it has busy set
			for i=start,#q do if (q[i].tree=="databaselist") then if (q[i].method=="send") then found=i; end end end	-- find a databaselist send
			if (found) then table.insert(q,q[found]); table.remove(q,found); found=nil; end						-- move it to the end of the list
			for i=start,#q do if (q[i].tree=="databaselist") then if (q[i].method=="poll") then found=i; end end end	-- find a databaselist poll
			if (found) then table.insert(q,q[found]); table.remove(q,found); end								-- move it to the end of the list
		end
		return t;
	end,
	-- remove a given number of events
	__sub = function(t,v)
		local q=rawget(util:GetRoot(t),"_queue_");
		while(v>0) do table.remove(q,1); v=v-1; end
		return t;
	end,
}

function dbc:Register(prefix,db,ef)
	local efn; if (ef) then efn=ef:GetName(); end
	self:Connect(db,prefix,efn or ef);
end

function dbc:Connect(db,prefix,ef)
	setmetatable(db,nil);							-- kill handlers
	if (not self.TheFrame) then dbc:Init(); end
	if (not db._global_) then db._global_={}; end
	if (not prefix) then if (not db._global_._prefix_) then return nil; end prefix=db._global_._prefix_; else db._global_._prefix_=prefix; end
	if (not ef) then ef=db._global_._eventframe_; else db._global_._eventframe_=ef; end
	if (type(ef)=="string") then ef=_G[ef]; end
	if (db._queue_) then wipe(db._queue_); db._queue_=nil; end
	DM.DatabaseQueue[prefix]={};
	DM.Talk[prefix]=time();
	db._queue_=DM.DatabaseQueue[prefix];		-- Set queue handler
	db._tree_=type(prefix):sub(1,1)..prefix;	-- Set root
	if (not db._str_) then db._str_={}; end
	if (not db._prn_) then db._prn_=DM_PRUNE_NORMAL; end
	if (not db._global_.Limit) then db._global_.Limit=5000; end
	local inter=DM.Com2:CreateInterface(prefix,dbc.Com2.RX,dbc.Com2.VC,dbc.Com2.DR,self.TheFrame,"GUILD");
	dbc.Addon[prefix]={ EventFrame=ef, dbL={}, ml={}, iface={ inter }, };
	dbd[prefix]=db;										-- Set reference
	util:SetInterface(db);							-- Set handlers
	util:SetDefaults(db);							-- Set all default globals
	util:RegisterSubDatabases(db);					-- add all nested databases
	-- Add shared guilds
	local buf=db[guildsDB] or {};
	for sGuild,sForum in d_pairs(buf) do

		if (sForum) then db:DM_AddGuild(sGuild,sForum); end
	end
	-- run prune for the database
	DM.MT:Run("Pruning "..tostring(db),util.MT.Prune,util.MT,prefix);	-- Fake call with self
	-- blind stuff for silent channels
--	db=db+{ method="EMPTY", };		-- add empty event to queue
	-- Set up to sync this database
--	self.Util:QueueForSend(prefix,"databaselist","",true);	-- Local only
	util:QueueForSend(nil,prefix,"databaselist");			-- everyone
	DM.QueueInited=true;
	return true;
end

--
-- DM.Database:Register(prefix,myDB[,eventFrame][,channel])
-- myDB:DM_SetLink("link name",otherDB)
-- myDB:DM_Get(key)
-- myDB:DM_Set(key,value)
-- myDB:DM_GetRedundancy(key)
-- myDB:DM_SetRedundancy(key,redundant)
-- myDB:DM_PruneMode([mode])
-- myDB:DM_AddGuild(guild,forum,mode)
-- myDB:DM_RemoveGuild(guild)
-- myDB:DM_ListGuilds()
-- myDB:DM_UniqueKey()
--

function util:SetInterface(db)
	rawset(db,"DM_SetLink",api.SetLink);				-- ducklib.com
	rawset(db,"DM_Get",api.Get);						-- ducklib.com
	rawset(db,"DM_Set",api.Set);						-- ducklib.com
	rawset(db,"DM_SetRedundancy",api.SetRedundancy);	-- ducklib.com
	rawset(db,"DM_GetRedundancy",api.GetRedundancy);	-- ducklib.com
	rawset(db,"DM_PruneMode",api.PruneMode);			-- ducklib.com
	rawset(db,"DM_AddGuild",api.AddGuild);				-- ducklib.com
	rawset(db,"DM_RemoveGuild",api.RemoveGuild);		-- ducklib.com
	rawset(db,"DM_ListGuilds",api.ListGuilds);			-- ducklib.com
	rawset(db,"DM_UniqueKey",api.UniqueKey);			-- ducklib.com
	rawset(db,"DM_Registered",api.DatabaseRegistered);
	rawset(db,"DM_EndPoint",api.EndPoint);
	rawset(db,"DM_ProgressIB",api.ProgressIB);
	setmetatable(db,DM.DatabaseMetaTable);				-- Set handlers
end

-- Convert database to a normal table. This will of course remove any
-- information regarding prune, redundancy and time of storage.
function totable(t)
	local _d_type=d_type;
	if (_d_type(t)=="table") then return t; end				-- it's already a table
	if (_d_type(t)~="database") then return {t}; end			-- it's not a database, so enclose and return
	local buf={};											-- fresh table
	for k,v in d_pairs(t) do								-- walk all entries
		if (_d_type(v)=="database") then buf[k]=totable(v);	-- convert db to table
		else buf[k]=v; end									-- add entry as is
	end
	return buf;												-- return the new table
end

-- Convert table to a database. This will add default special entries, and
-- time of storage for all entries will be set to the current time.
function todatabase(t)
	local _d_type,_rawget=d_type,rawget;
	if (_d_type(t)=="database") then return t; end		-- it's already a database
	if (_d_type(t)~="table") then t={t}; end				-- it's not a table, so make one
	if (_rawget(t,"_tree_") and _rawget(t,"_str_") and _rawget(t,"_prn_")) then
		util:SetInterface(t);								-- set handlers
		return t;
	end
	local buf={
		_tree_="",									-- create dummy locator
		_str_={},									-- create storage
		_prn_=DM_PRUNE_NORMAL,						-- set normal pruning
	};
	util:SetInterface(buf);							-- set handlers
	for k,v in pairs(t) do							-- regular is fine here as it is a table
		if (_d_type(v)=="table") then buf[k]=v;		-- create new level database from nested table
		else buf[k]=v; end							-- add key with data to database
	end
	return buf;										-- done this database
end

function api.SetLink(db,dbName,linkDB)
	if (not linkDB) then rawset(db[dbName],"_lnk_",nil); return nil; end	-- clear link (if any)
	db[dbName]=todatabase({});								-- Make empty database
	rawset(db[dbName],"_lnk_",linkDB._tree_);				-- Set tree for external database
end

function api.Get(db,key)
	db=util:GetRoot(db);
	if (not rawget(db,"_global_")) then return nil; end		-- Not a root db
	return db._global_[key];
end

function api.Set(db,key,value)
	db=util:GetRoot(db);
	if (not rawget(db,"_global_")) then return; end			-- Not a root db
	db._global_[key]=value;
end

function api.EndPoint(db,value)
	local old=rawget(db,"_endpoint_");
	rawset(db,"_endpoint_",(value and true) or nil);
end

function util:Handle(point,create)
	if (d_type(point)=="database") then										-- database supplied
		if (rawget(point,"_lnk_")) then return util:Handle(point._lnk_); end	-- I'm a link
		return point;														-- no link, so real database
	end
	if (d_type(point)~="string") then return nil; end						-- not a tree, so error
	local prefix;				-- don't use self prefix as point may differ
	if (not point:match("^(%w-)"..dbc.Separator)) then prefix=point:sub(2);	-- root
	else prefix=(point:match("^(%w-)"..dbc.Separator)):sub(2); end			-- sub
	point={strsplit(dbc.Separator,point)};
	local goal=DM.DatabaseData;
	for num,entry in ipairs(point) do
		local loc=entry:sub(2); if (entry:sub(1,1)=="n") then loc=tonumber(loc); end
--		if (not goal[loc]) then
		if (d_type(goal[loc])~="database") then
			if (not create) then return nil; end
			goal[loc]=todatabase({});
		end
		goal=goal[loc];
	end
	return goal;
end

function api.SetRedundancy(db,key,r)
	db=util:Handle(db);								-- in case of link
	local ls,typ,val=db._str_[key]:match("^(%d+):(%a%a):[%+%-](.*)$");
	if (r) then r="-"; else r="+"; end
	db._str_[key]=ls..":"..typ..":"..r..val;
end

function api.GetRedundancy(db,key)
	db=util:Handle(db);								-- in case of link
	local r=db._str_[key]:match("^%d+:%a%a:([%+%-]).*$");
	return (r=="-" and true or nil);
end

function api.PruneMode(db,mode)
	db=util:Handle(db);								-- in case of link
	if (not db) then return; end
	if (mode) then rawset(db,"_prn_",mode); end
	return rawget(db,"_prn_");
end

function api.AddGuild(db,guild,forum,mode)
	local myGuild=GetGuildInfo("player"); if (not guild or guild=="" or guild==myGuild) then return; end
	db=util:GetRoot(db);												-- rewind tree
	local prefix=db:DM_Get("_prefix_");									-- get unique prefix
	if (not mode) then mode="NORMAL"; end								-- no exceptions

	dbc.Info.WhoList[guild]=true;										-- flag for member retrieval
	if (not db[membersDB]) then db[membersDB]={}; db[membersDB]:DM_EndPoint(true); end	-- prep list
	if (d_type(db[guildsDB])~="database") then db[guildsDB]={}; db[guildsDB]:DM_EndPoint(true); db[guildsDB]:DM_PruneMode(DM_PRUNE_NEVER); end		-- prep list
	-- Set shareable guild member list database
	if (not db[membersDB][guild]) then db[membersDB][guild]={}; end		-- create member-list if missing
	db[guildsDB][guild]=forum;											-- set guild and share-name in registry
	-- we obviously have someone to share with, so force a roster update to make
	-- sure the other guild(s) gets an updated list of our own members.
	if (not dbc.Addon[prefix].OwnGuildRosterDone) then
		GuildRoster();

		util:QueueForSend(nil,nil,dbd[prefix]._tree_,myGuild);	-- Send to everyone
		dbc.Addon[prefix].OwnGuildRosterDone=true;
	end
	-- create communications interface for external guilds
	local inter=DM.Com2:CreateInterface(prefix,dbc.Com2.RX,dbc.Com2.VC,dbc.Com2.DR,ef,"WHISPER");
	table.insert(dbc.Addon[prefix].iface,inter);						-- add it to addon's interface list
	inter:SetReceiver(db[membersDB][guild],guild);						-- set list in interface

end


-- membersDB (_DMDB-share-guildlistdatabases)
--     guild 1
--         toon 1 =
--         toon n =
--     guild n
--         toon n =

-- guildsDB (_DMDB-share-guildlists)
--     guild 1 = forum 1
--     guild n = forum n

function api.RemoveGuild(db,guild)
	if (not guild or guild=="") then return; end
	db=util:GetRoot(db);
	local prefix=db:DM_Get("_prefix_");
	DM.Database.Info.WhoList[guild]=nil;

	local dbname;
	dbname=membersDB;
	if (db[dbname]) then db[dbname][guild]=nil; end
	dbname=guildsDB;
	if (db[dbname]) then db[dbname][guild]=nil; end
end

function api.ListGuilds(db)
	db=util:GetRoot(db);
	if (not db[guildsDB]) then return {}; end
	local list={};
	for g,gd in pairs(db[guildsDB]) do if (gd) then table.insert(list,g); end end
	return list;
end

function api.UniqueKey(db)
	while(true) do
		local key="auk_"..string.format("%X",tostring(math.random(0x7FFFFFFF))).."_auk";
		if (not db._str_[key] and not rawget(db,key)) then return key; end
	end
end

function api.DatabaseRegistered(db)
	return true;
end

function api.ProgressIB(db)
	return dbc.Addon[db:DM_Get("_prefix_")].iface[1]:ProgressIB();
end

function util:SetDefaults(db)
	if (not rawget(db,"_global_")) then return; end		-- Not a root db
	for k,v in pairs(db._global_) do
		if (d_type(k)=="string") then
			local key=k:match("^(.+)_Default$");
			if (key) then db:DM_Set(key,v); end
		end
	end
end

function util:GetRoot(db)
	if (not db._tree_) then return nil; end
	db=util:Handle(db);		-- resolve current link if any
	if (not db._tree_) then return nil; end
	local prefix;
	if (not db._tree_:match("^(%w-)"..dbc.Separator)) then prefix=db._tree_:sub(2);	-- root
	else prefix=(db._tree_:match("^(%w-)"..dbc.Separator)):sub(2); end
	return dbd[prefix];
end

function util:SendDatabaseList(inter,prefix)
	local list={};
	util:AddDatabaseToList(dbd[prefix],list);
--	MessageBoard_LS.DatalistOut=DM:CopyTable(list);
	util:InterSend(inter,prefix,list,UnitName("player"),{ cont="dbl" },nil);		-- not local only
end

--
--	Sending interface for databases. All registered guilds are cycled.
--
--		util:InterSend(inter,prefix,value,nil,{ cont="val", key=key, tree=tree, ver=dbc.LastStamp },nil);		-- not local only
--		util:InterSend(inter,prefix,handle._str_,nil,{ cont="db", tree=tree, ver=dbc.LastStamp },nil);		-- not local only
function util:InterSend(inter,prefix,data,name,meta,localonly)
--if (meta and meta.cont=="dbl") then
--  say(prefix,"Sending my own databaselist.");
--elseif (meta and meta.tree) then
--  say(prefix,"Sending:",util:GetLuaName(meta.tree,meta.key),(meta.cont and meta.cont) or "with no contents descriptor.");
--else
--  say("Sending with no tree for",prefix);
--end

	if (inter) then inter:Send(data,name,meta); return; end				-- specified interface only
	for _,inter in pairs(dbc.Addon[prefix].iface) do
		inter:Send(data,name,meta);
		if (localonly) then return; end		-- guild share is index 1
	end
end

--
--	Requesting interface for databases. All registered guilds are cycled.
--
function util:InterRequest(inter,prefix,ver,name,meta,localonly)
	if (inter) then inter:Request(name,meta,ver); return; end			-- specified interface only
	for _,inter in pairs(dbc.Addon[prefix].iface) do
		inter:Request(name,meta,ver);
		if (localonly) then return; end		-- guild share is index 1
	end
end

function util:AddDatabaseToList(db,list)
	if (db._tree_=="") then print("Empty tree detected in AddDatabaseToList()"); return; end
	list[db._tree_]=self:GetIDForDatabase(db);		-- add yourself
	for _,dbTable in d_pairs(db,true) do			-- add nested databases without following links
		if (d_type(dbTable)=="database") then self:AddDatabaseToList(dbTable,list); end
	end
--	Save created database list for MessageBoard
--	if (MessageBoard_DB) then
--		local p;
--		if (not MessageBoard_DB.dbLists) then MessageBoard_DB.dbLists={}; end
--		p=MessageBoard_DB.dbLists;
--		if (not p[UnitName("player")]) then p[UnitName("player")]={}; end
--		p=p[UnitName("player")];
--		p[time()]=DM:CopyTable(list);
--	end
end

-- _str_ and _prn_ already be set by creation or load
function util:RegisterSubDatabases_(parent,key,db)
	rawset(db,"_tree_",parent._tree_..dbc.Separator..d_type(key):sub(1,1)..key);
	self:SetInterface(db);
	for k,v in pairs(db._str_) do
		if (v:match("^%d+:(%a%a):[%+%-].*$")=="da") then
			self:RegisterSubDatabases_(db,k,db[k]);
		end
	end
end

function util:RegisterSubDatabases(db)
	if (d_type(db)~="database") then return; end	-- can't work with this
	for k,v in pairs(db._str_) do
		if (v:match("^%d+:(%a%a):[%+%-].*$")=="da") then
			self:RegisterSubDatabases_(db,k,db[k]);
		end
	end
end

-- This does NOT send nested databases
function util:SendValue(inter,prefix,tree,key)
	-- Database list
	if (tree=="databaselist") then
		util:SendDatabaseList(inter,prefix);			-- OK, and poll it to get it going
		util:QueueForPoll(inter,prefix,"databaselist",UnitName("player"),0);
		return true;
	end

	-- Single values
	if (key) then
		local handle=util:Handle(tree);
		if (not handle) then return true; end	-- fake valid transmission to remove this abomination
		local value=handle[key];
		if (d_type(value)=="database") then return self:SendValue(inter,prefix,rawget(value,"_tree_"),nil); end
		util:InterSend(inter,prefix,value,nil,{ cont="val", key=key, tree=tree, ver=dbc.LastStamp },nil);		-- not local only
		return true;
	end

	-- Complete databases
	local handle=util:Handle(tree);
	local _,_,_,stamp=self:GetLocationInfo(tree);
	util:InterSend(inter,prefix,handle._str_,nil,{ cont="db", tree=tree, ver=dbc.LastStamp },nil);		-- not local only
	return true;
end

-- accepts both databases and trees
function util:GetLocationInfo(tree,create)
--	tree=util:Handle(tree); if (not tree._tree_) then return nil; end		-- resolve current link if any
	tree=util:Handle(tree,create); if (not tree or not tree._tree_) then return nil; end		-- resolve current link if any
	tree=tree._tree_;
	if (not tree:find(dbc.Separator)) then return nil,"root","database",0; end	-- root - don't mess with it
	local parent,typ,name=tree:match("^(.+)"..dbc.Separator.."(%a)(.-)$");	-- last separator
	if (typ=="n") then name=tonumber(name); end
	parent=util:Handle(parent);		-- get parent
	if (not parent) then return "unreachable"; end
	local stamp=tonumber(parent._str_[name]:match("^(%d+):%a%a:[%+%-].*$"));
	return parent,name,d_type(parent[name]),stamp;	-- parentDB,key,type,stamp
end


function dbc.Com2.ReDirect(inter,meta)
	if (meta and meta.forum) then
		for id,db in pairs(dbd) do											-- walk all known databases
			local buf=db[guildsDB] or {};									-- grab forum list for this database
			for sg,gForum in d_pairs(buf) do								-- walk all its forum names
				if (gForum==meta.forum and (db[membersDB][sg] or {})[sender]) then	-- if a forum matches and member is in that guild
					return id,db;											-- found goal database ad 'db' with prefix 'id'
				end
			end
		end
	end
	return inter.ID,dbd[inter.ID];
end

--	t.cbRX=cbRX;	-- data RX
--		cbRX(interface,data,data name,metadata,sender toon,api used);
function dbc.Com2.RX(inter,data,name,meta,sender,api)
	local pf,db=dbc.Com2.ReDirect(inter,meta);
	if (not DM.DatabaseQueue[pf]) then return; end		-- Quick way out
	if (not db) then return; end						-- Quick way out


	if (meta.cont=="val") then						-- incoming value
		if (not meta.tree) then return; end				-- all types must have a tree
		if (not meta.key) then return; end;			-- no key
--say(pf,": Received value for",util:GetLuaName(meta.tree,meta.key));
		local goal=util:Handle(meta.tree,true);	-- Create it if need be
		local newdata=nil; if (not goal._str_[meta.key]) then newdata=true; end
		_=goal[meta.key];							-- Update stamp
		if (dbc.LastStamp>meta.ver) then
--say("My current data is newer, so aborting save.");
--saydate(dbc.LastStamp,meta.ver);
			return; end	-- mine's better
		-- check for guild member-lists and discard own guild listing
		if (meta.tree:match(membersDB.."$")) then	-- check for this as an ending
			local myGuild=GetGuildInfo("player");
			if (meta.key==myGuild) then return; end	-- don't accept own guild at all
		end
		-- normal storage
		util.LockedTime=meta.ver;					-- Override time
		goal[meta.key]=data;						-- Set value
		util.LockedTime=nil;						-- Re-enable time
--say("Saved.");
		if (not tostring(meta.key):match("^_DMDB-")) then		-- not a hidden share
			--print("Called util:SendEvent with pf: "..tostring(pf).." goal: "..tostring(goal).." meta.key: "..tostring(meta.key).." newdata: "..tostring(newdata))
			util:SendEvent(pf,"DMEVENT_DATABASE_SYNC",goal,meta.key,newdata);
		end
		if (inter.Channel:upper()=="WHISPER" and inter.Receiver) then	-- received from external source
--say("Private reception. Queuing for guild broadcast.");
			util:QueueForSend(nil,pf,meta.tree,meta.key,true);	-- distribute locally
		end
		return;
	elseif (meta.cont=="db") then					-- incoming database
		if (not meta.tree) then return; end				-- all types must have a tree
--say(pf,": Received database",util:GetLuaName(meta.tree));
		local parent,pkey=util:GetLocationInfo(meta.tree,true);	-- create if not there
		if (not parent and not pkey) then
--say("ERROR: Anon database without parent detected.");
		return; end	-- no parent, can't save (do not create it)
		if (parent) then
			if (parent._str_[pkey]) then
				_=parent[pkey];						-- update stamp reading and save it for single access
				if (d_type(_)~="database") then
					if (meta.ver<=dbc.LastStamp) then
--say("Current non-db data is newer than incoming database.");
--saydate(dbc.LastStamp,meta.ver);		-- first current, second incoming/external
					return; end	-- not newer received
					util.LockedTime=meta.ver; parent[pkey]={}; util.LockedTime=nil;
--say("Current older non-db data was overwritten by empty db at",date("%c",meta.ver));
				end
			else	-- key does not exist
				util.LockedTime=meta.ver; parent[pkey]={}; util.LockedTime=nil;
--say("Empty nested database created at",date("%c",meta.ver));
			end
		elseif (pkey~="root") then
--say("ERROR: Non-root database without parent detected.");
		return; end		-- some other problem
		-- merge entries
		local handle=util:Handle(meta.tree);
--say("Merging incoming database",util:GetLuaName(meta.tree));
		for entry,eData in pairs(data) do			-- data is remote raw _str_ table
			local newdata,savedata=nil,nil;
			local sNew,tNew=eData:match("^(%d+):(%a%a):[%+%-].*$"); sNew=tonumber(sNew);
			if (not handle._str_[entry]) then				--==>> it's new data
				newdata=true; savedata=true;
			else
				local sOld,dOld,tOld=handle._str_[entry]:match("^(%d+)(:(%a%a):[%+%-].*)$"); sOld=tonumber(sOld);
				if (tOld~="da" or tNew~="da") then			--==>> one is not db
					if (sNew>sOld) then savedata=true;
--say("Existing type",tOld,"will be overwritten. - Updated stamp:");
--saydate(sOld,sNew);		-- first current, second incoming/external
					end	--==>> do normal evaluation
				elseif (sNew>sOld) then						--==>> both are databases
					rawset(handle._str_,entry,sNew..dOld);	-- update stamp only
--say("Existing and incoming nested database detected:",util:GetLuaName(handle,entry),"- Updated stamp:");
--saydate(sOld,sNew);		-- first current, second incoming/external
				end
			end
			if (savedata) then
				if (tNew=="da") then						-- a database is incoming
					util.LockedTime=sNew;					-- set lock
					handle[entry]={};						-- overwrite/save by meta
					util.LockedTime=nil;					-- clear lock
--say("Creating empty nested database at",util:GetLuaName(handle,entry),"at",date("%c",sNew));
				else										-- non-database incoming
					handle._str_[entry]=eData;				-- store raw non-database entry
--say("Storing raw "..tNew.."-type data at",util:GetLuaName(handle,entry));
				end
				--print("Called util:SendEvent with pf: "..tostring(pf).." handle: "..tostring(handle).." entry: "..tostring(entry).." newdata: "..tostring(newdata))
				util:SendEvent(pf,"DMEVENT_DATABASE_SYNC",handle,entry,newdata);	-- updated
			end
		end
		if (inter.Channel:upper()=="WHISPER" and inter.Receiver) then	-- received from external source
--say("Private reception. Queuing for guild broadcast.");
			util:QueueForSend(nil,pf,meta.tree,meta.key,true);	-- distribute locally
		end
		return;
	elseif (meta.cont=="dbl") then					-- incoming database list
		if (not DM.Database.Addon[pf].dbL[name]) then DM.Database.Addon[pf].dbL[name]={}; end
		wipe(DM.Database.Addon[pf].dbL[name]);	-- better for garbagecollect to wipe first
		DM.Database.Addon[pf].dbL[name]=DM:CopyTable(data,DM.Database.Addon[pf].dbL[name]);
--say("Databaselist for",name,"received. Stored. No other action taken.");
		-- no need for local sync of incoming external databaselist
		return;
	end
end

--	t.cbVC=cbVC;	-- data version check typically for negotiation
--		cbVC(interface,data name,metadata);
function dbc.Com2.VC(inter,name,meta)
	local pf,db=dbc.Com2.ReDirect(inter,meta);
	if (not DM.DatabaseQueue[pf]) then return; end		-- Quick way out
	if (not db) then return; end						-- Quick way out

	-- A poll for value stamp
	if (meta.key and meta.tree) then
--say("Someone is checking my version of value",util:GetLuaName(meta.tree,meta.key));
		local handle=util:Handle(meta.tree); if (not handle) then return 0; end
		return tonumber(handle._str_[meta.key]:match("^(%d+):%a%a:[%+%-].*$"));
	end

	-- A poll for database ID/version/stamp
	if (not meta.key and meta.tree and meta.ve) then
		local handle=util:Handle(meta.tree); if (not handle) then
--say("Someone is checking my version for database",util:GetLuaName(meta.tree),"=> Don't have it.");
		return 0; end
		if (util:GetIDForDatabase(handle)==meta.ve) then
--say("Someone is checking my version for database",util:GetLuaName(meta.tree),"=> I have the same version.");
		return 0; end
--say("Someone is checking my version for database",util:GetLuaName(meta.tree),"=> Mine is different.");
		return 1;
	end

	-- A poll for database list ID (ID=toon=meta.key)
	if (meta.key and meta.dblist) then
		if (dbc.Addon[pf].dbL[meta.key]) then
			local list={};
			util:AddDatabaseToList(dbd[pf],list);				-- Make my own database list
			-- Remove same and non-existent based on own database
			for tree,id in pairs(dbc.Addon[pf].dbL[meta.key]) do	-- cycle external list
				if (list[tree] and list[tree]==id) then list[tree]=nil; end	-- Both have same
			end
			-- Walk it
			local stamper=0;
			for _ in pairs(list) do stamper=stamper+1; end
			return stamper;
		end
--say("Someone is checking databasedlist version for a missing list from user",meta.key);
		return 0;
	end

	-- default
--say("Someone is making an illegal version check:");
--dbg("meta",meta);
	return 0;
end

--	t.cbDR=cbDR;	-- a request for data to be sent (typ. from a negotiation)
--		cbDR(interface,data name,metadata);
function dbc.Com2.DR(inter,name,meta)
	local pf,db=dbc.Com2.ReDirect(inter,meta);
	if (not DM.DatabaseQueue[pf]) then return; end		-- Quick way out
	if (not db) then return; end						-- Quick way out

--dbg(pf..": DR called",meta);
--print(dbc.Addon[pf]);
	-- databaselist
	if (meta.key and meta.dblist) then									-- it's a database list
		if (not dbc.Addon[pf] or not dbc.Addon[pf].dbL[meta.key]) then return; end	-- i don't have any such stuff
--say("I need to queue databaselist differences for",meta.key,pf);
		local list={};
		util:AddDatabaseToList(db,list);								-- create my own database list
		if (DM:CompareTables(list,dbc.Addon[pf].dbL[meta.key])) then return; end	-- they're equal, so nothing to do
		for tree,id in pairs(dbc.Addon[pf].dbL[meta.key]) do		-- cycle external list
			if (list[tree] and list[tree]==id) then						-- i have the same entry
				list[tree]=nil;											-- so remove it from my list
			end
		end

--say("Differences for",meta.key,pf,"database are:");
		for tree in pairs(list) do util:QueueForSend(nil,pf,tree);	-- queue all databases that are different for send
--say("==>",util:GetLuaName(tree));
		end	-- all non-equal databases
		util:QueueForSend(nil,pf,"databaselist");				-- send own list when all done
		return;
	end

	-- database and value
--say(pf,": Generic request for",util:GetLuaName(meta.tree,meta.key),"queued for output.");
	util:QueueForSend(nil,pf,meta.tree,meta.key,nil);
end

-- get one-level unique(ish) ID for a database
function util:GetIDForDatabase(db)
	if (not db) then return 0; end								-- nothing i can use
	local id=0;													-- new id
	for _,eData in pairs(db._str_) do							-- walk storage
		id=id+tonumber(eData:match("^(%d+):%a%a:[%+%-].*$"));	-- add stamp
	end
--	for k,v in d_pairs(db,true) do								-- walk without following links
--		if (d_type(v)=="database") then							-- it's a database
--			k=tostring(k);										-- grab key as string
--			for i=1,k:len() do id=id+k:byte(i); end				-- add in all byte-values
--		end
--	end
	if (id==0) then id=1; end									-- id=0 is a special case, so avoid that
	return id;
end

local function getprefix(prefix,tree)
	if (prefix and prefix~="") then return prefix; end
	if (not tree:match("^(%w-)"..dbc.Separator)) then return tree:sub(2); end	-- root
	return (tree:match("^(%w-)"..dbc.Separator)):sub(2);						-- nested
end


--function DM.Database.Util:QueueForSend(prefix,tree,key,noguild)
function util:QueueForSend(inter,prefix,tree,key,localonly)
	if (tree=="") then return; end			-- skip unrooted databases (typically todatabase)
	prefix=getprefix(prefix,tree);			-- recreate prefix if not specified

--say(prefix,"Queue for send",util:GetLuaName(tree,key),(localonly and "=> local") or "global");
	DM.Talk[prefix]=time();					-- Delay further
	local q={ method="send", tree=tree, key=key, inter=inter };						-- basic metadata
	if (tree=="databaselist") then q.offset=20; end						-- additional wait
	q.localonly=localonly;

	dbd[prefix]=dbd[prefix]+q;				-- add event to queue
end

--function DM.Database.Util:QueueForPoll(prefix,tree,key,stamp,noguild)
function util:QueueForPoll(inter,prefix,tree,key,stamp)
	if (tree=="") then return; end			-- skip unrooted databases (typically todatabase)
	prefix=getprefix(prefix,tree);			-- recreate prefix if not specified
--say(prefix,"Queue for poll",util:GetLuaName(tree,key));
	DM.Talk[prefix]=time();					-- Delay further
	local q={ method="poll", tree=tree, key=key, inter=inter, instamp=stamp };
	dbd[prefix]=dbd[prefix]+q;
end

function d_next(t,k)
	if (d_type(t)~="database") then return next(t,k); end
	local kv=nil;
	repeat repeat
			k=next(t._str_,k);
			if (k) then kv=t[k]; end
		until (not tostring(k):match("^"..hiddenKey));	-- get next if this is a database zero (valid key, nil data)
	until (not k or kv);			-- Invoke metatable for data
	return k,kv;
end

-- Support for metatable types
function d_type(data)
	local m=getmetatable(data);
	if (m and m.__type) then return m.__type; end
	return type(data);
end

function table.dmsort(t,func)
	local buf={};
	for k,d in pairs(t._str_) do table.insert(buf,{k=k,d=d}); end
	table.sort(buf,func);
	return buf;
end

-- The iterator allows nil-values, and stops when the key does not exist.
function d_pairs_up(t,nl)
	if (d_type(t)~="database") then return pairs(t); end		-- fallback
	if (rawget(t,"_lnk_")) then
		if (t:DM_Get("PauseLinks")) then nl=true; end
		if (nl) then return function() return nil; end,t,nil; end		-- empty iterator
		t=util:Handle(rawget(t,"_lnk_"));
		if (not t) then return function() return nil; end,t,nil; end	-- empty iterator
	end
	local oi=table.dmsort(t,
				function(A,B)
					if (tonumber(A.d:match("^(%d+):%a%a:[%+%-].*$"))<tonumber(B.d:match("^(%d+):%a%a:[%+%-].*$"))) then return true; end
				end
			);
	return function (t,k)
		local kv=nil;
		repeat repeat
				k=table.remove(oi,1);	-- Pull top one and get its data
				if (not k) then return nil,nil; end
				if (k.k) then kv=t[k.k]; end	-- Invoke metatable for data
			until (not tostring(k.k):match("^"..hiddenKey));	-- get next if this is a database zero (valid key, nil data)
		until (not k.k or kv or not tostring(k.k):match("^"..hiddenKey));	-- get next if this is a database zero (valid key, nil data)
		return k.k,kv;
	end,t,nil;
end

-- The iterator allows nil-values, and stops when the key does not exist.
function d_pairs_dn(t,nl)
	if (d_type(t)~="database") then return pairs(t); end		-- fallback
	if (rawget(t,"_lnk_")) then
		if (t:DM_Get("PauseLinks")) then nl=true; end
		if (nl) then return function() return nil; end,t,nil; end		-- empty iterator
		t=util:Handle(rawget(t,"_lnk_"));
		if (not t) then return function() return nil; end,t,nil; end	-- empty iterator
	end
	local oi=table.dmsort(t,
				function(A,B)
					if (tonumber(A.d:match("^(%d+):%a%a:[%+%-].*$"))>tonumber(B.d:match("^(%d+):%a%a:[%+%-].*$"))) then return true; end
				end
			);
	return function (t,k)
		local kv=nil;
		repeat repeat
				k=table.remove(oi,1);	-- Pull top one and get its data
				if (not k) then return nil,nil; end
				if (k.k) then kv=t[k.k]; end	-- Invoke metatable for data
			until (not tostring(k.k):match("^"..hiddenKey));	-- get next if this is a database zero (valid key, nil data)
		until (not k.k or kv or not tostring(k.k):match("^"..hiddenKey));	-- get next if this is a database zero (valid key, nil data)
		return k.k,kv;
	end,t,nil;
end

-- The iterator allows nil-values, and stops when the key does not exist.
function d_pairs(t,nl)
	if (d_type(t)~="database") then return pairs(t); end
	if (rawget(t,"_lnk_")) then
		if (t:DM_Get("PauseLinks")) then nl=true; end
		if (nl) then return function() return nil; end,t,nil; end		-- empty iterator
		t=util:Handle(rawget(t,"_lnk_"));
		if (not t) then return function() return nil; end,t,nil; end	-- empty iterator
	end
	return function (t,k)
		local kv=nil;
		repeat repeat
				k=next(t._str_,k);
				if (k) then kv=t[k]; end
			until (not tostring(k):match("^"..hiddenKey));	-- get next if this is a database zero (valid key, nil data)
		until (not k or kv);	-- Invoke metatable for data
		return k,kv;
	end,t,nil;
end

-- The iterator allows nil-values, and stops when the key does not exist.
function d_ipairs(t,nl)
	if (d_type(t)~="database") then return ipairs(t); end
	if (rawget(t,"_lnk_")) then
		if (t:DM_Get("PauseLinks")) then nl=true; end
		if (nl) then return function() return nil; end,t,0; end		-- empty iterator
		t=util:Handle(rawget(t,"_lnk_"));
		if (not t) then return function() return nil; end,t,0; end	-- empty iterator
	end
	return function (t,k)
		k=k+1;
		if (t._str_[k]==nil) then return nil,nil; end
		return k,t[k];			-- Invoke metatable for data
	end,t,0;
end

--local function dbPF(comprefix,gettoon)
--	if (not DM.Net.DB[comprefix].Receiver) then return comprefix; end
--	local dbprefix,toon=comprefix:match("^(.+)"..DM.Net.Split1.."(.-)$");
--	if (not dbprefix or not toon) then return; end
--	if (gettoon) then return toon; end			-- to have only single-value returns
--	return dbprefix;
--end

--function dbc.Info:GetOnlineGuildMembers(guild)
--	SetWhoToUI(1);
--	FriendsFrame:UnregisterEvent("WHO_LIST_UPDATE");
--	SendWho("g-\""..guild.."\"");
--end

function util.MT:UpdateGuildLists()

	if (util.UpdatingGuildLists) then return; end
	util.UpdatingGuildLists="";

	SetWhoToUI(1);
	FriendsFrame:UnregisterEvent("WHO_LIST_UPDATE");
	repeat
		local guild,state=next(dbc.Info.WhoList);
		if (not guild or not state) then break; end		-- list is now empty
		local retry=0;
		repeat
			retry=retry+1;
			SendWho("g-\""..guild.."\"");		-- request these members
			local now=debugprofilestop();
			repeat DM.MT:Yield(true); until (debugprofilestop()-now>=5000 or util.UpdatingGuildLists==guild);
		until (util.UpdatingGuildLists==guild or retry>=3);		-- keep going until guild has been received or retried too many times
		if (util.UpdatingGuildLists==guild) then						-- got this guild
			-- set new list to interface here
			for prefix,db in pairs(dbd) do								-- do all prefixes
				for _,inter in pairs(dbc.Addon[prefix].iface) do		-- do all interfaces for this prefix
					if (inter.ReceiverListName==guild) then				-- there's an interface for this guild at this prefix
						inter:SetReceiver(db[membersDB][guild],guild);	-- so plug list in there
					end
				end
			end
		else
			dbc.Info.WhoList[guild]=nil;								-- most likely no members online
		end
	until (false);		-- forever
	FriendsFrame:RegisterEvent("WHO_LIST_UPDATE");

	util.UpdatingGuildLists=nil;
end

function dbc.Info:UpdateGuildMemberList(guild,list,merge)

	for prefix,db in pairs(dbd) do
--		DM.Script:SetValue(prefix,"UEGwait",-1);	-- tell potential scripts that a list is here
		if (db[membersDB]) then
			local oldlist,changed,valid=db[membersDB][guild],false,true;	-- select guild
			if (oldlist) then												-- old list exists
				for toon in pairs(oldlist) do if (not list[toon]) then changed=true; break; end end
				if (not changed) then
					for toon in pairs(list) do if (not oldlist[toon]) then changed=true; break; end end
				end
			elseif (guild==GetGuildInfo("player")) then changed=true;		-- always update own guild
			else valid=nil; end												-- not one of mine by AddGuild
			if (valid and changed) then
				if (merge) then list=DM:CopyTable(oldlist,list); end		-- merge old and new
				db[membersDB][guild]=list;	-- pack and store
			end
		end
	end
	util.UpdatingGuildLists=guild;			-- tell MT guild has been received
	dbc.Info.WhoList[guild]=nil;			-- pull it from the queue
end

-- Functions for coroutine usage
function util.MT:Prune(prefix)
	local pDB=dbd[prefix];
	local count=self:Prune_FullEntryCount(pDB)-pDB._global_.Limit;		-- Determine amount of pruning
-- if (count<0) then say(prefix,"There are",0-count,"entries available in the database.");
-- elseif (count>0) then say(prefix,"There are",count,"entries to remove from the database.");
-- else say(prefix,"The database is full with no need to remove anything yet."); end
	if (count>0) then
		local list=self:Prune_MakeList(pDB,count);				-- List X oldest redundant entries
		for _,e in ipairs(list) do
			local thisdb=util:Handle(e.tree);
			if (d_type(thisdb[e.key])=="database") then
				util.LockedTime=1;			-- set to not start sync, just something low as it will be wiped immediately anyway
				thisdb[e.key]=nil;			-- set via meta to correctly wipe associated table
				util.LockedTime=nil;		-- nothing
			end
			thisdb._str_[e.key]=nil;		-- wipe it
		end
	end
end

-- "Prune_MakeList" will seek to add everything that is not a populated
-- database or a link to another database.
-- args are (db,count) in initial call. For subsequent calls
-- args are (db,count,list).
function util.MT:Prune_MakeList(db,count,list)
	if (not list) then list={}; end
	db=util:Handle(db);
	for name,entry in d_pairs_up(db,true) do					-- no links
		if (d_type(entry)=="database" and next(entry._str_)) then self:Prune_MakeList(entry,count,list);	-- db with contents, add to list
		else self:Prune_AddOneTolist(count,list,db,name); DM.MT:Yield(true); end							-- a normal entry, add it
	end
	return list;
end

-- entry:
--   tree : the tree
--   key  : the key
--   r    : redundant
--   stamp: stamp
--   p    : prune-mode

--> SORT: if (A<B) then return true; end
-- "redundant" takes precedence over DM_PRUNE_NEVER.
function util.MT:Prune_AddOneTolist(count,list,db,key)
	local s,r=db._str_[key]:match("^(%d+):%a%a:([%+%-]).*$");
	if (db._prn_==DM_PRUNE_NEVER and r~="-") then return; end		-- non-redundant non-removable
	table.insert(list,{ tree=db._tree_, key=key, r=(r=="-"), stamp=tonumber(s), p=db._prn_ });
	if (#list<=1) then return; end			-- first entry, keep it

	table.sort(list,function (a,b)
		if (a.r~=b.r) then return a.r; end						-- set redundant before non-redundant
		if (a.r) then return (a.stamp<b.stamp); end				-- both redundant, only stamp sort
		-- non-redundant sorting
		if (a.p~=b.p) then																	-- different prune settings
			if (a.p>=DM_PRUNE_FIRST and b.p>=DM_PRUNE_FIRST) then return (a.p<b.p); end		-- both have fixed order, so sort by lowest
			return (b.p==DM_PRUNE_NORMAL);													-- one is normal, and is to be last (a above b)
		end
		-- non-redundant, same prune
		return (a.stamp<b.stamp);			-- same group, sort by stamp
	end );

	if (#list>count) then table.remove(list); end		-- kill last if list is too long
end

--	DM_PRUNE_NEVER=0,
--	DM_PRUNE_NORMAL=1,
--	DM_PRUNE_FIRST=10,		-- 11 second, 12 third, etc

function util.MT:Prune_FullEntryCount(db)
	local count=0;
	for _,pDB in d_pairs(db,true) do
		count=count+1;
		if (d_type(pDB)=="database") then
			count=count+self:Prune_FullEntryCount(pDB);
			DM.MT:Yield(true);		-- Immediate yield for each database
		end
	end
	return count;
end


--
-- deprecated by adding channel loopback
--
-- dummy event for channels that are slow to get going
--function wq:EMPTY(inter,prefix,tree,key)
--say("From",(inter and inter.ID) or "no ID","or",prefix,"Queue: Empty kickstart");
--	dbd[prefix]=dbd[prefix]-1;	-- Remove directly
--end

-- transmit stuff
function wq:send(inter,prefix,tree,key)
--say("From",(inter and inter.ID) or "no ID","or",prefix,"Queue: Send",util:GetLuaName(tree,key));
	util:SendValue(inter,prefix,tree,key);
	dbd[prefix]=dbd[prefix]-1;
end

-- poll stuff
function wq:poll(inter,prefix,tree,key,ver)
--say("From",(inter and inter.ID) or "no ID","or",prefix,"Queue: Poll",util:GetLuaName(tree,key));
	if (tree=="databaselist") then		-- Poll database list
		util:InterRequest(inter,prefix,ver,nil,{ dblist=true, key=key },nil);	-- set databaselist type
		dbd[prefix]=dbd[prefix]-1;
		return;
	end

	-- if "key" is nil, a database will be polled
	util:InterRequest(inter,prefix,ver,nil,{ tree=tree, key=key },nil);
	dbd[prefix]=dbd[prefix]-1;
end

function dbc:Heartbeat(elapsed)
	dbc.LastPrefixIterator=next(dbd,dbc.LastPrefixIterator);
	if (dbc.LastPrefixIterator) then dbc:Heartbeat_(dbc.LastPrefixIterator,time()); end
end

function dbc:Heartbeat_(prefix,now)
	if (not DM.Talk or not DM.Talk[prefix]) then return; end
	if (not dbc.Addon[prefix]) then return; end

	-- Get next queue item
	local entry=dbd[prefix]("event_queue_peek"); if (not entry or entry.busy) then return; end			-- peek next event
	if (not entry.offset) then entry.offset=0; end						-- extra time offset/hold check

	-- run window
	local delay=0;	-- neg is 2s, may need some consideration here, maybe not as inter should buffer and hold
	if (entry.offset>0) then delay=7; end
	if (now-DM.Talk[prefix]<math.random()*10+delay+entry.offset) then return; end
	entry.busy=true;									-- prevent re-entry for this prefix

	-- event dispatcher
	if (wq[entry.method]) then
		wq[entry.method](wq,entry.inter,prefix,entry.tree,entry.key,entry.instamp);
	else
		dbd[prefix]=dbd[prefix]-1;	-- Remove unknown entry
	end
end

--		util:SendEvent(inter.ID,"DMEVENT_DATABASE_SYNC",goal,meta.key,newdata);
function util:SendEvent(prefix,event,handle,key,newdata)
	if (not dbc.Addon[prefix] or not dbc.Addon[prefix].EventFrame) then return; end
	local handler=dbc.Addon[prefix].EventFrame:GetScript("OnEvent");
	if (not handler) then return; end
--	EventFrame:OnEvent(event,database,key,newdata)
	handler(dbc.Addon[prefix].EventFrame,event,handle,key,newdata);
end

-- accepts both databases and trees
function util:Name(db)
	local tree=db;
	if (d_type(db)=="database") then tree=db._tree_; end
	tree={strsplit(dbc.Separator,tree)};
	for num,entry in ipairs(tree) do
		local loc=entry:sub(2); if (entry:sub(1,1)=="n") then loc=tonumber(loc); end
		tree[num]=loc;
	end
--	tree[1]=self:GetRoot(db);
	return tree;
end

function util:GetLuaName(tree,key)
	if (not key and d_type(tree)=="table") then tree,key=tree.tree,tree.key; end
--print("GetLoc in:",tree:gsub(dbc.Separator,"%^"));
	if (tree=="databaselist") then
		if (key and tostring(key)~="") then return "databaselist ("..key..")"; end
		return "databaselist";
	end
	local tree=util:Name(tree);
	if (key and key~="") then tree[(#tree)+1]=key; end	-- numeric key
	local loc=tree[1];
	for i,value in ipairs(tree) do
--print("Name",i,": ",value);
		if (i>1) then
			if (d_type(value)=="number") then loc=loc.."["..value.."]";
			elseif (not value:find("[%s%-%+]")) then loc=loc.."."..value;
			else loc=loc.."[\""..value.."\"]"; end
		end
	end
	if (not key) then loc=loc.." (db)"; elseif (key~="") then loc=loc.." (key)"; end
	return loc;
end

