--[[
	MessageBoard - A World of Warcraft In-game Forum
	Copyright (C) 2009-2015  Darrell Keys

	MessageBoard is free software: you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.

	MessageBoard is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with MessageBoard.  If not, see <http://www.gnu.org/licenses/>.
]]
--      Many thanks to former author Dag Bakken on helping me get this up and running.
local _, MessageBoard = ...
local L = MessageBoard.L;
local addonName, PRIVATE = ... -- Blizzard passes us the addon name and a table to all addon files in the ellipsis variable
PRIVATE.Functions = {} -- create a subtable to store our utility functions so we can call them from anywhere in our addon without making them global
local MBFunc = PRIVATE.Functions -- get a reference to our functions table with a friendly/short name
DuckLibMB=DuckMod[2.9991];
local DuckLib=DuckLibMB;
local Net=DuckLib.Net;
local DB=DuckLib.Database;
local ren=DuckLibMB.Render;
local pairs=d_pairs;
local ipairs=d_ipairs;
local pairs_up=d_pairs_up;
local pairs_dn=d_pairs_dn;
local type=d_type;
local CategoryLevel;
BINDING_HEADER_MESSAGEBOARD = "MessageBoard bindings";
BINDING_NAME_TOGGLEMESSAGEBOARD = "Toggle MessageBoard";
SLASH_MESSAGEBOARD1 = "/mboard";
SLASH_MESSAGEBOARD2 = "/mb";
local MB_PREFIX = "MB2005";
local MB_HEARTBEAT = 1/3;
local L = LibStub("AceLocale-3.0"):GetLocale("MessageBoard")
local version = "2.05"
local iVersion = 2005
local MBDialog = LibStub("LibDialog-1.0")
local Ambiguate = Ambiguate
local _MBLDB
local _minimapIcon
local vface

MessageBoard_LS = {
	IconPosition=180,
	Debug=false,
	Quoted=nil,
	NewPostNotify=true,
	NewPostNotifyPopup=true,
	Background="Interface\\Addons\\MessageBoard\\Images\\MessageBoardBG-MOP",
	OptBackground="Interface\\Addons\\MessageBoard\\Images\\MessageBoard_Options-MOP",
	buttonbg="Interface\\Addons\\MessageBoard\\Images\\buttonbg-MOP",
	BackgroundName=L["BACKGROUND_MOP"],
	Draft=nil,
	DraftTitle=nil,
};

MessageBoard_DB = {
}

MessageBoardLDBIconDB = {
}

MessageBoardXML = {
	Editor={
		ID_MANAGEMENTNOTE = 1, -- management note
		ID_NEWTOPIC       = 2, -- new topic
		ID_NEWMESSAGE     = 3, -- reply
		ID_EDITMESSAGE    = 4, -- edit
		ID_NEWPOLL		  = 5, -- new poll
	},
	Admin={},
	MyRank=nil,
	Administrator=nil;
};

local MessageBoard = {
	HaveRoster=nil,
	GuildName=nil,
	Ready=nil,
	Selected = {
		Topic=nil,
	},
	Timers = {
		Last=0,
	},
	Administrator=nil,
	selected=nil,
	disabledButton=nil,
	PositionViewport=nil,
	FirstRun=true,
	LoggedIn=false,
};

local Icon={
	ANNOUNCE = "Interface\\Buttons\\UI-GuildButton-MOTD-Disabled",
	ANNOUNCE_new = "Interface\\Buttons\\UI-GuildButton-MOTD-Up",
	STICKY = "Interface\\Buttons\\UI-GuildButton-OfficerNote-Disabled",
	STICKY_new = "Interface\\Buttons\\UI-GuildButton-OfficerNote-Up",
	NORMAL = "Interface\\Buttons\\UI-GuildButton-PublicNote-Disabled",
	NORMAL_new = "Interface\\Buttons\\UI-GuildButton-PublicNote-Up",
	LOCKED = "Interface\\AddOns\\MessageBoard\\Images\\topic-icon-lockedread",
	LOCKED_new = "Interface\\AddOns\\MessageBoard\\Images\\topic-icon-lockedunread",
	POLL = "Interface\\AddOns\\MessageBoard\\Images\\topic-icon-pollread",
	POLL_new = "Interface\\AddOns\\MessageBoard\\Images\\topic-icon-pollunread",
};

local Colour={
	Unread={ r=0.30, g=0.30, b=0.85, },
	Read={ r=1.0, g=0.8196079, b=0.0, },
};

-- function MessageBoard:OnInitialize()
	-- vface=DuckLib.Com2:CreateInterface("MBMBV",MBverRX);   -- create communication interface
-- end

function MessageBoard_OnLoad(self)
	self:RegisterEvent("ADDON_LOADED");	
end

function MessageBoard_OnEvent(self, event, ...)
	if (event=="ADDON_LOADED" and ... == "MessageBoard") then
		self:UnregisterEvent("ADDON_LOADED");
		self:RegisterEvent("PLAYER_LOGIN");
		vface=DuckLib.Com2:CreateInterface("MBMBV",MBverRX);   -- create communication interface
		if not MBDb then MBDb = {} end
		MBFunc.InitializeMinimap()
	end
	if (event=="PLAYER_LOGIN") then
		tinsert(UISpecialFrames,"MessageBoard_Main");
		tinsert(UISpecialFrames,"MessageBoard_OFrame");
		SlashCmdList["MESSAGEBOARD"]=function(msg) MessageBoard.Slasher(msg) end;
		if (not MessageBoard_LS.Passwords) then MessageBoard_LS.Passwords={}; end -- setup per character savevariables
		if MessageBoard_LS.Quoted==nil then MessageBoard_LS.Quoted=true; end -- setup default quoted text
		if MessageBoardLDBIconDB.hide==nil then MessageBoardLDBIconDB.hide=true; end
		if MessageBoard_LS.NewPostNotify==nil then MessageBoard_LS.NewPostNotify=true; end -- setup default notification text
		if MessageBoard_LS.NewPostNotifyPopup==nil then MessageBoard_LS.NewPostNotifyPopup=true; end -- setup default notification text popup
		if MessageBoard_LS.Background==nil then MessageBoard_LS.Background="Interface\\Addons\\MessageBoard\\Images\\MessageBoardBG-MOP"; end
		if MessageBoard_LS.OptBackground==nil then MessageBoard_LS.OptBackground="Interface\\Addons\\MessageBoard\\Images\\MessageBoard_Options-MOP"; end
		if MessageBoard_LS.buttonbg==nil then MessageBoard_LS.buttonbg="Interface\\Addons\\MessageBoard\\Images\\buttonbg-MOP"; end
		if MessageBoard_LS.BackgroundName==nil then MessageBoard_LS.BackgroundName="Mists of Pandaria"; end
		MessageBoard_MainLowerFrameBG:SetTexture(MessageBoard_LS.Background); -- set bg
		for i=1,9 do
			_G["MessageBoard_MainNavBarButton"..i.."BG"]:SetTexture(MessageBoard_LS.buttonbg)
		end
		DuckLib:Init();
		MessageBoard_MainNewTopic:SetText(L["MAIN_NEW_TOPIC"]);
		MessageBoard_MainMsgReply:SetText(L["MAIN_NEW_REPLY"]);
		MessageBoard_MainNewPoll:SetText(L["MAIN_NEW_POLL"]);
		self:RegisterEvent("GUILD_ROSTER_UPDATE");
		self:UnregisterEvent("PLAYER_LOGIN");
		self:RegisterEvent("CHAT_MSG_ADDON");
		self:UnregisterEvent("PLAYER_LOGIN");
		MessageBoard.LoggedIn=true;
		MBFunc.MBGuildFrame_UpdateTabard()
	end
	if (event=="GUILD_ROSTER_UPDATE") then
		local seq=1;
		local me=tostring(UnitName("player"))
		local name,_,_,_,_,_,pnote,onote=GetGuildRosterInfo(seq);
		while(name) do
			name = Ambiguate(name,"none")
			if (name==me) then
				if (onote and string.find(onote,"#MB:A")) or (MessageBoardXML:Setting("UseAlternateAdmin") and pnote and string.find(pnote,"#MB:A")) then
					MessageBoard.Administrator=true; MessageBoardXML.Administrator=true;
				else
					MessageBoard.Administrator=nil; MessageBoardXML.Administrator=nil;
				end
				MessageBoard.HaveRoster=true;
				break;
			end
			seq=seq+1;
			name,_,_,_,_,_,pnote,onote=GetGuildRosterInfo(seq);
		end
	end
	if (event=="DMEVENT_WIDGET_LEAVE") then
		GameTooltip:Hide()
	end
	if (event=="DMEVENT_WIDGET_ENTER") then
		local event,canvas,tip,context,param = ...
		GameTooltip:SetOwner(UIParent,"ANCHOR_CURSOR") -- Set the header to the entry-text. This also clears all previous text in it (I think)
		GameTooltip:SetText(context.topic)
		GameTooltip:AddLine(context.author)  -- Set some text
		local term=context.term              -- You add this from somewhere
		local text=context.text              -- You add this from somewhere
		local txts,txte=text:lower():find(term:lower(),1,true)          -- Find the searched term in the text
		if (not txts) then txts=0; txte=2; end
		local showoffstart,showoffend="(...) "," (...)"                 -- Just some show-off in case we need to trunc
		txts=math.floor((txte+txts)/2)-50                               -- Go 50 characters to the left from term center
		if (txts<1) then txts=1; showoffstart=""; end                   -- Clip to beginning
		txte=txts+100; if (text:len()<=txte) then showoffend=""; end    -- Clip to end
		text=text:sub(txts,txte)                                        -- Copy 100 characters
		txts,txte=text:lower():find(term:lower(),1,true)                -- Re-find the term in the new string
		if (txts) then
			text=text:sub(1,txts-1).."|cffffff00"..text:sub(txts,txte).."|r"..text:sub(txte+1)    -- Add the yellow color
		end
		GameTooltip:AddLine(showoffstart..text..showoffend,1,1,1,1)     -- Add text to tooltip
		GameTooltip:SetWidth(425)
		GameTooltip:Show()
	end
	if (event=="DMEVENT_HYPERLINK_ENTER") then
		local _,linkdata,link,button=...;
		if (linkdata:find("http:",1,true)==1) then
			GameTooltip:SetOwner(UIParent,"ANCHOR_CURSOR");
			GameTooltip:AddLine("Click to Copy");
			GameTooltip:Show();
		else
			local ttlink=linkdata..link;
			GameTooltip:SetOwner(UIParent,"ANCHOR_CURSOR");
			GameTooltip:SetHyperlink(ttlink);
			GameTooltip:Show();
		end
		--return
	end
	if (event=="DMEVENT_HYPERLINK_LEAVE") then
		GameTooltip:Hide();
		--return
	end
	if (event=="DMEVENT_HYPERLINK_CLICKED") then
		local _,linkdata,link,button=...;
		MBDialog:Register("MessageBoardUrlCopyDialog", {
		  text = L["POPUP_URL_COPY"],
		  buttons = {
			{
			  text = "Close",
			},
		  },
		  editboxes = {
			{
			auto_focus = true,
			width=200,
			},
		  },
		  on_show = function(self, data)
			self.text:SetFormattedText("\n"..L["POPUP_URL_COPY"].."\n")
			self.editboxes[1]:SetText(linkdata)
			self.editboxes[1]:HighlightText()
		  end,
		  hide_on_escape = true,
		  show_while_dead = true,
		})
		if (linkdata:find("http:",1,true)==1) then
			linkdata=string.sub(linkdata,6); -- removes the temporary http: in front
			MBDialog:Spawn("MessageBoardUrlCopyDialog")
		else MessageBoardXML:OnHyperlinkShow(frame,linkdata,link,button); end
		--return
	end
	if (event=="DMEVENT_BUTTON_CLICKED") then
		local canvas,context,params=...;
		if (not context or not params) then return; end    -- Data not provided, so something is wrong somewhere
		local base=MessageBoard:ValidateBoardDB(); if (not base) then return; end
		if (context.pollid and params.action=="submit") then
			local form=canvas:GetForm("pollid",context.pollid);
			for _,t in pairs(form) do
				if (t.param.id and t.data) then
					-- you've just found a checked box, so save it
					local buf=unpacktable(base[tonumber(t.context.pollc)][t.context.pollt][t.context.pollm]);    -- get the message
					local name=UnitName("player");
					name = Ambiguate(name,"none")
					if (type(buf.Voted)~="table") then buf.Voted={}; end    -- create voting table
					buf.Voted[name]=tonumber(t.param.id);	-- add myself as voter with the vote
					base[tonumber(t.context.pollc)][t.context.pollt][t.context.pollm]=packtable(buf);    -- put message back and sync db
					MessageBoard:ViewThread(tonumber(t.context.pollc),t.context.pollt);
					break;    -- only one option allowed selected by this method
				end
			end
			return
		end
		if (context.category) then context.category=tonumber(context.category); end
		if (params.id=="topic") then
			MessageBoard:ViewThread(tonumber(params.category),params.topic);
		elseif params.id=="deletetopic" then
			base[tonumber(params.category)][params.topic]=nil;    -- make the topic nil
			base[tonumber(params.category)]:DM_SetRedundancy(params.topic,true);    -- it's now redundant
            MessageBoardXML:Update();
            MessageBoardXML:ViewTopics();        -- view remaining topics
		elseif (params.id=="edit") then
			local buf=unpacktable(base[context.category][context.topic][context.msg])
			buf.Text=DuckLib.Code:Unprotect(buf.Text,MessageBoard_LS.Passwords["CategoryLevel"..context.category]);
			MessageBoardXML.Editor:Init(MessageBoardXML.Editor.ID_EDITMESSAGE,context.category,context.topic,buf,base[context.category][context.topic].State,context.r,context.g,context.b);
			MessageBoard_EditorTopicName:Hide();MessageBoard_Editor_Quote:Hide();MessageBoard_EditorTopicNameNonEdit:Show();MessageBoard_EditorTopicNameNonEdit:SetText(ren:toASCII(context.topic));
			MessageBoard_PollEditor:Hide();
		elseif (params.id=="reply") then
			local buf=unpacktable(base[context.category][context.topic][context.msg])
			local writepermission=MessageBoard:GetCategoryWritePermissions(CategoryLevel);
			if writepermission then
				buf.Text=DuckLib.Code:Unprotect(buf.Text,MessageBoard_LS.Passwords["CategoryLevel"..CategoryLevel]);
				buf.Text="\n\n\n|Cff888888=== "..L["EDITOR_MESSAGE_QUOTE"].." "..buf.Author.." ===\n\n"..buf.Text.."|r";
				MessageBoardXML.Editor:Init(MessageBoardXML.Editor.ID_NEWMESSAGE,CategoryLevel,context.topic,buf); --color was |cFFB8860B
				MessageBoard_Editor_Quote:Show();MessageBoard_Editor_Announce:Hide(); MessageBoard_Editor_Sticky:Hide(); MessageBoard_Editor_Locked:Hide();
				MessageBoard_EditorTopicName:Hide();MessageBoard_EditorTopicNameNonEdit:Show(); MessageBoard_EditorTopicNameNonEdit:SetText(ren:toASCII(context.topic));
			else DuckLibMB:Chat("MessageBoard: |cFFFFFFFF"..L["NO_REPLY_ALLOWED"],0,0.9,0.9);
			end
		elseif (params.id=="delete") then
            base[CategoryLevel][context.topic][context.msg]=nil;
			local dbcount=0;
			for msg,mTable in pairs(base[CategoryLevel][context.topic]) do
				if (type(mTable)=="string") then
					dbcount=dbcount+1;
					if (dbcount>1) then break; end
				end
			end
			if dbcount < 1 then
                base[CategoryLevel][context.topic]=nil;     -- make the topic nil
                base[CategoryLevel]:DM_SetRedundancy(context.topic,true);    -- it's now redundant
                MessageBoardXML:Update();
                MessageBoardXML:ViewTopics();        -- view remaining topics
            else
                base[CategoryLevel][context.topic]:DM_SetRedundancy(context.msg,true);    -- it's now redundant
                MessageBoard_Main_MessageScroller.SkipScroller=MessageBoard_Main_MessageScrollerScrollBar:GetValue();
				MessageBoard:ViewThread(CategoryLevel,context.topic); 
            end
        end
		--return
	end
	if (event=="DMEVENT_DATABASE_SYNC") then
		local base,_,_=MessageBoard:ValidateBoardDB();
		local handle,key,newdata=...;
		local tree=DuckLib.Database.Util:Name(handle);
		MessageBoardXML:Debug("key: "..tostring(key)) -- auk_xxxxxx_auk or State
		MessageBoardXML:Debug("handle[key]: "..tostring(handle[key])) -- text, time, and author / or state number
		MessageBoardXML:Debug("type handle[key]: "..tostring(type(handle[key]))) -- text, time, and author / or state number
		MessageBoardXML:Debug("tree1: "..tostring(tree[1])) -- MBPR1-9
		MessageBoardXML:Debug("tree2: "..tostring(tree[2])) -- topic name
		MessageBoardXML:Debug("newdata: "..tostring(newdata)) -- newdata seems to be true or nil
		if (tree[1] == "MBPR1" or "MBPR2" or "MBPR3" or "MBPR4" or "MBPR5" or "MBPR6" or "MBPR7" or "MBPR8" or "MBPR9") and (newdata == true) then -- if sync event comes from message base 
			MessageBoardXML:Debug("---Post Received---");
			local cat = string.sub (tree[1],5,5) -- new 9-3-13
			if (MessageBoard:GetCategoryReadPermissions(cat) and tree[2] and type(handle[key])=="string" and key ~= "State") then
				if (handle[key]) then
					local newcat = MessageBoardXML:Setting("Category"..cat.."Value") -- new 9-3-14
						if newcat then 
						if MessageBoard_LS.NewPostNotify==true then -- if new post notify option is on
							DuckLibMB:Chat("MessageBoard: |cFFFFFFFF"..L["NEW_POST_IN"]..ren:toASCII(newcat),0,0.9,0.9);   -- print notification to chat
						end
						if MessageBoard_LS.NewPostNotifyPopup==true then -- if new post notify option is on
							MBFunc.SpawnToast("MessageBoard","|cFFFFFFFF"..L["NEW_POST_IN"]..ren:toASCII(newcat)..".|r", 1,1,1);  -- print notification to popup
						end
					end
				else
					MessageBoardXML:Debug("---Deleted Message Received---");
				end
				-- MessageBoard:SetReadFlag(tree[3],tree[4],key,not handle[key]);
				MessageBoard:SetReadFlag(cat,tree[2],key,not handle[key]); -- new 9-3-14
				if MessageBoard_Main_TopicScroller:IsVisible() then MessageBoardXML:ViewTopics(); end
				MessageBoardXML:Update();
			end
		elseif (key=="Settings") then
			MessageBoardXML:Debug("---Settings Received---");
			MessageBoard:UpdatePermissions()
			MessageBoardXML:CheckCategories();
			--if (key=="MessageLimit") then
			--	base:DM_Set("Limit",handle[key]);
			--end
		elseif (tree[2]=="Avatar") then MessageBoardXML:Debug("---Avatar List Received---");
			MessageBoardXML:Update();
		end
		MessageBoardXML:Update();
	end
end

function MessageBoard:UpdateRoster() -- Update guild roster
	MessageBoardXML:Debug("Updating Guild Roster");
	if (not IsInGuild()) then return; end
	GuildRoster();
end

function MessageBoard.Slasher(msg) -- There's slashing to be done
	if (not msg) then msg=""; end   -- Validate input
	local omsg=msg;
	if (strlen(msg)>0) then msg=strlower(msg); end
	if (msg=="ver" or msg=="version") then
		MessageBoardXML:CheckV()
		return;
	end
	if (msg=="debug") then
		if (MessageBoard_LS.Debug==true) then MessageBoard_LS.Debug=false; DuckLib:Chat("Debug OFF");
		else MessageBoard_LS.Debug=true; DuckLib:Chat("Debug ON"); end
		return;
	end
	if (msg=="?" or msg=="help") then
		DuckLibMB:Chat("MessageBoard: |cFFFFFFFF"..L["SLASH_HELP_1"],0,0.9,0.9)
		DuckLibMB:Chat("MessageBoard: |cFFFFFFFF"..L["SLASH_HELP_2"],0,0.9,0.9)
		DuckLibMB:Chat("MessageBoard: |cFFFFFFFF"..L["SLASH_HELP_3"],0,0.9,0.9)
		DuckLibMB:Chat("MessageBoard: |cFFFFFFFF"..L["SLASH_HELP_4"],0,0.9,0.9)
		DuckLibMB:Chat("MessageBoard: |cFFFFFFFF"..L["SLASH_HELP_5"],0,0.9,0.9)
		return;
	end
	MessageBoard:ToggleMainFrame();
end

-- Generic mouse-wheel handler for scrolling-panes.
function MessageBoardXML:MouseWheel(slider,delta)
 if (not slider) then slider=self; end
 local min,max=slider:GetMinMaxValues(); if (not min or not max) then return; end
 local step=slider:GetValueStep(); if (not step) then return; end
 local value=slider:GetValue(); if (not value) then return; end
 if (delta==1 and (value-step)>=min) then slider:SetValue(value-step);
 elseif (delta==-1 and (value+step)<=max) then slider:SetValue(value+step); end
end

function MessageBoard_Main_OnShow() -- When the main window is shown
	if (not IsInGuild()) then MessageBoard_WaitingText:SetText(L["JOIN_A_GUILD"]); MessageBoard:DisableBoard(); return; end
	if (not MessageBoard.HaveRoster) then MessageBoard:UpdateRoster(); end
	MessageBoard:UpdatePermissions();
	MessageBoardXML:CheckCategories();
	MessageBoardXML:Update();
end

function MessageBoard_Main_OnHide() -- When the main window is hidden
	MessageBoard_OFrame:Hide();
end

-- MessageBoard heartbeat
-- o The naming-convention is because the use it not as an OnUpdate
--   handler. It's a slow beat going on MB_HEARTBEAT speed, and written
--   so that it will get slowed down by hiccups like changing zone and
--   heavy loading from disk when entering a highly populated zone.
function MessageBoard_HeartBeat(self,elapsed)
	MessageBoard.Timers.Last=MessageBoard.Timers.Last+elapsed;      -- Update this
	if (MessageBoard.Timers.Last<MB_HEARTBEAT) then return; end;    -- Heartbeat is 1/3 sec
	MessageBoard.Timers.Last=0;      								-- Restart
	if (MessageBoard.Ready~=true and MessageBoard.LoggedIn and MessageBoard.HaveRoster) then    -- "LoggedIn" must guarrantee data has been loaded
		MessageBoard:UpdatePermissions();
		MessageBoardXML:Update();
		MessageBoard.Ready=true;
		DuckLibMB:Chat("MessageBoard: |cFFFFFFFF"..L["CONNECTED_TO_THE"]..MessageBoard:GetGuildName()..L["GUILD_FORUM"],0,0.9,0.9);
		MBFunc.SpawnToast("MessageBoard","|cFFFFFFFF"..L["CONNECTED_TO_THE"]..MessageBoard:GetGuildName()..L["GUILD_FORUM"], 1,1,1);
		MessageBoard_MainTitle:SetText(MessageBoard:GetGuildName()..L["GUILD_FORUM"]);
		MessageBoard_Back:SetText(L["MAIN_BACK_BUTTON"])
		--MessageBoard_WaitingText:SetText(L["MB_WAITING_SYNC"])
	end
end

-- Retrieves the total number of messages by user. Don't count deleted messages
-- Author provided returns that user's message count.
function MessageBoardXML:UserMessageCount(author)
	local base=MessageBoard:ValidateBoardDB(); if (not base) then return nil; end
	local count=0;
	for category,cTable in pairs(base) do
		for topic,tTable in pairs(cTable) do
			if (type(tTable)=="database") then
				for msg,mTable in pairs(tTable) do
					if (type(mTable)=="string") then
						local buf=unpacktable(base[category][topic][msg]);
						if (not author or (author and buf.Author==author)) then
							count=count+1;
						end
					end
				end
			end
		end
	end
	return count;
end

function MessageBoard:CleanUp()
	MessageBoardXML:Debug("Cleaning Up...",1);
	local base=MessageBoard:ValidateBoardDB(); if (not base) then return nil; end
	-- Kill empty topics
	for category,cTable in pairs(base) do
		for topic,tTable in pairs(cTable) do
			if (type(tTable)=="database") then
				for msg,mTable in pairs(tTable) do
					if (type(mTable)=="database") then
						break;
					else
						cTable[topic]=nil;
					end
				end
			end
		end
	end
end

-- Only runs when the main window is visible
function MessageBoard_Main_OnUpdate()
	if (DuckLib.RenderWnCounter and DuckLib.RenderWnCounter>=4) then
        if (MessageBoard.PositionViewport) then -- search scroller viewport update
            MessageBoard_Main_MessageScroller_Browser_Frame:GoTo(MessageBoard.PositionViewport);
            MessageBoard.PositionViewport=nil;
        elseif (MessageBoard_Main_MessageScroller.SkipScroller) then
            MessageBoard_Main_MessageScrollerScrollBar:SetValue(MessageBoard_Main_MessageScroller.SkipScroller);
            MessageBoard_Main_MessageScroller.SkipScroller=nil;
		end
    end
end

-- Item-link support
function MessageBoardXML:OnHyperlinkShow(frame,link,text,button)
	ChatFrame_OnHyperlinkShow(frame,link,text,button);
end

-- Return the state of the provided topic
-- o Returns State,Icon,Unread
function MessageBoard:TopicState(category,topic)
	local base=MessageBoardXML:ValidateBoardDB();
	local newmsg=false;
	if (not base[category][topic]) then return 0,"",nil; end
	if type(base[category][topic])~="database" then -- something wrong, set nil, return nil, record debug.
		local dbDebug={};
		MessageBoard_DB.dbDebug=totable(base[category][topic]);
		base[category][topic]=nil;
		return 0,"",nil;
	end 
	local state=base[category][topic].State or 0;
	if (state < 0) then return 0,"",nil; end
	for msg,mTable in pairs(base[category][topic]) do
		if (type(mTable)=="string") then  -- if true, the message exists so no need to check for that
			if (not self:GetReadFlag(category,topic,msg)) then newmsg=true; break; end
		end
	end
	local nm=""; if (newmsg) then nm="_new"; end
	if state== 0 then return state,Icon["NORMAL"..nm],newmsg; end
	if state==10 then return state,Icon["STICKY"..nm],newmsg; end
	if state==20 then return state,Icon["ANNOUNCE"..nm],newmsg; end
	if state==30 then return state,Icon["LOCKED"..nm],newmsg; end
	if state==40 then return state,Icon["POLL"..nm],newmsg; end
	return 0,"",nil;
end

-- Some helpers to get past later updates to the WoW code
function MessageBoardXML.SetButtonText(button,text)
	_G[button:GetName().."Text"]:SetText(text);
end
function MessageBoardXML.GetButtonText(button,text)
	return _G[button:GetName().."Text"]:GetText(text);
end
function MessageBoardXML.SetButtonTextColor(button,r,g,b,a)
	_G[button:GetName().."Text"]:SetTextColor(r,g,b,a);
end

-- Update the main window
-- Use when anything visual has changed
function MessageBoardXML:Update()
	MessageBoardXML:Debug("Updating Visuals");
	MBFunc.MBGuildFrame_UpdateTabard();
	local base,avatar,settings=self:ValidateBoardDB(); if (not base or not avatar or not settings) then return; end
	MessageBoardXML:Debug(MessageBoardXML.MyRank);
	MessageBoardXML:UpdateDisplay()
	MessageBoard:CheckDatabase()
	MessageBoardXML:SetCategories()
	MBFunc.MBUpdateMinimap()
	local seq=1;
	local me=tostring(UnitName("player"))
	local name,_,_,_,_,_,pnote,onote=GetGuildRosterInfo(seq);
	while(name) do
		name = Ambiguate(name,"none")
		if (name==me) then
			if (onote and string.find(onote,"#MB:A")) or (MessageBoardXML:Setting("UseAlternateAdmin") and pnote and string.find(pnote,"#MB:A")) then
				MessageBoard.Administrator=true; MessageBoardXML.Administrator=true;
			else
				MessageBoard.Administrator=nil; MessageBoardXML.Administrator=nil;
			end
			MessageBoard.HaveRoster=true;
			break;
		end
		seq=seq+1;
		name,_,_,_,_,_,pnote,onote=GetGuildRosterInfo(seq);
	end
end

function MessageBoardXML:GetNewMsgCount()
	local base=self:ValidateBoardDB(); if (not base) then return nil; end
	local unread=nil;
	local unreadcount=0;
	for category,cTable in pairs(base) do
		if (type(cTable)~="database") then
			base[category]=nil;
		else
			for topic,tTable in pairs(cTable) do
				local readpermission=MessageBoard:GetCategoryReadPermissions(category);
				local _,_,unread=MessageBoard:TopicState(category,topic);
				if unread and readpermission then unreadcount=unreadcount+1; end
			end
		end
	end
	return unreadcount;
end

function MessageBoardXML:UpdateDisplay()
	-- noop, hooked by displays
end

-- Format TOD
function MessageBoard:FormatTime(native)
	if (not native) then return "<unknown time>"; end
	return date(MessageBoardLANG.Widget.NearDate,native);
end

function MessageBoard:GetCategoryReadPermissions(category)
	if (not category) then return false; end
	if MessageBoardXML.MyRank=="Administrator" then	return true;
	elseif MessageBoardXML.MyRank=="Officer" then return MessageBoardXML:Setting("Category"..category.."OfficerRead");
	elseif MessageBoardXML.MyRank=="Normal User" then return MessageBoardXML:Setting("Category"..category.."UserRead");
	elseif MessageBoardXML.MyRank=="Alt" then return MessageBoardXML:Setting("Category"..category.."UserRead");
	elseif MessageBoardXML.MyRank=="No Access" then
		return false;
	end
end

function MessageBoard:GetCategoryWritePermissions(category)
	if (not category) then return false; end
	if MessageBoardXML.MyRank=="Administrator" then	return true;
	elseif MessageBoardXML.MyRank=="Officer" then return MessageBoardXML:Setting("Category"..category.."OfficerWrite");
	elseif MessageBoardXML.MyRank=="Normal User" then return MessageBoardXML:Setting("Category"..category.."UserWrite");
	elseif MessageBoardXML.MyRank=="Alt" then return false;
	elseif MessageBoardXML.MyRank=="No Access" then	return false;
	end
end

function MessageBoardXML:MakeNewTopic()
	if (MessageBoard:GetCategoryWritePermissions(CategoryLevel)) then -- check ranks
		MessageBoardXML.Editor:Init(MessageBoardXML.Editor.ID_NEWTOPIC,CategoryLevel,"",{});
		MessageBoard_Editor_Quote:Hide();MessageBoard_EditorTopicNameNonEdit:Hide();
		MessageBoardXML.Editor:ResetPoll();
	else DuckLibMB:Chat("MessageBoard: |cFFFFFFFF"..L["NOT_ALLOWED_POST"],0,0.9,0.9);
	end
end

function MessageBoardXML:MakeNewPoll()
	if (MessageBoard:GetCategoryWritePermissions(CategoryLevel)) then -- check ranks
		MessageBoardXML.Editor:Init(MessageBoardXML.Editor.ID_NEWPOLL,CategoryLevel,"",{});
		MessageBoard_Editor_Quote:Hide();MessageBoard_EditorTopicNameNonEdit:Hide();
		MessageBoardXML.Editor:ResetPoll();
	else DuckLibMB:Chat("MessageBoard: |cFFFFFFFF"..L["NOT_ALLOWED_POST"],0,0.9,0.9);
	end
end

function MessageBoardXML:Reply() -- new reply with no quotes
	local base=MessageBoardXML:ValidateBoardDB(); if (not base) then return; end
	if (not MessageBoard.Selected.Topic) then return; end
	if (base[CategoryLevel][MessageBoard.Selected.Topic].State==30) then -- if locked
		DuckLibMB:Chat("MessageBoard: |cFFFFFFFF"..L["TOPIC_IS_LOCKED"],0,0.9,0.9);
		return;
	end
	if (MessageBoard:GetCategoryWritePermissions(CategoryLevel)) then -- check ranks
		--buf.Text=DuckLib.Code:Unprotect(base[CategoryLevel][MessageBoard.Selected.Topic].Text,MessageBoard_LS.Passwords["CategoryLevel"..CategoryLevel]);
		MessageBoardXML.Editor:Init(MessageBoardXML.Editor.ID_NEWMESSAGE,CategoryLevel,MessageBoard.Selected.Topic,{});
		MessageBoard_Editor_Quote:Hide();MessageBoard_Editor_Announce:Hide();MessageBoard_Editor_Sticky:Hide();
		MessageBoard_Editor_Locked:Hide();MessageBoard_EditorTopicName:Hide();MessageBoard_EditorTopicNameNonEdit:Show();
		MessageBoard_EditorTopicNameNonEdit:SetText(ren:toASCII(MessageBoard.Selected.Topic));
	else DuckLibMB:Chat("MessageBoard: |cFFFFFFFF"..L["NO_REPLY_ALLOWED"],0,0.9,0.9);
	end
end

-- Entry-function when editor is done
-- o The text will be saved accordingly
function MessageBoardXML:EditedText(id,category,topic,data,state,colorr,colorg,colorb)
	if not MBFunc.isBlank(data.Text) then data.Text=DuckLib.Code:Protect(data.Text,MessageBoard_LS.Passwords["CategoryLevel"..CategoryLevel]); end
	local now=time();
	local base,_,_=MessageBoardXML:ValidateBoardDB();
	if (id==MessageBoardXML.Editor.ID_NEWTOPIC or id==MessageBoardXML.Editor.ID_NEWPOLL) then
		if (not base[category]) then MessageBoardXML:Debug("Category \""..category.."\" was not found",1); return; end
		if (base[category][topic]) then
			DuckLibMB:Chat("MessageBoard: |cFFFFFFFF"..L["TOPIC_IS_EXIST"],0,0.9,0.9);
			return;   -- can't allow same-named topics in one category
		end
		base[category][topic]={ State=state };
	end
	if (id==MessageBoardXML.Editor.ID_NEWTOPIC or id==MessageBoardXML.Editor.ID_NEWMESSAGE or id==MessageBoardXML.Editor.ID_NEWPOLL) then
		MessageBoardXML:Debug("Called EditedText NEWTOPIC NEWMESSAGE OR NEWPOLL")
		if (not base[category]) then MessageBoardXML:Debug("Category \""..category.."\" was not found",1); return; end
		if (not base[category][topic]) then MessageBoardXML:Debug("Topic \""..topic.."\" was not found",1); return; end
		local msg=base[category][topic]:DM_UniqueKey();
		local name=UnitName("player");
		name = Ambiguate(name,"none")
		data.Time=now;
		data.Author=name;
		data.colorR=colorr;
		data.colorG=colorg;
		data.colorB=colorb;
		base[category][topic][msg] = packtable(data);
		MessageBoard:SetReadFlag(category,topic,msg,true); -- set read flag to true for a message you just wrote
		local buf=unpacktable(base[category][topic][msg]);
	elseif (id==MessageBoardXML.Editor.ID_EDITMESSAGE) then
		MessageBoardXML:Debug("Called EditedText EDITMESSAGE")
		if (not base[category]) then MessageBoardXML:Debug("Category \""..category.."\" was not found",1); return; end
		if (not base[category][topic]) then MessageBoardXML:Debug("Topic \""..topic.."\" was not found",1); return; end
		base[category][topic].State=state;
		local tempmsg;
		local msg;
		for msg,mTable in pairs_dn(base[category][topic]) do
        if (type(mTable)=="string") then
			local unpmsg=unpacktable(mTable);
            if (unpmsg.Time==data.Time) then tempmsg=msg; break; end
			end
		end
		if (not tempmsg) then MessageBoardXML:Debug("Original message not found",1); return; end
		base[category][topic][tempmsg]=packtable(data);
		MessageBoard:SetReadFlag(category,topic,tempmsg,true);
	end
	MessageBoardXML:Update();
end

function MessageBoardXML:Button_OnEnter()
	if (not self.Help) then return; end
	GameTooltip:SetOwner(self,"ANCHOR_RIGHT");
	GameTooltip:SetText(MessageBoardLANG.Help.Usage);
	GameTooltip:AddLine(self.Help,1,1,1,true);
	GameTooltip:Show();
end

function MessageBoard:GetGuildName()
	MessageBoard.GuildName=GetGuildInfo("player");
	if (not MessageBoard.GuildName) then MessageBoard.GuildName="<UnGuilded>"; end
	return MessageBoard.GuildName;
end

-- Check and create board-base only
function MessageBoard:ValidateBoard(namesonly)
	if (not MessageBoard_DB) then return nil; end						-- can't work without that
	local gn=MessageBoard:GetGuildName(); if (not gn or gn=="<UnGuilded>") then return nil; end	-- no guild, no share
	--local realm=GetRealmName(); if (not realm) then return nil; end		-- no realm, no world
	--local dbn=realm.." - "..gn;											-- NOTE: swapped realm and guild to not kill old stuff
	local dbn=gn;														-- NOTE: swapped realm and guild to not kill old stuff
	if (namesonly) then return dbn; end									-- don't remember what this was for, but I kept it in there
	if (MessageBoard_DB[dbn] and MessageBoard_DB[dbn].DM_Registered) then return MessageBoard_DB[dbn]; end
	if (not MessageBoard_DB[dbn]) then									-- database does not exists
	MessageBoard_DB[dbn]=todatabase({                          		    -- so create it
            Avatar={}, Settings=packtable({}),						    -- compatibility with previous functionality
            Base={},                                                    -- and the category base as usual
        });
		for i=1,9 do													-- do 9
			MessageBoard_DB[dbn..i]={};									-- create this category
			DB:Register("MBPR"..i,MessageBoard_DB[dbn..i],MessageBoard_Main);	-- register it as a database
		end
		DB:Register("MBPR",MessageBoard_DB[dbn],MessageBoard_Main);		-- register root database
		for i=1,9 do MessageBoard_DB[dbn].Base:DM_SetLink(i,MessageBoard_DB[dbn..i]); end	-- link all 9 categories
	else
		for i=1,9 do DB:Register("MBPR"..i,MessageBoard_DB[dbn..i],MessageBoard_Main); end	-- register categories' databases
		DB:Register("MBPR",MessageBoard_DB[dbn],MessageBoard_Main);		-- register root database
	end
	return MessageBoard_DB[dbn];										-- upsies
end

-- Check and create message base
function MessageBoardXML:ValidateBoardDB()
	local board=MessageBoard:ValidateBoard(); if (not board) then return nil; end
	return board.Base,board.Avatar,board.Settings
end

function MessageBoard:ValidateBoardDB()
	return MessageBoardXML:ValidateBoardDB();
end

function MessageBoardXML:ValidateBoard(namesonly)
	return MessageBoard:ValidateBoard(namesonly);
end

function MessageBoard:ToggleMainFrame()
	if (MessageBoard_Main:IsVisible()) then MessageBoard_Main:Hide();
	else MessageBoard_Main:Show(); end
	DB.EventFrame=MessageBoard_Main;
end

-- MessageBoard Debug
function MessageBoardXML:Debug(text,r,g,b)
	if (not MessageBoard_LS.Debug) then return; end
	DuckLib:Chat("MBDEBUG: "..tostring(text),r,g,b);
end

local LibStub = _G.LibStub;
local Toast = LibStub("LibToast-1.0");

-- enter password for categories  --
function MessageBoardXML:SetPassword(category)
	MBDialog:Register("MESSAGEBOARD_SET_PASSWORD", {
	  text = L["POPUP_ENTER_PASSWORD"],
	  buttons = {
		{
		  text = _G.ACCEPT,
		  on_click = function(self, cat)
			if (not MessageBoard_LS.Passwords) then MessageBoard_LS.Passwords={}; end
				if (cat>=1 and cat<=9) then
					MessageBoard_LS.Passwords["CategoryLevel"..cat]=self.editboxes[1]:GetText()
					if (MessageBoard_LS.Passwords["CategoryLevel"..cat]=="") then
						MessageBoard_LS.Passwords["CategoryLevel"..cat]=nil;
						DuckLibMB:Chat("MessageBoard: |cFFFFFFFF"..L["CAT_PASS_FOR"]..cat..L["CAT_PASS_CLEARED"],0,0.9,0.9)
					else
						DuckLibMB:Chat("MessageBoard: |cFFFFFFFF"..L["CAT_PASS_FOR"]..cat..L["CAT_PASS_SETTO"]..MessageBoard_LS.Passwords["CategoryLevel"..cat]..".",0,0.9,0.9)
					end
				end
		  end,
		},
		{
		  text = _G.CANCEL,
		},
	  },
	  editboxes = {
		{
		  auto_focus = true,
		},
	  },
	  on_show = function(self,data)
		self.text:SetFormattedText(L["POPUP_ENTER_PASSWORD"])
		self.editboxes[1]:SetText("")
	  end,
	  hide_on_escape = true,
	  show_while_dead = true,
	})	
	local dialog=MBDialog:Spawn("MESSAGEBOARD_SET_PASSWORD")
	if (dialog) then dialog.data=category; end
end

-- Update User Permissions
function MessageBoard:UpdatePermissions()
	MessageBoardXML:Debug("Updating Permissions");
	local guildName, guildRankName, guildRankIndex = GetGuildInfo("player");
	local myrank = guildRankIndex + 1;
	if MessageBoard.Administrator then MessageBoardXML.MyRank="Administrator";
	elseif (MessageBoardXML:Setting("OfficerRank"..myrank) == true or IsGuildLeader()) then	MessageBoardXML.MyRank="Officer";
	elseif MessageBoardXML:Setting("AltRank"..myrank) == true then MessageBoardXML.MyRank="Alt";
	elseif MessageBoardXML:Setting("NoPostRank"..myrank) == true then MessageBoardXML.MyRank="No Access";
	elseif MessageBoardXML:Setting("NormalRank"..myrank) == true then MessageBoardXML.MyRank="Normal User";
	end
end

function MessageBoardXML:CheckCategories()
	MessageBoardXML:Debug("Checking Categories");
	local i
	local count=0
	for i=1,9 do
		if (MessageBoardXML:Setting("Category"..i.."Value")==nil or MBFunc.isBlank(MessageBoardXML:Setting("Category"..i.."Value"))) then
			_G["MessageBoard_MainNavBarButton"..i]:Hide();
			count=count+1
		else
			_G["MessageBoard_MainNavBarButton"..i]:Show();
		end
	end
	if count==9 then MessageBoard:DisableBoard(); return end
	if MessageBoard.Administrator then
		MessageBoardXML:SetCategories();
		MessageBoard:EnableBoard();
	else
		MessageBoardXML:SetCategories();
		if MessageBoard_Waiting:IsShown() then
			MessageBoard:EnableBoard();
		end
	end
end

function MessageBoard:DisableBoard() -- run when settings are not yet received/setup
	MessageBoardXML:Debug("No settings, not Administrator, or not in a guild, disabling forum");
	MessageBoard_MainNavBar:Hide();
	MessageBoard_MainCategory:Hide();
	--MessageBoard_MainStats:Hide();
	MessageBoard_Main_SearchScroller:Hide();
	MessageBoard_CurrentTopic:Hide();
	MessageBoard_Main_TopicScroller:Hide();
	MessageBoard_Main_SearchScroller:Hide();
	MessageBoard_Main_MessageScroller:Hide();
	MessageBoard_Main_MessageScroller_Browser_Frame:Hide();
	MessageBoard_Main_LeftPanel:Hide();
	MessageBoard_Main_RightPanel:Hide();
	MessageBoard_Main_LeftDisplay:Hide();
	MessageBoard_Main_RightDisplay:Hide();
	if MessageBoard.Administrator then
		MessageBoard_Waiting:Show();
		MessageBoard_OpenSettings:Show();
	else
		MessageBoard_OpenSettings:Hide();
		MessageBoard_Waiting:Show();
	end
end

function MessageBoard:EnableBoard()
	if MessageBoard.FirstRun then
		MessageBoardXML:Debug("-- First Run --")
		MessageBoard_OpenSettings:Show();
		MessageBoard_MainCategory:Show();
		MessageBoard_Waiting:Hide();
		--MessageBoard_MainStats:Show();
		MessageBoard_MainNavBar:Show();
		MessageBoard_Main_TopicScroller:Show();
		MessageBoard_Main_SearchScroller:Hide();
		MessageBoard_Main_MessageScroller:Hide();
		MessageBoard_Main_MessageScroller_Browser_Frame:Hide();
		MessageBoard_CurrentTopic:Hide();
		MessageBoard_Main_LeftPanel:Show();
		MessageBoard_Main_RightPanel:Show();
		MessageBoard_Main_LeftDisplay:Show();
		MessageBoard_Main_RightDisplay:Show();
		-- set initial category based on first accessible
		local i=1;
		while i<10 do
			local readpermission=MessageBoard:GetCategoryReadPermissions(i);
			if readpermission then
				MessageBoardXML:SetCategoryLevel(i);
				_G["MessageBoard_MainNavBarButton"..i]:Click(); break;
			end
			i=i+1;
		end
		if (readpermission) then MessageBoardXML:ViewTopics(); end
		MessageBoard.FirstRun=false;
	end
	if MessageBoard.Selected.Topic then MessageBoard:ViewThread(CategoryLevel,MessageBoard.Selected.Topic); end
end

function MessageBoard_Search()
	MessageBoard_Main_TopicScroller:Hide();MessageBoard_Main_MessageScroller:Hide();MessageBoard_Main_MessageScroller_Browser_Frame:Hide()
	MessageBoard_Main_SearchScroller:Hide();MessageBoard_CurrentTopic:Hide();
	local text=MessageBoard_MainNavBarSearchBox:GetText() or "";
	text=text:lower();
	local base,avatar=MessageBoard:ValidateBoardDB(); if (not base) then return nil; end
	local count=0;
	local page=[[<wn><meta><page background="#00000000"/><font color="#FFFFFFFF" size="14"/><font name="Blue" color="#f00f"/><font name="Green" color="#f0f0"/><font name="Green24" inherits="Green" size="24"/><elements><button inheritSys="MB_Topic_Buttons"/></elements></meta><body>]];
	for category,cTable in pairs(base) do -- cycle through the categories
		for topic,tTable in pairs(cTable) do -- cycle through the topics
			if (type(tTable)=="database") then
				for msg,mTable in pairs(tTable) do
					if (type(mTable)=="string") then
						local buf=unpacktable(base[category][topic][msg]);
						local readpermission=MessageBoard:GetCategoryReadPermissions(category);
						if (type(buf.Text)=="string" and type(buf.Author)=="string" and buf.Text and readpermission) then -- dont search deleted messages --
							local addit=nil;
							if (buf.Text:lower():find(text,1,true)) then     -- Check article text
								MessageBoardXML:Debug("Found article in: "..topic,0,1,1); -- name of the topic with the text debug
								addit=true;
							elseif (buf.Author:lower():find(text,1,true)) then      -- Check article author
								MessageBoardXML:Debug("Found author: "..buf.Author,0,1,1); -- name of the topic with the text debug
								addit=true;
							end
							if (addit) then
								local pic=nil;
								if (avatar[buf.Author]) then pic=avatar[buf.Author]; else pic="Interface\\Icons\\INV_Misc_QuestionMark"; end
								page=page..[[<img align="left" pad="0" height="20" width="20">]]..pic..[[</img><button height="30" text="]]..buf.Text..[[" term="]]..text..[[" tip="" author="]].. buf.Author..[[" category="]]..category..[[" topic="]]..topic..[[" id="topic" pad="0" width="85%" glue="left">]]..topic..[[</button>]];
								count=count+1;
							end
						end
					end
				end
			end
		end
	end
	page=page..[[</body></wn>]];
	MessageBoard_Main_SearchScrollerNumResults:SetText(L["NUMBER_SEARCH_RESULTS"]..count);
	MessageBoard_Main_MessageScroller:Show();MessageBoard_Main_MessageScroller_Browser_Frame:Show();
    DuckLib.Render:Render(MessageBoard_Main_MessageScroller_Browser_Frame,page);
end

function MessageBoardXML:ViewTopics()
	if (not MessageBoard:GetCategoryReadPermissions(CategoryLevel)) then return; end
    local base=MessageBoard:ValidateBoardDB(); if (not base) then return nil; end
    CategoryLevel=CategoryLevel or 1;    -- set to 1 if it is nil
    MessageBoard_MainCategory:SetText(MessageBoardXML:Setting("Category"..CategoryLevel.."Value") or "");
	--MessageBoard_MainLowerFrame_LeftDisplayCategory:SetText(MessageBoardXML:Setting("Category"..CategoryLevel.."Value") or "");
    MessageBoard_Main_SearchScroller:Hide();
    MessageBoard_Main_MessageScroller:Hide();MessageBoard_Main_MessageScroller_Browser_Frame:Hide();MessageBoard_CurrentTopic:Hide();
    local page=[[<wn><meta><page background="#00000000"/><font color="#FFFFFFFF" size="14"/><font name="Blue" color="#f00f"/><font name="Green" color="#f0f0"/><font name="Green24" inherits="Green" size="24"/><elements><button inheritSys="MB_Topic_Buttons"/></elements></meta><body>]];
	-- this routine assumes "state" to be the same as "tTable.State"
	-- if not, exchange them
	local _state,_unread=20,true;										-- first selection in the generated list
	repeat
		-- cycle all thingies
		for topic,tTable in pairs_up(base[CategoryLevel]) do repeat		-- cycle with an extra inner loop
		if (type(tTable)~="database") then break; end				-- break inner repeat-until loop
		local state,icon,unread=MessageBoard:TopicState(CategoryLevel,topic);	-- get what-not
		if state==_state and unread==_unread then					-- are we at where we are in the loop?
			page=page..[[<img align="left" pad="0" height="20" width="20">]]..icon..[[</img><button height="30" category="]]..CategoryLevel..[[" topic="]]..topic..[[" id="topic" pad="0" width="85%" glue="left">]]..topic..[[</button>]]
			if MessageBoard.Administrator then page=page..[[<button inheritSys="MessageBoard_PostButtonsTemplate" category="]]..CategoryLevel..[[" topic="]]..topic..[[" id="deletetopic" pad="10" width="50" glue="left">Delete</button>]]; end
			end
		until(true); end										-- close loops
		-- select next options
		if (not _unread) then 
			if _state==20 then _state=10;						-- third selection in list
				elseif _state==10 then _state=40;				-- fifth selection in list
				elseif _state==40 then _state=0;				-- seventh selection in list
				elseif _state==0 then _state=30;				-- ninth selection in list
				elseif _state==30 then _state=nil;				-- eleventh selection in list
			end
		end
		_unread=not _unread										-- up odd number selections to even
	until(not _state);
    page=page..[[</body></wn>]];
    MessageBoard_Main_MessageScroller:Show();MessageBoard_Main_MessageScroller_Browser_Frame:Show();
    DuckLib.Render:Render(MessageBoard_Main_MessageScroller_Browser_Frame,page);
end

function MessageBoard:ViewThread(category,topic)
	MessageBoard.Selected.Topic = topic
	MorePostsBelow:Hide(); MessageBoard_Main_TopicScroller:Hide();
	local base,avatar,_=self:ValidateBoardDB();
	local otopic=topic;
	if (topic:find("<",1,true)) then topic=topic:gsub("<"," "); end
	if (topic:find(">",1,true)) then topic=topic:gsub(">"," "); end
	if (topic:find('"',1,true)) then topic=topic:gsub('"'," "); end
	if (otopic~=topic) then
		base[category][topic]=base[category][otopic];
		base[category][otopic]=nil;
	end
	local page=[[<wn><meta><page background="#00000000"/><font size="14"/><font name="Blue" color="#f00f"/><font name="Green" color="#f0f0"/><font name="Green24" inherits="Green" size="24"/><elements><button inheritSys="MessageBoard_PostButtonsTemplate"/></elements></meta><body>]];
	page=page..[[<h2 glue="left" pad="10">|cFFC0C0C0]]..topic..[[|r</h2>]];
	local icon=Icon.NORMAL;
	if (not base[category][topic]) then MessageBoardXML:ViewTopics(); return; end -- topic has been deleted
	if (base[category][topic].State==10) then icon=Icon.STICKY;
	elseif (base[category][topic].State==20) then icon=Icon.ANNOUNCE;
	elseif (base[category][topic].State==30) then icon=Icon.LOCKED;
	elseif (base[category][topic].State==40) then icon=Icon.POLL; end	
	page=page..[[<img glue="left" pad="10" width="25" height="25">]]..icon..[[</img>]];
	page=page..[[<img align="center" pad="10" height="12" width="100%">Interface\FriendsFrame\UI-FriendsFrame-OnlineDivider</img>]];
	local msglist={};			-- local unpacked storage
	for msg,d in pairs(base[category][topic]) do			-- get all message keys and data
		if (type(d)=="string") then table.insert(msglist,{msg=msg,Time=unpacktable(d).Time}); end	-- insert key and time
	end
	table.sort(msglist,function(A,B)			-- let lua sort it
		if (type(A)~="table" or type(B)~="table") then return nil; end	-- only sort tables (messages)
		return (A.Time<B.Time);				-- the sorting parameters
	end );
	for _,mit in ipairs(msglist) do			-- read all sorted message indexes
		local msg,mTable=mit.msg,base[category][topic][mit.msg];	-- recreate "msg" and its data
		--if type(mTable)=="table" then	-- can remove this line and its corresponding "end"
    -- end new
			local buf=unpacktable(base[category][topic][msg]);
			local msgTxt=DuckLib.Code:Unprotect(buf.Text,MessageBoard_LS.Passwords["CategoryLevel"..CategoryLevel]); -- decrypt based on password
				if (msgTxt:match("(%a+://%S+)")) then -- grab the http links
				msgTxt=msgTxt:gsub("(%a+://%S+)","|cFF4682B4|Hhttp:%1|h[%1]|h|r"); end
				if (msgTxt:match("%f[%S](www%.[-%w_%%]+%.%S+)")) then -- grab the http links
				msgTxt=msgTxt:gsub("%f[%S](www%.[-%w_%%]+%.%S+)","|cFF4682B4|Hhttp:%1|h[%1]|h|r"); end
				page=page..[[<focus id="]]..topic..msg..[["/>]];
				if buf.colorR then
					page=page..[[<set r="]]..buf.colorR..[[" g="]]..buf.colorG..[[" b="]]..buf.colorB..[[" category="]]..category..[[" topic="]]..topic..[[" msg="]]..msg..[["/>]];
				else
					page=page..[[<set category="]]..category..[[" topic="]]..topic..[[" msg="]]..msg..[["/>]];
				end
				page=page..[[<img width="60" height="60" pad="10">]];
				page=page..(avatar[buf.Author] or [[Interface\Icons\INV_Misc_QuestionMark]]);
				page=page..[[</img><text size="12" pad="10" glue="left">|cFFC0C0C0by: ]];
				page=page..buf.Author.." - "..MessageBoard:FormatTime(buf.Time).."|r</text>";
				page=page..[[<text size="8">|cFFC0C0C0Posts: ]]..MessageBoardXML:UserMessageCount(buf.Author);
				page=page..[[|r</text>]];
				
				-- check for polls and display --
				if buf.Poll then
					-- check if previously answered or have permission to write
					local writepermission=MessageBoard:GetCategoryWritePermissions(CategoryLevel); -- check permission --
					local name=UnitName("player");
					name = Ambiguate(name,"none")
					if (not buf.Voted or not buf.Voted[name] and writepermission) then
						-- set context variables that indicates widgets that are members of the poll
						page=page..[[<set pollid="]]..tostring(category)..topic..msg..[[" pollc="]]..category..[[" pollt="]]..topic..[[" pollm="]]..msg..[["/>]];
						page=page..[[<h2 align="center" pad="10">|cFFC0C0C0]]..buf.Poll.Question..[[|r</h2>]];
						for polln,polla in ipairs(buf.Poll) do	-- do numeric indices (skipping the question)
							if ((polla or "")~="") then
								page=page..[[<set group="]]..buf.Poll.Question..[["/><radio align="center" id="]]..polln..[[" pad="5" width="150" height="50">|cFFC0C0C0]]..polla..[[|r</radio>]];
							end
						end
						page=page..[[<button align="center" action="submit">Submit</button>]];	-- the "Submit" button
						page=page..[[<set pollid="" pollc="" pollt="" pollm=""/>]];	-- clear poll context
					else -- if answered then show results
						local votecount=0;
						page=page..[[<h2 align="center" pad="10">|cFFC0C0C0]]..buf.Poll.Question..[[|r</h2>]];
						local i,o,high;
						
						for i,o in ipairs(buf.Poll) do -- find index of last option
							--print("i:"..tostring(i).." o:"..tostring(o))
							if (o~="") then
								high=i;
							else break;
							end
						end	

						local bar={}; for i=1,high do bar[i]=0; end	-- init bar sizes
						high=1;                                 	-- minimum max value
						for _,i in pairs(buf.Voted) do	            -- cycle all votes
							--print("blank: "..tostring(_).." i: "..tostring(i));
							votecount=votecount+1;
							bar[i]=bar[i]+1;                     	-- add vote to correct bar
							if (bar[i]>high) then high=bar[i] end	-- update highest vote-count
						end
						-- draw bars on screen
						local loopcount=0;
						for _,i in ipairs(bar) do
							loopcount=loopcount+1;
							--print("Starting Loop "..tostring(loopcount));
							-- get answers again for display --
							for polln,polla in ipairs(buf.Poll) do	-- do numeric indices (skipping the question)
								if (polln==loopcount) then
									--print(buf.Poll[polln]); 
									--page=page..[[<bar pad="20" align="center" empty="#FF222222" fill="Interface\AddOns\MessageBoard\Images\Smoothv2.tga" max="]]..high..[[">]]..i..[[</bar><text font="GameFontLarge" size="16" pad="20" glue="right">]]..buf.Poll[polln]..[[</text><text glue="left" pad="20" font="GameFontHighlightSmall" size="8">|cFFC0C0C0(]]..i..[[)|r</text>]]; --#FF69CCF0
									  page=page..[[<bar width="150" pad="20" align="center" empty="#FF222222" inheritsys="Votingbarborder" fill="Interface\AddOns\MessageBoard\Images\Smoothv2.tga" max="]]..high..[[">]]..i..[[</bar><text font="GameFontLarge" size="16" pad="20" glue="right">|cFFC0C0C0]]..buf.Poll[polln]..[[  (|r|cFF00FF96]]..i..[[|r|cFFC0C0C0)|r</text>]]; --#FF69CCF0
								end
							end							
						end
						page=page..[[<text align="center" pad="20" font="GameFontHighlightSmall" size="8">|cFFC0C0C0 Votes: ]]..votecount..[[|r</text>]];
					end
				end
				local mcolor
				if buf.colorR then
					mcolor = string.format("#FF%02X%02X%02X",buf.colorR*255,buf.colorG*255,buf.colorB*255)
					mcolor = mcolor:sub(2)
				else
					mcolor="FFF5F5F5"
				end
				page=page..[[<text pad="25">|C]]..mcolor..msgTxt..[[|r</text>]];
				-- Buttons Start --
				local buttoncount=0;
				local left;
				local name=UnitName("player");
				name = Ambiguate(name,"none")
				if not(base[category][topic].State==30) then -- if not locked state then
					buttoncount=buttoncount+1;
					if (buf.Author==name or MessageBoard.Administrator) then
						buttoncount=buttoncount+2;
					end
				else
					if (MessageBoard.Administrator) then -- if locked and admin then add edit and delete
						buttoncount=buttoncount+2;
					end
				end
				if (buttoncount==1) then left=245; elseif (buttoncount==2) then left=200; else left=110; end -- show buttons
				page=page..[[<img color="0000" pad="40" height="1" width="]]..left..[["></img>]];
				if not(base[category][topic].State==30) then -- if not locked state then
					page=page..[[<button id="reply" pad="40" glue="left" >]]..L["TOPIC_SCROLLER_REPLY"]..[[</button>]];
					if (buf.Author==name or MessageBoard.Administrator) then -- if message owner or admin add edit and delete
						page=page..[[<button id="edit" pad="17" glue="left">]]..L["TOPIC_SCROLLER_EDIT"]..[[</button>]];
						page=page..[[<button id="delete" pad="17" glue="left">]]..L["TOPIC_SCROLLER_DELETE"]..[[</button>]]; else end
				else
					if (MessageBoard.Administrator) then -- if locked and admin then add edit and delete
						page=page..[[<button id="edit" pad="17" glue="left" >]]..L["TOPIC_SCROLLER_EDIT"]..[[</button>]];
						page=page..[[<button id="delete" pad="17" glue="left">]]..L["TOPIC_SCROLLER_DELETE"]..[[</button>]]; else end
				end
				-- Buttons End --
				page=page..[[<img align="center" pad="40" height="12" width="90%">Interface\FriendsFrame\UI-FriendsFrame-OnlineDivider</img>]];
			MessageBoard:SetReadFlag(category,topic,msg,true);
	end	
	page=page..[[</body></wn>]];
	MessageBoard_Main_MessageScroller:Show();MessageBoard_Main_MessageScroller_Browser_Frame:Show();MessageBoard_CurrentTopic:Show();
	MessageBoard_CurrentTopicText:SetText(ren:toASCII(topic));MessageBoard_CurrentTopic:SetWidth(strlen(ren:toASCII(topic)) * 6  + 40);
	MessageBoard_CurrentTopic:Disable();
	MessageBoardXML:Update()
	DuckLib.Render:Render(MessageBoard_Main_MessageScroller_Browser_Frame,page);
	wipe(msglist)
end

function MessageBoard:SetReadCount(cat,top)
	local base,_,settings=self:ValidateBoardDB(); if (not base) then return nil; end
	base[cat][top].Count=(base[cat][top].Count or 0) + 1;	-- init to zero and add one in one go
end

function MessageBoardXML:SetCategories()
	local base=MessageBoard:ValidateBoardDB(); if (not base) then return nil; end
	local unread=nil;
	local level=nil;
	local unreadcount1=0;
	local unreadcount2=0;
	local unreadcount3=0;
	local unreadcount4=0;
	local unreadcount5=0;
	local unreadcount6=0;
	local unreadcount7=0;
	local unreadcount8=0;
	local unreadcount9=0;
	local readpermission;
	local unreadpermission;
	local lock="Interface\\AddOns\\MessageBoard\\Images\\Lock";
	-- Set Category Buttons Names
	local o=1;
	for index=1,9 do
		_G["MessageBoard_MainNavBarButton"..index.."Text"]:SetText(MessageBoardXML:Setting("Category"..index.."Value") or " ");
		if MessageBoardXML:Setting("Category"..index.."Value") then
			if strlen(MessageBoardXML:Setting("Category"..index.."Value")) > 0 then
				local length = strlen(_G["MessageBoard_MainNavBarButton"..index.."Text"]:GetText())
				if length > 15 then _G["MessageBoard_MainNavBarButton"..index.."Text"]:SetSize(100, 30) end
			else
				_G["MessageBoard_MainNavBarButton"..index]:Hide()
			end
		end
	end
	
	while o<10 do -- Set Category Buttons Icons and Lock Icon if needed
		if MessageBoardXML:Setting("Category"..o.."Icon") then
			_G["MessageBoard_MainNavBarButton"..o.."Icon"]:SetTexture(MessageBoardXML:Setting("Category"..o.."Icon"))
		end
		if MessageBoard:GetCategoryReadPermissions(o) then
			_G["MessageBoard_MainNavBarButton"..o.."LockIcon"]:Hide()
		else
			_G["MessageBoard_MainNavBarButton"..o.."LockIcon"]:Show()
		end
	o=o+1;
	end
	
	local i=1; -- Set Category Button Unread Post Count
	while i<10 do
		_G["MessageBoard_MainNavBarButton"..i.."UnreadBox"]:Hide();
		_G["MessageBoard_MainNavBarButton"..i.."Unread"]:Hide();
		for topic,tTable in pairs(base[i]) do
			if type(tTable)=="database" then
				local _,_,unread=MessageBoard:TopicState(i,topic); -- get unread
				if (i==1 and unread and MessageBoard:GetCategoryReadPermissions(i)) then unreadcount1=unreadcount1+1;
					_G["MessageBoard_MainNavBarButton"..i.."Unread"]:Show();
					_G["MessageBoard_MainNavBarButton"..i.."Unread"]:SetText(unreadcount1);
					_G["MessageBoard_MainNavBarButton"..i.."UnreadBox"]:Show();
				end
				if (i==2 and unread and MessageBoard:GetCategoryReadPermissions(i)) then unreadcount2=unreadcount2+1;
					_G["MessageBoard_MainNavBarButton"..i.."Unread"]:Show();
					_G["MessageBoard_MainNavBarButton"..i.."Unread"]:SetText(unreadcount2);
					_G["MessageBoard_MainNavBarButton"..i.."UnreadBox"]:Show();
				end
				if (i==3 and unread and MessageBoard:GetCategoryReadPermissions(i)) then unreadcount3=unreadcount3+1;
					_G["MessageBoard_MainNavBarButton"..i.."Unread"]:Show();
					_G["MessageBoard_MainNavBarButton"..i.."Unread"]:SetText(unreadcount3);
					_G["MessageBoard_MainNavBarButton"..i.."UnreadBox"]:Show();
				end
				if (i==4 and unread and MessageBoard:GetCategoryReadPermissions(i)) then unreadcount4=unreadcount4+1;
					_G["MessageBoard_MainNavBarButton"..i.."Unread"]:Show();
					_G["MessageBoard_MainNavBarButton"..i.."Unread"]:SetText(unreadcount4);
					_G["MessageBoard_MainNavBarButton"..i.."UnreadBox"]:Show();
				end
				if (i==5 and unread and MessageBoard:GetCategoryReadPermissions(i)) then unreadcount5=unreadcount5+1;
					_G["MessageBoard_MainNavBarButton"..i.."Unread"]:Show();
					_G["MessageBoard_MainNavBarButton"..i.."Unread"]:SetText(unreadcount5);
					_G["MessageBoard_MainNavBarButton"..i.."UnreadBox"]:Show();
				end
				if (i==6 and unread and MessageBoard:GetCategoryReadPermissions(i)) then unreadcount6=unreadcount6+1;
					_G["MessageBoard_MainNavBarButton"..i.."Unread"]:Show();
					_G["MessageBoard_MainNavBarButton"..i.."Unread"]:SetText(unreadcount6);
					_G["MessageBoard_MainNavBarButton"..i.."UnreadBox"]:Show();
				end
				if (i==7 and unread and MessageBoard:GetCategoryReadPermissions(i)) then unreadcount7=unreadcount7+1;
					_G["MessageBoard_MainNavBarButton"..i.."Unread"]:Show();
					_G["MessageBoard_MainNavBarButton"..i.."Unread"]:SetText(unreadcount7);
					_G["MessageBoard_MainNavBarButton"..i.."UnreadBox"]:Show();
				end
				if (i==8 and unread and MessageBoard:GetCategoryReadPermissions(i)) then unreadcount8=unreadcount8+1;
					_G["MessageBoard_MainNavBarButton"..i.."Unread"]:Show();
					_G["MessageBoard_MainNavBarButton"..i.."Unread"]:SetText(unreadcount8);
					_G["MessageBoard_MainNavBarButton"..i.."UnreadBox"]:Show();
				end
				if (i==9 and unread and MessageBoard:GetCategoryReadPermissions(i)) then unreadcount9=unreadcount9+1;
					_G["MessageBoard_MainNavBarButton"..i.."Unread"]:Show();
					_G["MessageBoard_MainNavBarButton"..i.."Unread"]:SetText(unreadcount9);
					_G["MessageBoard_MainNavBarButton"..i.."UnreadBox"]:Show();
				end
		    end
		end
	i=i+1;
	end
end

function MessageBoard:SetReadFlag(cat,top,msg,flag)
	local gn=self:GetGuildName(); if (not gn or gn=="<UnGuilded>") then return nil; end
	--local realm=GetRealmName(); if (not realm) then return nil; end
	if (not MessageBoard_DB) then return nil; end
	--local bn=gn.." - "..realm;
	local bn=gn;
	-- Verify local storage
	if (not MessageBoard_LS[bn]) then MessageBoard_LS[bn]={}; end
	if (not MessageBoard_LS[bn].Base) then MessageBoard_LS[bn].Base={}; end
	if (not MessageBoard_LS[bn].Base[cat]) then MessageBoard_LS[bn].Base[cat]={}; end
	if (not MessageBoard_LS[bn].Base[cat][top]) then MessageBoard_LS[bn].Base[cat][top]={}; end
	MessageBoard_LS[bn].Base[cat][top][msg]=flag;
end

function MessageBoard:GetReadFlag(cat,top,msg)
	local gn=self:GetGuildName(); if (not gn or gn=="<UnGuilded>") then return nil; end
	--local realm=GetRealmName(); if (not realm) then return nil; end
	if (not MessageBoard_DB) then return nil; end
	--local bn=gn.." - "..realm;
	local bn=gn;
	-- Verify local storage
	if (not MessageBoard_LS[bn]) then return nil; end
	if (not MessageBoard_LS[bn].Base) then return nil; end
	if (not MessageBoard_LS[bn].Base[cat]) then return nil; end
	if (not MessageBoard_LS[bn].Base[cat][top]) then return nil; end
	return MessageBoard_LS[bn].Base[cat][top][msg];
end

function MessageBoard:SetPrunes()
	MessageBoardXML:Debug("Init Prune Settings");
	local base,avatar,_=MessageBoard:ValidateBoardDB(); if (not base) then return nil; end
	local bn=MessageBoard:ValidateBoard(true);
	local limit=MessageBoardXML:Setting("MessageLimit") or 1000;
	MessageBoard_DB[bn]:DM_Set("Limit",5000);
	for category,cTable in pairs(base) do
		for topic,tTable in pairs(cTable) do
			if (type(tTable)=="database" and tTable.State==0) then -- normal
				tTable:DM_PruneMode(DM_PRUNE_FIRST+10);
			elseif (type(tTable)=="database" and tTable.State==40) then -- poll
				tTable:DM_PruneMode(DM_PRUNE_FIRST+11);
			elseif (type(tTable)=="database" and tTable.State==10) then -- sticky
				tTable:DM_PruneMode(DM_PRUNE_FIRST+12);
			elseif (type(tTable)=="database" and tTable.State==20) then -- announcement
				tTable:DM_PruneMode(DM_PRUNE_FIRST+13);
			elseif (type(tTable)=="database" and tTable.State==30) then -- locked
				tTable:DM_PruneMode(DM_PRUNE_FIRST+14);
			end
		end
	end	
	avatar:DM_PruneMode(DM_PRUNE_NEVER);
	MessageBoard:ValidateBoard():DM_PruneMode(DM_PRUNE_NEVER);
	base:DM_PruneMode(DM_PRUNE_NEVER);
end

function MBFunc.MBGuildFrame_UpdateTabard()
	MessageBoardXML:Debug("Updating Guild Emblem");
	MBFunc.MBSetLargeGuildTabardTextures("player", MessageBoard_MainTabardEmblem, MessageBoard_MainTabardBackground, MessageBoard_MainTabardBorder);
	MBFunc.MBSetLargeGuildTabardTextures("player", MessageBoard_EditorTabardEmblem, MessageBoard_EditorTabardBackground, MessageBoard_EditorTabardBorder);
	MBFunc.MBSetLargeGuildTabardTextures("player", SideTabardEmblem, SideTabardBackground, SideTabardBorder);
	MessageBoard_EditorTabardEmblem:SetAlpha(.10);
	SideTabardEmblem:SetAlpha(1.0);
end

function MBFunc.MBSetLargeGuildTabardTextures(unit, emblemTexture, backgroundTexture, borderTexture, tabardData)
	-- texure dimensions are 1024x1024, icon dimensions are 64x64
	local emblemSize, columns, offset;
	if ( emblemTexture ) then
		emblemSize = 64 / 1024;
		columns = 16
		offset = 0;
		emblemTexture:SetTexture("Interface\\GuildFrame\\GuildEmblemsLG_01");
	end
	MBFunc.MBSetGuildTabardTextures(emblemSize, columns, offset, unit, emblemTexture, backgroundTexture, borderTexture, tabardData);
end

function MBFunc.MBSetSmallGuildTabardTextures(unit, emblemTexture, backgroundTexture, borderTexture, tabardData)
	-- texure dimensions are 256x256, icon dimensions are 16x16, centered in 18x18 cells
	local emblemSize, columns, offset;
	if ( emblemTexture ) then
		emblemSize = 18 / 256;
		columns = 14;
		offset = 1 / 256;
		emblemTexture:SetTexture("Interface\\GuildFrame\\GuildEmblems_01");
	end
	MBFunc.MBSetGuildTabardTextures(emblemSize, columns, offset, unit, emblemTexture, backgroundTexture, borderTexture, tabardData);
end

function MBFunc.MBSetDoubleGuildTabardTextures(unit, leftEmblemTexture, rightEmblemTexture, backgroundTexture, borderTexture, tabardData)
	if ( leftEmblemTexture and rightEmblemTexture ) then
		SetGuildTabardTextures(nil, nil, nil, unit, leftEmblemTexture, backgroundTexture, borderTexture, tabardData);
		rightEmblemTexture:SetTexture(leftEmblemTexture:GetTexture());
		rightEmblemTexture:SetVertexColor(leftEmblemTexture:GetVertexColor());
	end
end

function MBFunc.MBSetGuildTabardTextures(emblemSize, columns, offset, unit, emblemTexture, backgroundTexture, borderTexture, tabardData)
	local bkgR, bkgG, bkgB, borderR, borderG, borderB, emblemR, emblemG, emblemB, emblemFilename;
	if ( tabardData ) then
		bkgR = tabardData[1];
		bkgG = tabardData[2];
		bkgB = tabardData[3];
		borderR = tabardData[4];
		borderG = tabardData[5];
		borderB = tabardData[6];
		emblemR = tabardData[7];
		emblemG = tabardData[8];
		emblemB = tabardData[9];
		emblemFilename = tabardData[10];
	else
		bkgR, bkgG, bkgB, borderR, borderG, borderB, emblemR, emblemG, emblemB, emblemFilename = GetGuildLogoInfo(unit);
	end
	if ( emblemFilename ) then
		if ( backgroundTexture ) then
			backgroundTexture:SetVertexColor(bkgR / 255, bkgG / 255, bkgB / 255);
		end
		if ( borderTexture ) then
			borderTexture:SetVertexColor(borderR / 255, borderG / 255, borderB / 255);
		end
		if ( emblemSize ) then
			local index = emblemFilename:match("([%d]+)");
			if ( index) then
				index = tonumber(index);
				local xCoord = mod(index, columns) * emblemSize;
				local yCoord = floor(index / columns) * emblemSize;
				emblemTexture:SetTexCoord(xCoord + offset, xCoord + emblemSize - offset, yCoord + offset, yCoord + emblemSize - offset);
			end
			emblemTexture:SetVertexColor(emblemR / 255, emblemG / 255, emblemB / 255);
		elseif ( emblemTexture ) then
			emblemTexture:SetTexture(emblemFilename);
			emblemTexture:SetVertexColor(emblemR / 255, emblemG / 255, emblemB / 255);
		end
	else
		-- tabard lacks design
		if ( backgroundTexture ) then
			backgroundTexture:SetVertexColor(0.2245, 0.2088, 0.1794);
		end
		if ( borderTexture ) then
			borderTexture:SetVertexColor(0.2, 0.2, 0.2);
		end
		if ( emblemTexture ) then
			if ( emblemSize ) then
				if ( emblemSize == 18 / 256 ) then
					emblemTexture:SetTexture("Interface\\GuildFrame\\GuildLogo-NoLogoSm");
				else
					emblemTexture:SetTexture("Interface\\GuildFrame\\GuildLogo-NoLogo");
				end
				emblemTexture:SetTexCoord(0, 1, 0, 1);
				emblemTexture:SetVertexColor(1, 1, 1, 1);
			else
				emblemTexture:SetTexture("");
			end
		end
	end
end

Toast:Register("MessageBoard_Toast", function(toast, ...)
	toast:SetTitle("MessageBoard");
	toast:SetText(...);
	toast:SetIconTexture([[Interface\Icons\Inv_holiday_brewfestbuff_01]]);
end)

function MBFunc.SpawnToast(source, text, r, g, b, ...)
	Toast:Spawn("MessageBoard_Toast", text);
end

function MBFunc.tprint (tbl, indent)
  if not indent then indent = 0 end
  local formatting
  for k, v in pairs(tbl) do
    formatting = string.rep(" ", indent) .. k .. ": "
    if type(v) == "table" then
      print(formatting)
      MBFunc.tprint(v, indent+1)
    else
      print(formatting .. v)
    end
  end
end

function MBFunc.isBlank(x)
  return not not tostring(x):find("^%s*$")
end

function MessageBoardXML:Setting(key)
    local _,_,settings=MessageBoardXML:ValidateBoardDB();
    if (type(settings)=="string") then return unpacktable(settings)[key]; end   -- packed table
    return settings[key];      													-- "old" database
end

function MessageBoard:CheckDatabase()
	local root=MessageBoard:ValidateBoard();
	local gn=MessageBoard:ValidateBoard(true);		-- Get names
	-- Do local database
	if (not root or type(root.Base)~="database") then
		-- Missing message database. Recreating everything from scratch.
		-- Check for real databases. Unregister if registered.
		if (type(MessageBoard_DB[gn])=="database") then setmetatable(MessageBoard_DB[gn],nil); end
		-- Wipe tables if any
		if (type(MessageBoard_DB[gn])=="table") then wipe(MessageBoard_DB[gn]); end
		MessageBoard_DB[gn]=nil;		-- unlink storage
		MessageBoard:ValidateBoard();	-- Create new database
		return;							-- No need to check a fresh database
	end
	if (type(root.Avatar)~="database") then
										-- Missing avatar database. Creating new.
		root.Avatar={};					-- Create blank database
	end
	-- Verify Root
	for _,c in pairs(root) do
		if strfind(_, "auk") then
			self:CheckDatabase_Cleaner(root,_);
		end
	end
	-- Verify categories
	for e,d in pairs(root.Base) do
		if (type(e)~="number" or e<1 or e>9) then		-- Kill anything with a wrong name
			self:CheckDatabase_Cleaner(root.Base,e);	-- Clean it up
		end
	end
    for e=1,9 do
        if (type(root.Base[e])~="database") then -- Kill anything with a wrong name
            self:CheckDatabase_Cleaner(root.Base,e); -- Clean it up
            root.Base[e]={}; -- re-insert it
        end
    end
	-- verify topics
	for _,c in pairs(root.Base) do
		for e,t in pairs(c) do
			if (type(t)~="database") then
				self:CheckDatabase_Cleaner(c,e);	-- Clean it up
			end
		end
	end
	 -- Hooking... Do NOT return before it has been restored!
    MessageBoard.PE_Original=DuckLib.Table.PullEntry;
    DuckLib.Table.PullEntry=MessageBoard.DM_Table_PullEntry;

    if (type(root.Settings)~="string") then -- Not a packed string
        root.Settings=packtable({}); -- Create blank
    end
    MessageBoard.MessageErrorDetected=nil; -- No error yet
    local buf=unpacktable(root.Settings); -- Unpack this message
    if (MessageBoard.MessageErrorDetected) then -- Broken message
	if (not MessageBoard_DB._PTE) then MessageBoard_DB._PTE={}; end table.insert(MessageBoard_DB._PTE,root.Settings);
        root.Settings=packtable({}); -- Create blank
    end

	-- verify topic integrity
	for _,c in pairs(root.Base) do				-- Get categories
		for _,t in pairs(c) do					-- Get topics
			for e,m in pairs(t) do				-- Get messages, e = auk, m = text
				local cleanit=nil;				-- No cleaning yet
				if e~="State" and not strfind(e, "auk") then
			       --print("we found an error in "..tostring(e))
				   cleanit=true;
				end
				if t.State==nil then cleanit=true; end
				if (type(m)=="number") then		-- Stored data is number
					if (e~="State") then cleanit=true; end	-- Clean it up if it's not "State"
				elseif (type(m)=="string") then	-- Messages are strings at this level
					if (m:sub(1,2)~=(DuckLib.Table.Version..DuckLib.Table.ThisVersion)) then
					else
						MessageBoard.MessageErrorDetected=nil;		-- No error yet
						local buf=unpacktable(m);					-- Unpack this message
						if (MessageBoard.MessageErrorDetected) then	-- Broken message
							cleanit=true;
						else
						end
					end
				else cleanit=true; end			-- Invalid data type
				if (cleanit) then self:CheckDatabase_Cleaner(t,e); end	-- Clean it up
			end
		end
	end
	MessageBoard.MessageErrorDetected=nil;

	-- Restore hook
	DuckLib.Table.PullEntry=MessageBoard.PE_Original;
--	collectgarbage("collect");
end

function MessageBoard:DM_Table_PullEntry(...)
	local r1,r2,r3=MessageBoard.PE_Original(DuckLib.Table,...);
	if (not r1) then r1=""; r3=""; MessageBoard.MessageErrorDetected=true; end
	if (not r2) then r2=""; r3=""; MessageBoard.MessageErrorDetected=true; end
	if (not r3) then r3=""; MessageBoard.MessageErrorDetected=true; end
	return r1,r2,r3;
end

function MessageBoard:CheckDatabase_Cleaner(db,e)
	DuckLib.Database.Util.LockedTime=0;		-- Deactivate sync
	db[e]=nil;							-- Invoke database internal clean-up
	DuckLib.Database.Util.LockedTime=nil;	-- Reactivate sync
	rawset(db,e,nil);					-- Wipe the rest
end

function MessageBoardCategory_OnClick(self,button,down)
	local newId = self:GetID()
	local oldId = CategoryLevel
	local readpermission=MessageBoard:GetCategoryReadPermissions(newId)
	if readpermission then
		if ( oldId ~= newId ) then
			selectedEntryID = nil
		end
		selectedPageNum = 1
		MessageBoard_SetSelectedButton(self,newId)
		MessageBoardXML:SetCategoryLevel(newId)
		PlaySound("UI_igStore_PageNav_Button")
		if button=="RightButton" then
			if MessageBoardXML.Administrator then
				local menu=DMMenuCreate(self)
				menu.Timer=0.1;    -- value in seconds
				menu:Add("Enter Password",function () MessageBoardXML:SetPassword(newId) end)
				menu:Show()
			end
		end
	end
end

function MessageBoard_SetSelectedButton(self,id,button)
	for i = 1,9 do
		_G["MessageBoard_MainNavBarButton"..i].SelectedBG:Hide()
	end
	self.SelectedBG:Show()
end

-- Set Category Level
function MessageBoardXML:SetCategoryLevel(category)
	if MessageBoard:GetCategoryReadPermissions(category) then
		CategoryLevel=category;	MessageBoard.Selected.Topic=nil; MessageBoardXML:ViewTopics();
	else DuckLibMB:Chat("MessageBoard: |cFFFFFFFF"..L["NOT_ALLOWED_ACCESS"],0,0.9,0.9); end
end

function MessageBoardXML:MakeBool(value)
	if (type(value)=="boolean") then return value; end
	if (not value) then return nil; end
	if (type(value)=="string") then value=tonumber(value); end
	if (type(value)=="number") then
		if (value==0) then return nil; end
		return true;
	end

	return nil;
end

function MessageBoardXML:PrintProgress()
	local base=MessageBoard:ValidateBoardDB();
    print("Inbound:",ceil(base:DM_ProgressIB()*100));
end

function MessageBoardXML:CheckV()
	vface:Send(nil,"SendVersion");
	local me=tostring(UnitName("player"))
	me = Ambiguate(me,"none")
	DuckLibMB:Chat("MessageBoard: |cFFFFFFFF"..L["VER_CHECK_TITLE"],0,0.9,0.9)
	DuckLibMB:Chat("MessageBoard: |cFFFFFFFF ---------------------------------------",0,0.9,0.9)
	DuckLibMB:Chat("MessageBoard: |cFFFFFFFF"..me..L["VER_CHECK_USER"].."|cFFFFFFFF"..iVersion,0,0.9,0.9)
end

local VersionPage
local verTable={}
function MBverRX(iface,data,name,_,toon)    -- handler for incoming communication
	Ambiguate(toon,"none")
    if (name=="SendVersion") then           -- the entry used for version request
        iface:Send(iVersion,"Version");    -- assuming variable "iVersion" contains your own version
	end
	if (name=="Version") then               -- The name of the data in this example is "Version", so the data should be version information for "toon".
		verTable[toon]=data
		for vtoon,vdata in pairs(verTable) do
			DuckLibMB:Chat("MessageBoard: |cFFFFFFFF"..vtoon..L["VER_CHECK_USER"].."|cFFFFFFFF"..vdata,0,0.9,0.9);
		end
	end
end


-- Messageboard Minimap Button

function MBFunc.InitializeMinimap()
    _MBLDB = LibStub("LibDataBroker-1.1"):NewDataObject("MessageBoard", {
        type = "launcher",
        text = "MessageBoard",
        icon = "Interface\\Buttons\\UI-GuildButton-PublicNote-Disabled",
        OnClick = function()            
            MessageBoard:ToggleMainFrame()
        end,
        OnTooltipShow = function(tt)
            tt:AddLine("MessageBoard", 1, 1, 1);
        end	
    });
    MBFunc.MBUpdateMinimap()
end

function MBFunc.MBUpdateMinimap()
	if not _minimapIcon then 
		_minimapIcon = LibStub("LibDBIcon-1.0")
        _minimapIcon:Register("MessageBoard", _MBLDB, MessageBoardLDBIconDB)
	end     
	local count = MessageBoardXML:GetNewMsgCount() or 0
	if count > 0 then
		_MBLDB.icon = 'Interface\\Buttons\\UI-GuildButton-PublicNote-Up'
		_MBLDB.OnTooltipShow = function(tt)
            tt:AddLine("MessageBoard", 1, 1, 1);
            tt:AddLine(" ");
            tt:AddLine(count.." new messages")
        end	
	else
		_MBLDB.icon = 'Interface\\Buttons\\UI-GuildButton-PublicNote-Disabled'
		_MBLDB.OnTooltipShow = function(tt)
            tt:AddLine("MessageBoard", 1, 1, 1);
            tt:AddLine(" ");
            tt:AddLine("No new messages")
        end	
	end
	
	local icon = LibStub("LibDBIcon-1.0", true)
	if MessageBoardLDBIconDB.hide then
		icon:Show("MessageBoard")
	else
		icon:Hide("MessageBoard")
	end
end